@@ -28,22 +28,19 @@ static void flattenOutline(NSMutableArray *titles, NSMutableArray *pages, fz_out
 	indent[level * 4] = 0;
 	while (outline)
 	{
-		if (outline->dest.kind == FZ_LINK_GOTO)
+		int page = outline->page;
+		if (page >= 0 && outline->title)
 		{
-			int page = outline->dest.ld.gotor.page;
-			if (page >= 0 && outline->title)
-			{
-				NSString *title = [NSString stringWithUTF8String: outline->title];
-				[titles addObject: [NSString stringWithFormat: @"%s%@", indent, title]];
-				[pages addObject: [NSNumber numberWithInt: page]];
-			}
+			NSString *title = @(outline->title);
+			[titles addObject: [NSString stringWithFormat: @"%s%@", indent, title]];
+			[pages addObject: @(page)];
 		}
 		flattenOutline(titles, pages, outline->down, level + 1);
 		outline = outline->next;
 	}
 }
 
-static char *tmp_path(char *path)
+static char *tmp_path(const char *path)
 {
 	int f;
 	char *buf = malloc(strlen(path) + 6 + 1);
@@ -67,7 +64,7 @@ static char *tmp_path(char *path)
 	}
 }
 
-static void saveDoc(char *current_path, fz_document *doc)
+static void saveDoc(const char *current_path, fz_document *doc)
 {
 	char *tmp;
 	pdf_document *idoc = pdf_specifics(ctx, doc);
@@ -129,7 +126,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 	fz_document *doc;
 	MuDocRef *docRef;
 	NSString *key;
-	char *filePath;
+	NSString *_filePath;
 	BOOL reflowMode;
 	MuOutlineController *outline;
 	UIScrollView *canvas;
@@ -158,9 +155,9 @@ static void saveDoc(char *current_path, fz_document *doc)
 	BOOL _isRotating;
 }
 
-- (id) initWithFilename: (NSString*)filename path:(char *)cstr document: (MuDocRef *)aDoc
+- (instancetype) initWithFilename: (NSString*)filename path:(NSString *)cstr document: (MuDocRef *)aDoc
 {
-	self = [super init];
+	self = [super initWithNibName:nil bundle:nil];
 	if (!self)
 		return nil;
 
@@ -168,25 +165,16 @@ static void saveDoc(char *current_path, fz_document *doc)
 	if ([self respondsToSelector:@selector(automaticallyAdjustsScrollViewInsets)])
 		self.automaticallyAdjustsScrollViewInsets = NO;
 #endif
-	key = [filename retain];
+	key = [filename copy];
 	docRef = [aDoc retain];
 	doc = docRef->doc;
-	filePath = strdup(cstr);
+	_filePath = [cstr copy];
+
+	//  this will be created right before the outline is shown
+	outline = nil;
 
 	dispatch_sync(queue, ^{});
 
-	fz_outline *root = fz_load_outline(ctx, doc);
-	if (root) {
-		NSMutableArray *titles = [[NSMutableArray alloc] init];
-		NSMutableArray *pages = [[NSMutableArray alloc] init];
-		flattenOutline(titles, pages, root, 0);
-		if ([titles count])
-			outline = [[MuOutlineController alloc] initWithTarget: self titles: titles pages: pages];
-		[titles release];
-		[pages release];
-		fz_drop_outline(ctx, root);
-	}
-
 	return self;
 }
 
@@ -199,7 +187,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 	else
 	{
 		UIView *buttonView;
-		BOOL iOS7Style = ([[UIDevice currentDevice].systemVersion floatValue] >= 7.0f);
+		BOOL iOS7Style = (([UIDevice currentDevice].systemVersion).floatValue >= 7.0f);
 		UIButton *button = [UIButton buttonWithType:iOS7Style ? UIButtonTypeSystem : UIButtonTypeCustom];
 		[button setImage:[UIImage imageNamed:resource] forState:UIControlStateNormal];
 		[button addTarget:self action:selector forControlEvents:UIControlEventTouchUpInside];
@@ -219,8 +207,8 @@ static void saveDoc(char *current_path, fz_document *doc)
 		[array addObject:outlineButton];
 	[array addObject:reflowButton];
 	[array addObject:linkButton];
-	[[self navigationItem] setRightBarButtonItems: array ];
-	[[self navigationItem] setLeftBarButtonItem:backButton];
+	self.navigationItem.rightBarButtonItems = array ;
+	self.navigationItem.leftBarButtonItem = backButton;
 }
 
 - (void) loadView
@@ -232,16 +220,16 @@ static void saveDoc(char *current_path, fz_document *doc)
 		current = 0;
 
 	UIView *view = [[UIView alloc] initWithFrame: CGRectZero];
-	[view setAutoresizingMask: UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight];
+	view.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
 	[view setAutoresizesSubviews: YES];
 	view.backgroundColor = [UIColor grayColor];
 
 	canvas = [[UIScrollView alloc] initWithFrame: CGRectMake(0,0,GAP,0)];
-	[canvas setAutoresizingMask: UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight];
+	canvas.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
 	[canvas setPagingEnabled: YES];
 	[canvas setShowsHorizontalScrollIndicator: NO];
 	[canvas setShowsVerticalScrollIndicator: NO];
-	[canvas setDelegate: self];
+	canvas.delegate = self;
 
 	UITapGestureRecognizer *tapRecog = [[UITapGestureRecognizer alloc] initWithTarget: self action: @selector(onTap:)];
 	tapRecog.delegate = self;
@@ -259,34 +247,43 @@ static void saveDoc(char *current_path, fz_document *doc)
 	scroll_animating = NO;
 
 	indicator = [[UILabel alloc] initWithFrame: CGRectZero];
-	[indicator setAutoresizingMask: UIViewAutoresizingFlexibleLeftMargin | UIViewAutoresizingFlexibleRightMargin | UIViewAutoresizingFlexibleTopMargin];
-	[indicator setText: @"0000 of 9999"];
+	indicator.autoresizingMask = UIViewAutoresizingFlexibleLeftMargin | UIViewAutoresizingFlexibleRightMargin | UIViewAutoresizingFlexibleTopMargin;
+	indicator.text = @"0000 of 9999";
 	[indicator sizeToFit];
-	[indicator setCenter: CGPointMake(0, INDICATOR_Y)];
-	[indicator setTextAlignment: NSTextAlignmentCenter];
-	[indicator setBackgroundColor: [[UIColor blackColor] colorWithAlphaComponent: 0.5]];
-	[indicator setTextColor: [UIColor whiteColor]];
+	indicator.center = CGPointMake(0, INDICATOR_Y);
+	indicator.textAlignment = NSTextAlignmentCenter;
+	indicator.backgroundColor = [[UIColor blackColor] colorWithAlphaComponent: 0.5];
+	indicator.textColor = [UIColor whiteColor];
 
 	[view addSubview: canvas];
 	[view addSubview: indicator];
 
 	slider = [[UISlider alloc] initWithFrame: CGRectZero];
-	[slider setMinimumValue: 0];
-	[slider setMaximumValue: fz_count_pages(ctx, doc) - 1];
+	slider.minimumValue = 0;
+	slider.maximumValue = fz_count_pages(ctx, doc) - 1;
 	[slider addTarget: self action: @selector(onSlide:) forControlEvents: UIControlEventValueChanged];
 
-	if ([[[UIDevice currentDevice] systemVersion] floatValue] < 7.0)
+	if ([UIDevice currentDevice].systemVersion.floatValue < 7.0)
 	{
 		sliderWrapper = [[UIBarButtonItem alloc] initWithCustomView: slider];
 
-		[self setToolbarItems: [NSArray arrayWithObjects: sliderWrapper, nil]];
+		self.toolbarItems = @[sliderWrapper];
 	}
 
 	// Set up the buttons on the navigation and search bar
 
-	if (outline) {
+	fz_outline *outlineRoot;
+	fz_try(ctx)
+		outlineRoot = fz_load_outline(ctx, doc);
+	fz_catch(ctx)
+		outlineRoot = NULL;
+	if (outlineRoot)
+	{
+		//  only show the outline button if there is an outline
 		outlineButton = [self newResourceBasedButton:@"ic_list" withAction:@selector(onShowOutline:)];
+		fz_drop_outline(ctx, outlineRoot);
 	}
+
 	linkButton = [self newResourceBasedButton:@"ic_link" withAction:@selector(onToggleLinks:)];
 	cancelButton = [self newResourceBasedButton:@"ic_cancel" withAction:@selector(onCancel:)];
 	searchButton = [self newResourceBasedButton:@"ic_magnifying_glass" withAction:@selector(onShowSearch:)];
@@ -305,8 +302,8 @@ static void saveDoc(char *current_path, fz_document *doc)
 	deleteButton = [self newResourceBasedButton:@"ic_trash" withAction:@selector(onDelete:)];
 	searchBar = [[UISearchBar alloc] initWithFrame: CGRectMake(0,0,50,32)];
 	backButton = [self newResourceBasedButton:@"ic_arrow_left" withAction:@selector(onBack:)];
-	[searchBar setPlaceholder: @"Search"];
-	[searchBar setDelegate: self];
+	searchBar.placeholder = @"Search";
+	searchBar.delegate = self;
 
 	[prevButton setEnabled: NO];
 	[nextButton setEnabled: NO];
@@ -315,7 +312,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 	// TODO: add activityindicator to search bar
 
-	[self setView: view];
+	self.view = view;
 	[view release];
 }
 
@@ -345,7 +342,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 	[tickButton release]; tickButton = nil;
 	[deleteButton release]; deleteButton = nil;
 	[canvas release]; canvas = nil;
-	free(filePath); filePath = NULL;
+	[_filePath release]; _filePath = NULL;
 
 	[outline release];
 	[key release];
@@ -355,55 +352,55 @@ static void saveDoc(char *current_path, fz_document *doc)
 - (void) viewWillAppear: (BOOL)animated
 {
 	[super viewWillAppear:animated];
-	[self setTitle: [key lastPathComponent]];
+	self.title = key.lastPathComponent;
 
-	[slider setValue: current];
+	slider.value = current;
 
-	if ([[[UIDevice currentDevice] systemVersion] floatValue] >= 7.0)
-		[[[self navigationController] toolbar] addSubview:slider];
+	if ([UIDevice currentDevice].systemVersion.floatValue >= 7.0)
+		[self.navigationController.toolbar addSubview:slider];
 
-	[indicator setText: [NSString stringWithFormat: @" %d of %d ", current+1, fz_count_pages(ctx, doc)]];
+	indicator.text = [NSString stringWithFormat: @" %d of %d ", current+1, fz_count_pages(ctx, doc)];
 
-	[[self navigationController] setToolbarHidden: NO animated: animated];
+	[self.navigationController setToolbarHidden: NO animated: animated];
 }
 
 - (void) viewWillLayoutSubviews
 {
-	CGSize size = [canvas frame].size;
+	CGSize size = canvas.frame.size;
 	int max_width = fz_max(width, size.width);
 
 	width = size.width;
 	height = size.height;
 
-	[canvas setContentInset: UIEdgeInsetsZero];
-	[canvas setContentSize: CGSizeMake(fz_count_pages(ctx, doc) * width, height)];
-	[canvas setContentOffset: CGPointMake(current * width, 0)];
+	canvas.contentInset = UIEdgeInsetsZero;
+	canvas.contentSize = CGSizeMake(fz_count_pages(ctx, doc) * width, height);
+	canvas.contentOffset = CGPointMake(current * width, 0);
 
 	[sliderWrapper setWidth: SLIDER_W];
-	[searchBar setFrame: CGRectMake(0,0,SEARCH_W,32)];
-	if ([[[UIDevice currentDevice] systemVersion] floatValue] >= 7.0)
+	searchBar.frame = CGRectMake(0,0,SEARCH_W,32);
+	if ([UIDevice currentDevice].systemVersion.floatValue >= 7.0)
 	{
-		CGRect r = [[self navigationController] toolbar].frame;
+		CGRect r = self.navigationController.toolbar.frame;
 		r.origin.x = 0;
 		r.origin.y = 0;
-		[slider setFrame:r];
+		slider.frame = r;
 	}
 
-	[[[self navigationController] toolbar] setNeedsLayout]; // force layout!
+	[self.navigationController.toolbar setNeedsLayout]; // force layout!
 
 	// use max_width so we don't clamp the content offset too early during animation
-	[canvas setContentSize: CGSizeMake(fz_count_pages(ctx, doc) * max_width, height)];
-	[canvas setContentOffset: CGPointMake(current * width, 0)];
+	canvas.contentSize = CGSizeMake(fz_count_pages(ctx, doc) * max_width, height);
+	canvas.contentOffset = CGPointMake(current * width, 0);
 
-	for (UIView<MuPageView> *view in [canvas subviews]) {
-		if ([view number] == current) {
-			[view setFrame: CGRectMake([view number] * width, 0, width-GAP, height)];
+	for (UIView<MuPageView> *view in canvas.subviews) {
+		if (view.number == current) {
+			view.frame = CGRectMake(view.number * width, 0, width-GAP, height);
 			[view willRotate];
 		}
 	}
-	for (UIView<MuPageView> *view in [canvas subviews]) {
-		if ([view number] != current) {
-			[view setFrame: CGRectMake([view number] * width, 0, width-GAP, height)];
+	for (UIView<MuPageView> *view in canvas.subviews) {
+		if (view.number != current) {
+			view.frame = CGRectMake(view.number * width, 0, width-GAP, height);
 			[view willRotate];
 		}
 	}
@@ -418,34 +415,34 @@ static void saveDoc(char *current_path, fz_document *doc)
 - (void) viewWillDisappear: (BOOL)animated
 {
 	[super viewWillDisappear:animated];
-	if ([[[UIDevice currentDevice] systemVersion] floatValue] >= 7.0)
+	if ([UIDevice currentDevice].systemVersion.floatValue >= 7.0)
 		[slider removeFromSuperview];
 
-	[self setTitle: @"Resume"];
+	self.title = @"Resume";
 	[[NSUserDefaults standardUserDefaults] removeObjectForKey: @"OpenDocumentKey"];
-	[[self navigationController] setToolbarHidden: YES animated: animated];
+	[self.navigationController setToolbarHidden: YES animated: animated];
 }
 
 - (void) showNavigationBar
 {
-	if ([[self navigationController] isNavigationBarHidden]) {
+	if (self.navigationController.navigationBarHidden) {
 		[sliderWrapper setWidth: SLIDER_W];
-		if ([[[UIDevice currentDevice] systemVersion] floatValue] >= 7.0)
+		if ([UIDevice currentDevice].systemVersion.floatValue >= 7.0)
 		{
-			CGRect r = [[self navigationController] toolbar].frame;
+			CGRect r = self.navigationController.toolbar.frame;
 			r.origin.x = 0;
 			r.origin.y = 0;
-			[slider setFrame:r];
+			slider.frame = r;
 		}
-		[[self navigationController] setNavigationBarHidden: NO];
-		[[self navigationController] setToolbarHidden: NO];
+		[self.navigationController setNavigationBarHidden: NO];
+		[self.navigationController setToolbarHidden: NO];
 		[indicator setHidden: NO];
 
 		[UIView beginAnimations: @"MuNavBar" context: NULL];
 
-		[[[self navigationController] navigationBar] setAlpha: 1];
-		[[[self navigationController] toolbar] setAlpha: 1];
-		[indicator setAlpha: 1];
+		self.navigationController.navigationBar.alpha = 1;
+		self.navigationController.toolbar.alpha = 1;
+		indicator.alpha = 1;
 
 		[UIView commitAnimations];
 	}
@@ -453,16 +450,16 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) hideNavigationBar
 {
-	if (![[self navigationController] isNavigationBarHidden]) {
+	if (!self.navigationController.navigationBarHidden) {
 		[searchBar resignFirstResponder];
 
 		[UIView beginAnimations: @"MuNavBar" context: NULL];
 		[UIView setAnimationDelegate: self];
 		[UIView setAnimationDidStopSelector: @selector(onHideNavigationBarFinished)];
 
-		[[[self navigationController] navigationBar] setAlpha: 0];
-		[[[self navigationController] toolbar] setAlpha: 0];
-		[indicator setAlpha: 0];
+		self.navigationController.navigationBar.alpha = 0;
+		self.navigationController.toolbar.alpha = 0;
+		indicator.alpha = 0;
 
 		[UIView commitAnimations];
 	}
@@ -470,20 +467,41 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) onHideNavigationBarFinished
 {
-	[[self navigationController] setNavigationBarHidden: YES];
-	[[self navigationController] setToolbarHidden: YES];
+	[self.navigationController setNavigationBarHidden: YES];
+	[self.navigationController setToolbarHidden: YES];
 	[indicator setHidden: YES];
 }
 
 - (void) onShowOutline: (id)sender
 {
-	[[self navigationController] pushViewController: outline animated: YES];
+	//  rebuild the outline in case the layout has changed
+	fz_outline *root;
+	fz_try(ctx)
+		root = fz_load_outline(ctx, doc);
+	fz_catch(ctx)
+		root = NULL;
+	if (root)
+	{
+		NSMutableArray *titles = [[NSMutableArray alloc] init];
+		NSMutableArray *pages = [[NSMutableArray alloc] init];
+		flattenOutline(titles, pages, root, 0);
+		[outline release];
+		if (titles.count)
+			outline = [[MuOutlineController alloc] initWithTarget: self titles: titles pages: pages];
+		[titles release];
+		[pages release];
+		fz_drop_outline(ctx, root);
+	}
+
+	//  now show it
+
+	[self.navigationController pushViewController: outline animated: YES];
 }
 
 - (void) onToggleLinks: (id)sender
 {
 	showLinks = !showLinks;
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
 		if (showLinks)
 			[view showLinks];
@@ -496,12 +514,12 @@ static void saveDoc(char *current_path, fz_document *doc)
 {
 	reflowMode = !reflowMode;
 
-	[annotButton setEnabled:!reflowMode];
-	[searchButton setEnabled:!reflowMode];
-	[linkButton setEnabled:!reflowMode];
-	[moreButton setEnabled:!reflowMode];
+	annotButton.enabled = !reflowMode;
+	searchButton.enabled = !reflowMode;
+	linkButton.enabled = !reflowMode;
+	moreButton.enabled = !reflowMode;
 
-	[[canvas subviews] makeObjectsPerformSelector:@selector(removeFromSuperview)];
+	[canvas.subviews makeObjectsPerformSelector:@selector(removeFromSuperview)];
 	[self scrollViewDidScroll:canvas];
 }
 
@@ -510,20 +528,20 @@ static void saveDoc(char *current_path, fz_document *doc)
 	NSMutableArray *rightbuttons = [NSMutableArray arrayWithObjects:printButton, shareButton, nil];
 	if (docRef->interactive)
 		[rightbuttons insertObject:annotButton atIndex:0];
-	[[self navigationItem] setRightBarButtonItems:rightbuttons];
-	[[self navigationItem] setLeftBarButtonItem:cancelButton];
+	self.navigationItem.rightBarButtonItems = rightbuttons;
+	self.navigationItem.leftBarButtonItem = cancelButton;
 
 	barmode = BARMODE_MORE;
 }
 
 - (void) showAnnotationMenu
 {
-	[[self navigationItem] setRightBarButtonItems:[NSArray arrayWithObjects:inkButton, strikeoutButton, underlineButton, highlightButton, nil]];
-	[[self navigationItem] setLeftBarButtonItem:cancelButton];
+	self.navigationItem.rightBarButtonItems = @[inkButton, strikeoutButton, underlineButton, highlightButton];
+	self.navigationItem.leftBarButtonItem = cancelButton;
 
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
-		if ([view number] == current)
+		if (view.number == current)
 			[view deselectAnnotation];
 	}
 
@@ -532,7 +550,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) update
 {
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 		[view update];
 }
 
@@ -575,8 +593,8 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) shareDocument
 {
-	NSURL *url = [NSURL fileURLWithPath:[NSString stringWithUTF8String:filePath]];
-	UIActivityViewController *cont = [[UIActivityViewController alloc] initWithActivityItems:[NSArray arrayWithObject:url] applicationActivities:nil];
+	NSURL *url = [NSURL fileURLWithPath:_filePath];
+	UIActivityViewController *cont = [[UIActivityViewController alloc] initWithActivityItems:@[url] applicationActivities:nil];
 	cont.popoverPresentationController.barButtonItem = shareButton;
 	[self presentViewController:cont animated:YES completion:nil];
 	[cont release];
@@ -599,17 +617,17 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) textSelectModeOn
 {
-	[[self navigationItem] setRightBarButtonItems:[NSArray arrayWithObject:tickButton]];
-	for (UIView<MuPageView> *view in [canvas subviews])
+	self.navigationItem.rightBarButtonItems = @[tickButton];
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
-		if ([view number] == current)
+		if (view.number == current)
 			[view textSelectModeOn];
 	}
 }
 
 - (void) textSelectModeOff
 {
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
 		[view textSelectModeOff];
 	}
@@ -617,23 +635,23 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) inkModeOn
 {
-	[[self navigationItem] setRightBarButtonItems:[NSArray arrayWithObject:tickButton]];
-	for (UIView<MuPageView> *view in [canvas subviews])
+	self.navigationItem.rightBarButtonItems = @[tickButton];
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
-		if ([view number] == current)
+		if (view.number == current)
 			[view inkModeOn];
 	}
 }
 
 - (void) deleteModeOn
 {
-	[[self navigationItem] setRightBarButtonItems:[NSArray arrayWithObject:deleteButton]];
+	self.navigationItem.rightBarButtonItems = @[deleteButton];
 	barmode = BARMODE_DELETE;
 }
 
 - (void) inkModeOff
 {
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
 		[view inkModeOff];
 	}
@@ -665,10 +683,9 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) onShowSearch: (id)sender
 {
-	[[self navigationItem] setRightBarButtonItems:
-		[NSArray arrayWithObjects: nextButton, prevButton, nil]];
-	[[self navigationItem] setLeftBarButtonItem: cancelButton];
-	[[self navigationItem] setTitleView: searchBar];
+	self.navigationItem.rightBarButtonItems =	@[nextButton, prevButton];
+	self.navigationItem.leftBarButtonItem = cancelButton;
+	self.navigationItem.titleView = searchBar;
 	[searchBar becomeFirstResponder];
 	barmode = BARMODE_SEARCH;
 }
@@ -676,22 +693,22 @@ static void saveDoc(char *current_path, fz_document *doc)
 - (void) onTick: (id)sender
 {
 
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
-		if ([view number] == current)
+		if (view.number == current)
 		{
 			switch (barmode)
 			{
 				case BARMODE_HIGHLIGHT:
-					[view saveSelectionAsMarkup:FZ_ANNOT_HIGHLIGHT];
+					[view saveSelectionAsMarkup:PDF_ANNOT_HIGHLIGHT];
 					break;
 
 				case BARMODE_UNDERLINE:
-					[view saveSelectionAsMarkup:FZ_ANNOT_UNDERLINE];
+					[view saveSelectionAsMarkup:PDF_ANNOT_UNDERLINE];
 					break;
 
 				case BARMODE_STRIKE:
-					[view saveSelectionAsMarkup:FZ_ANNOT_STRIKEOUT];
+					[view saveSelectionAsMarkup:PDF_ANNOT_STRIKE_OUT];
 					break;
 
 				case BARMODE_INK:
@@ -705,9 +722,9 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) onDelete: (id)sender
 {
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
-		if ([view number] == current)
+		if (view.number == current)
 			[view deleteSelectedAnnotation];
 	}
 	[self showAnnotationMenu];
@@ -724,7 +741,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 			/* fallthrough */
 		case BARMODE_ANNOTATION:
 		case BARMODE_MORE:
-			[[self navigationItem] setTitleView: nil];
+			[self.navigationItem setTitleView: nil];
 			[self addMainMenuButtons];
 			barmode = BARMODE_MAIN;
 			break;
@@ -757,7 +774,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 	}
 	else
 	{
-		[[self navigationController] popViewControllerAnimated:YES];
+		[self.navigationController popViewControllerAnimated:YES];
 	}
 }
 
@@ -766,10 +783,10 @@ static void saveDoc(char *current_path, fz_document *doc)
 	if ([CloseAlertMessage isEqualToString:alertView.message])
 	{
 		if (buttonIndex == 1)
-			saveDoc(filePath, doc);
+			saveDoc(_filePath.UTF8String, doc);
 
 		[alertView dismissWithClickedButtonIndex:buttonIndex animated:YES];
-		[[self navigationController] popViewControllerAnimated:YES];
+		[self.navigationController popViewControllerAnimated:YES];
 	}
 
 	if ([ShareAlertMessage isEqualToString:alertView.message])
@@ -777,7 +794,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 		[alertView dismissWithClickedButtonIndex:buttonIndex animated:NO];
 		if (buttonIndex == 1)
 		{
-			saveDoc(filePath, doc);
+			saveDoc(_filePath.UTF8String, doc);
 			[self shareDocument];
 		}
 	}
@@ -786,7 +803,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 - (void) resetSearch
 {
 	searchPage = -1;
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 		[view clearSearchResults];
 }
 
@@ -795,8 +812,8 @@ static void saveDoc(char *current_path, fz_document *doc)
 	printf("search found match on page %d\n", number);
 	searchPage = number;
 	[self gotoPage: number animated: NO];
-	for (UIView<MuPageView> *view in [canvas subviews])
-		if ([view number] == number)
+	for (UIView<MuPageView> *view in canvas.subviews)
+		if (view.number == number)
 			[view showSearchResults: count];
 		else
 			[view clearSearchResults];
@@ -815,10 +832,10 @@ static void saveDoc(char *current_path, fz_document *doc)
 	else
 		start = current;
 
-	needle = strdup([[searchBar text] UTF8String]);
+	needle = strdup(searchBar.text.UTF8String);
 
 	searchField = nil;
-	for (id view in [searchBar subviews])
+	for (id view in searchBar.subviews)
 		if ([view isKindOfClass: [UITextField class]])
 			searchField = view;
 
@@ -858,7 +875,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 			[searchField setEnabled: YES];
 			UIAlertView *alert = [[UIAlertView alloc]
 				initWithTitle: @"No matches found for:"
-				message: [NSString stringWithUTF8String: needle]
+				message: @(needle)
 				delegate: nil
 				cancelButtonTitle: @"Close"
 				otherButtonTitles: nil];
@@ -887,7 +904,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 - (void) searchBar: (UISearchBar*)sender textDidChange: (NSString*)searchText
 {
 	[self resetSearch];
-	if ([[searchBar text] length] > 0) {
+	if (searchBar.text.length > 0) {
 		[prevButton setEnabled: YES];
 		[nextButton setEnabled: YES];
 	} else {
@@ -898,9 +915,9 @@ static void saveDoc(char *current_path, fz_document *doc)
 
 - (void) onSlide: (id)sender
 {
-	int number = [slider value];
-	if ([slider isTracking])
-		[indicator setText: [NSString stringWithFormat: @" %d of %d ", number+1, fz_count_pages(ctx, doc)]];
+	int number = slider.value;
+	if (slider.tracking)
+		indicator.text = [NSString stringWithFormat: @" %d of %d ", number+1, fz_count_pages(ctx, doc)];
 	else
 		[self gotoPage: number animated: NO];
 }
@@ -915,13 +932,13 @@ static void saveDoc(char *current_path, fz_document *doc)
 - (void) onTap: (UITapGestureRecognizer*)sender
 {
 	CGPoint p = [sender locationInView: canvas];
-	CGPoint ofs = [canvas contentOffset];
+	CGPoint ofs = canvas.contentOffset;
 	float x0 = (width - GAP) / 5;
 	float x1 = (width - GAP) - x0;
 	p.x -= ofs.x;
 	p.y -= ofs.y;
 	__block BOOL tapHandled = NO;
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
 		CGPoint pp = [sender locationInView:view];
 		if (CGRectContainsPoint(view.bounds, pp))
@@ -976,7 +993,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 	} else if (p.x > x1) {
 		[self gotoPage: current+1 animated: YES];
 	} else {
-		if ([[self navigationController] isNavigationBarHidden])
+		if (self.navigationController.navigationBarHidden)
 			[self showNavigationBar];
 		else if (barmode == BARMODE_MAIN)
 			[self hideNavigationBar];
@@ -997,7 +1014,7 @@ static void saveDoc(char *current_path, fz_document *doc)
 	if (sender.state == UIGestureRecognizerStateEnded)
 		scale = sender.scale;
 
-	for (UIView<MuPageView> *view in [canvas subviews])
+	for (UIView<MuPageView> *view in canvas.subviews)
 	{
 		// Zoom only the visible page until end of gesture
 		if (view.number == current || sender.state == UIGestureRecognizerStateEnded)
@@ -1029,21 +1046,21 @@ static void saveDoc(char *current_path, fz_document *doc)
 	if (scroll_animating)
 		return; // don't mess with layout during animations
 
-	float x = [canvas contentOffset].x + width * 0.5f;
+	float x = canvas.contentOffset.x + width * 0.5f;
 	current = x / width;
 
 	[[NSUserDefaults standardUserDefaults] setInteger: current forKey: key];
 
-	[indicator setText: [NSString stringWithFormat: @" %d of %d ", current+1, fz_count_pages(ctx, doc)]];
-	[slider setValue: current];
+	indicator.text = [NSString stringWithFormat: @" %d of %d ", current+1, fz_count_pages(ctx, doc)];
+	slider.value = current;
 
 	// swap the distant page views out
 
 	NSMutableSet *invisiblePages = [[NSMutableSet alloc] init];
-	for (UIView<MuPageView> *view in [canvas subviews]) {
-		if ([view number] != current)
+	for (UIView<MuPageView> *view in canvas.subviews) {
+		if (view.number != current)
 			[view resetZoomAnimated: YES];
-		if ([view number] < current - 2 || [view number] > current + 2)
+		if (view.number < current - 2 || view.number > current + 2)
 			[invisiblePages addObject: view];
 	}
 	for (UIView<MuPageView> *view in invisiblePages)
@@ -1064,8 +1081,8 @@ static void saveDoc(char *current_path, fz_document *doc)
 	if (number < 0 || number >= fz_count_pages(ctx, doc))
 		return;
 	int found = 0;
-	for (UIView<MuPageView> *view in [canvas subviews])
-		if ([view number] == number)
+	for (UIView<MuPageView> *view in canvas.subviews)
+		if (view.number == number)
 			found = 1;
 	if (!found) {
 		UIView<MuPageView> *view
@@ -1103,18 +1120,18 @@ static void saveDoc(char *current_path, fz_document *doc)
 		[UIView setAnimationDelegate: self];
 		[UIView setAnimationDidStopSelector: @selector(onGotoPageFinished)];
 
-		for (UIView<MuPageView> *view in [canvas subviews])
+		for (UIView<MuPageView> *view in canvas.subviews)
 			[view resetZoomAnimated: NO];
 
-		[canvas setContentOffset: CGPointMake(number * width, 0)];
-		[slider setValue: number];
-		[indicator setText: [NSString stringWithFormat: @" %d of %d ", number+1, fz_count_pages(ctx, doc)]];
+		canvas.contentOffset = CGPointMake(number * width, 0);
+		slider.value = number;
+		indicator.text = [NSString stringWithFormat: @" %d of %d ", number+1, fz_count_pages(ctx, doc)];
 
 		[UIView commitAnimations];
 	} else {
-		for (UIView<MuPageView> *view in [canvas subviews])
+		for (UIView<MuPageView> *view in canvas.subviews)
 			[view resetZoomAnimated: NO];
-		[canvas setContentOffset: CGPointMake(number * width, 0)];
+		canvas.contentOffset = CGPointMake(number * width, 0);
 	}
 	current = number;
 }
@@ -1158,8 +1175,8 @@ static void saveDoc(char *current_path, fz_document *doc)
 	// We need to set these here, because during the animation we may use a wider
 	// size (the maximum of the landscape/portrait widths), to avoid clipping during
 	// the rotation.
-	[canvas setContentSize: CGSizeMake(fz_count_pages(ctx, doc) * width, height)];
-	[canvas setContentOffset: CGPointMake(current * width, 0)];
+	canvas.contentSize = CGSizeMake(fz_count_pages(ctx, doc) * width, height);
+	canvas.contentOffset = CGPointMake(current * width, 0);
 }
 
 @end
