@@ -3,5 +3,5 @@
 @class MuDocumentController;
 
 @interface MuOutlineController : UITableViewController
-- (id) initWithTarget: (id)aTarget titles: (NSMutableArray*)aTitles pages: (NSMutableArray*)aPages;
+- (instancetype) initWithTarget: (id)aTarget titles: (NSMutableArray*)aTitles pages: (NSMutableArray*)aPages;
 @end
