@@ -3,11 +3,7 @@
 #import "MuPageView.h"
 
 @interface MuPageViewReflow : UIWebView <UIWebViewDelegate,MuPageView>
-{
-	int number;
-	float scale;
-}
 
--(id) initWithFrame:(CGRect)frame document:(MuDocRef *)aDoc page:(int)aNumber;
+-(instancetype) initWithFrame:(CGRect)frame document:(MuDocRef *)aDoc page:(int)aNumber;
 
 @end
