@@ -8,7 +8,7 @@
 
 void fz_var_imp(void *var)
 {
-	UNUSED(var); /* Do nothing */
+	/* Do nothing */
 }
 
 void fz_flush_warnings(fz_context *ctx)
@@ -16,20 +16,17 @@ void fz_flush_warnings(fz_context *ctx)
 	if (ctx->warn->count > 1)
 	{
 		fprintf(stderr, "warning: ... repeated %d times ...\n", ctx->warn->count);
-		LOGE("warning: ... repeated %d times ...\n", ctx->warn->count);
 	}
 	ctx->warn->message[0] = 0;
 	ctx->warn->count = 0;
 }
 
-void fz_warn(fz_context *ctx, const char *fmt, ...)
+void fz_vwarn(fz_context *ctx, const char *fmt, va_list ap)
 {
-	va_list ap;
 	char buf[sizeof ctx->warn->message];
 
-	va_start(ap, fmt);
-	vsnprintf(buf, sizeof buf, fmt, ap);
-	va_end(ap);
+	fz_vsnprintf(buf, sizeof buf, fmt, ap);
+	buf[sizeof(buf) - 1] = 0;
 #ifdef USE_OUTPUT_DEBUG_STRING
 	OutputDebugStringA(buf);
 	OutputDebugStringA("\n");
@@ -43,12 +40,19 @@ void fz_warn(fz_context *ctx, const char *fmt, ...)
 	{
 		fz_flush_warnings(ctx);
 		fprintf(stderr, "warning: %s\n", buf);
-		LOGE("warning: %s\n", buf);
 		fz_strlcpy(ctx->warn->message, buf, sizeof ctx->warn->message);
 		ctx->warn->count = 1;
 	}
 }
 
+void fz_warn(fz_context *ctx, const char *fmt, ...)
+{
+	va_list ap;
+	va_start(ap, fmt);
+	fz_vwarn(ctx, fmt, ap);
+	va_end(ap);
+}
+
 /* Error context */
 
 /* When we first setjmp, code is set to 0. Whenever we throw, we add 2 to
@@ -82,15 +86,14 @@ void fz_warn(fz_context *ctx, const char *fmt, ...)
 
 FZ_NORETURN static void throw(fz_context *ctx)
 {
-	if (ctx->error->top >= 0)
+	if (ctx->error->top >= ctx->error->stack)
 	{
-		ctx->error->stack[ctx->error->top].code += 2;
-		fz_longjmp(ctx->error->stack[ctx->error->top].buffer, ctx->error->stack[ctx->error->top].code);
+		ctx->error->top->code += 2;
+		fz_longjmp(ctx->error->top->buffer, 1);
 	}
 	else
 	{
 		fprintf(stderr, "uncaught exception: %s\n", ctx->error->message);
-		LOGE("uncaught exception: %s\n", ctx->error->message);
 #ifdef USE_OUTPUT_DEBUG_STRING
 		OutputDebugStringA("uncaught exception: ");
 		OutputDebugStringA(ctx->error->message);
@@ -100,15 +103,47 @@ FZ_NORETURN static void throw(fz_context *ctx)
 	}
 }
 
-void fz_push_try(fz_context *ctx)
+/* Only called when we hit the bottom of the exception stack.
+ * Do the same as fz_throw, but don't actually throw. */
+static int fz_fake_throw(fz_context *ctx, int code, const char *fmt, ...)
+{
+	va_list args;
+	ctx->error->errcode = code;
+	va_start(args, fmt);
+	fz_vsnprintf(ctx->error->message, sizeof ctx->error->message, fmt, args);
+	ctx->error->message[sizeof(ctx->error->message) - 1] = 0;
+	va_end(args);
+
+	if (code != FZ_ERROR_ABORT)
+	{
+		fz_flush_warnings(ctx);
+		fprintf(stderr, "error: %s\n", ctx->error->message);
+#ifdef USE_OUTPUT_DEBUG_STRING
+		OutputDebugStringA("error: ");
+		OutputDebugStringA(ctx->error->message);
+		OutputDebugStringA("\n");
+#endif
+	}
+
+	/* We need to arrive in the always/catch block as if throw
+	 * had taken place. */
+	ctx->error->top++;
+	ctx->error->top->code = 2;
+	return 0;
+}
+
+int fz_push_try(fz_context *ctx)
 {
 	/* If we would overflow the exception stack, throw an exception instead
-	 * of entering the try block. */
-	if (ctx->error->top + 1 >= nelem(ctx->error->stack))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "exception stack overflow!");
+	 * of entering the try block. We assume that we always have room for
+	 * 1 extra level on the stack here - i.e. we throw the error on us
+	 * starting to use the last level. */
+	if (ctx->error->top + 2 >= ctx->error->stack + nelem(ctx->error->stack))
+		return fz_fake_throw(ctx, FZ_ERROR_GENERIC, "exception stack overflow!");
 
 	ctx->error->top++;
-	ctx->error->stack[ctx->error->top].code = 0;
+	ctx->error->top->code = 0;
+	return 1;
 }
 
 int fz_caught(fz_context *ctx)
@@ -123,19 +158,16 @@ const char *fz_caught_message(fz_context *ctx)
 	return ctx->error->message;
 }
 
-void fz_throw(fz_context *ctx, int code, const char *fmt, ...)
+void fz_vthrow(fz_context *ctx, int code, const char *fmt, va_list ap)
 {
-	va_list args;
 	ctx->error->errcode = code;
-	va_start(args, fmt);
-	vsnprintf(ctx->error->message, sizeof ctx->error->message, fmt, args);
-	va_end(args);
+	fz_vsnprintf(ctx->error->message, sizeof ctx->error->message, fmt, ap);
+	ctx->error->message[sizeof(ctx->error->message) - 1] = 0;
 
 	if (code != FZ_ERROR_ABORT)
 	{
 		fz_flush_warnings(ctx);
 		fprintf(stderr, "error: %s\n", ctx->error->message);
-		LOGE("error: %s\n", ctx->error->message);
 #ifdef USE_OUTPUT_DEBUG_STRING
 		OutputDebugStringA("error: ");
 		OutputDebugStringA(ctx->error->message);
@@ -146,40 +178,93 @@ void fz_throw(fz_context *ctx, int code, const char *fmt, ...)
 	throw(ctx);
 }
 
+void fz_throw(fz_context *ctx, int code, const char *fmt, ...)
+{
+	va_list ap;
+	va_start(ap, fmt);
+	fz_vthrow(ctx, code, fmt, ap);
+	va_end(ap);
+}
+
 void fz_rethrow(fz_context *ctx)
 {
 	assert(ctx && ctx->error && ctx->error->errcode >= FZ_ERROR_NONE);
 	throw(ctx);
 }
 
-void fz_rethrow_message(fz_context *ctx, const char *fmt, ...)
-{
-	va_list args;
-
-	assert(ctx && ctx->error && ctx->error->errcode >= FZ_ERROR_NONE);
-
-	va_start(args, fmt);
-	vsnprintf(ctx->error->message, sizeof ctx->error->message, fmt, args);
-	va_end(args);
-
-	if (ctx->error->errcode != FZ_ERROR_ABORT)
-	{
-		fz_flush_warnings(ctx);
-		fprintf(stderr, "error: %s\n", ctx->error->message);
-		LOGE("error: %s\n", ctx->error->message);
-#ifdef USE_OUTPUT_DEBUG_STRING
-		OutputDebugStringA("error: ");
-		OutputDebugStringA(ctx->error->message);
-		OutputDebugStringA("\n");
-#endif
-	}
-
-	throw(ctx);
-}
-
 void fz_rethrow_if(fz_context *ctx, int err)
 {
 	assert(ctx && ctx->error && ctx->error->errcode >= FZ_ERROR_NONE);
 	if (ctx->error->errcode == err)
 		fz_rethrow(ctx);
 }
+
+/* Android specific code to take fprintf to LOG */
+
+#ifdef __ANDROID__
+#include <android/log.h>
+
+#define LOG_TAG "libmupdf"
+
+static char android_log_buffer[4096];
+static int android_log_fill = 0;
+
+static char android_log_buffer2[4096];
+
+int fz_android_fprintf(FILE *file, const char *fmt, ...)
+{
+	va_list args;
+	char *p, *q;
+
+	/* Just in case someone has some magic fprintf redirection code working */
+	va_start(args, fmt);
+	vfprintf(file, fmt, args);
+	va_end(args);
+
+	if (file != stdout && file != stderr)
+		return 0;
+
+	va_start(args, fmt);
+	vsnprintf(android_log_buffer2, sizeof(android_log_buffer2)-1, fmt, args);
+	va_end(args);
+
+	/* Ensure we are always null terminated */
+	android_log_buffer2[sizeof(android_log_buffer2)-1] = 0;
+
+	p = android_log_buffer2;
+	q = p;
+	do
+	{
+		/* Find the end of the string, or the next \n */
+		while (*p && *p != '\n')
+			p++;
+
+		/* We need to output from q to p. Limit ourselves to what
+		 * will fit in the existing buffer. */
+		if (p - q >= sizeof(android_log_buffer)-1 - android_log_fill)
+			p = q + sizeof(android_log_buffer)-1 - android_log_fill;
+
+		memcpy(&android_log_buffer[android_log_fill], q, p-q);
+		android_log_fill += p-q;
+		if (*p == '\n')
+		{
+			android_log_buffer[android_log_fill] = 0;
+			__android_log_print(ANDROID_LOG_ERROR, LOG_TAG, "%s", android_log_buffer);
+			usleep(1); /* Hack to avoid the logcat buffer losing data */
+			android_log_fill = 0;
+			p++; /* Skip over the \n */
+		}
+		else if (android_log_fill >= sizeof(android_log_buffer)-1)
+		{
+			android_log_buffer[sizeof(android_log_buffer2)-1] = 0;
+			__android_log_print(ANDROID_LOG_ERROR, LOG_TAG, "%s", android_log_buffer);
+			usleep(1); /* Hack to avoid the logcat buffer losing data */
+			android_log_fill = 0;
+		}
+		q = p;
+	}
+	while (*p);
+
+	return 0;
+}
+#endif
