@@ -11,18 +11,18 @@ struct fz_flate_s
 	unsigned char buffer[4096];
 };
 
-static void *zalloc(void *opaque, unsigned int items, unsigned int size)
+static void *zalloc_flate(void *opaque, unsigned int items, unsigned int size)
 {
 	return fz_malloc_array_no_throw(opaque, items, size);
 }
 
-static void zfree(void *opaque, void *ptr)
+static void zfree_flate(void *opaque, void *ptr)
 {
 	fz_free(opaque, ptr);
 }
 
 static int
-next_flated(fz_context *ctx, fz_stream *stm, int required)
+next_flated(fz_context *ctx, fz_stream *stm, size_t required)
 {
 	fz_flate *state = stm->state;
 	fz_stream *chain = state->chain;
@@ -39,7 +39,7 @@ next_flated(fz_context *ctx, fz_stream *stm, int required)
 
 	while (zp->avail_out > 0)
 	{
-		zp->avail_in = fz_available(ctx, chain, 1);
+		zp->avail_in = (uInt)fz_available(ctx, chain, 1);
 		zp->next_in = chain->rp;
 
 		code = inflate(zp, Z_SYNC_FLUSH);
@@ -111,8 +111,8 @@ fz_open_flated(fz_context *ctx, fz_stream *chain, int window_bits)
 		state = fz_malloc_struct(ctx, fz_flate);
 		state->chain = chain;
 
-		state->z.zalloc = zalloc;
-		state->z.zfree = zfree;
+		state->z.zalloc = zalloc_flate;
+		state->z.zfree = zfree_flate;
 		state->z.opaque = ctx;
 		state->z.next_in = NULL;
 		state->z.avail_in = 0;
