@@ -21,66 +21,66 @@ fz_write_pwg_page_header(fz_context *ctx, fz_output *out, const fz_pwg_options *
 	fz_write(ctx, out, pwg ? pwg->media_color : zero, 64);
 	fz_write(ctx, out, pwg ? pwg->media_type : zero, 64);
 	fz_write(ctx, out, pwg ? pwg->output_type : zero, 64);
-	fz_write_int32be(ctx, out, pwg ? pwg->advance_distance : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->advance_media : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->collate : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->cut_media : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->duplex : 0);
-	fz_write_int32be(ctx, out, xres);
-	fz_write_int32be(ctx, out, yres);
+	fz_write_int32_be(ctx, out, pwg ? pwg->advance_distance : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->advance_media : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->collate : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->cut_media : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->duplex : 0);
+	fz_write_int32_be(ctx, out, xres);
+	fz_write_int32_be(ctx, out, yres);
 	/* CUPS format says that 284->300 are supposed to be the bbox of the
 	 * page in points. PWG says 'Reserved'. */
 	for (i=284; i < 300; i += 4)
 		fz_write(ctx, out, zero, 4);
-	fz_write_int32be(ctx, out, pwg ? pwg->insert_sheet : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->jog : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->leading_edge : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->insert_sheet : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->jog : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->leading_edge : 0);
 	/* CUPS format says that 312->320 are supposed to be the margins of
 	 * the lower left hand edge of page in points. PWG says 'Reserved'. */
 	for (i=312; i < 320; i += 4)
 		fz_write(ctx, out, zero, 4);
-	fz_write_int32be(ctx, out, pwg ? pwg->manual_feed : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->media_position : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->media_weight : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->mirror_print : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->negative_print : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->num_copies : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->orientation : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->output_face_up : 0);
-	fz_write_int32be(ctx, out, w * 72/ xres);	/* Page size in points */
-	fz_write_int32be(ctx, out, h * 72/ yres);
-	fz_write_int32be(ctx, out, pwg ? pwg->separations : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->tray_switch : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->tumble : 0);
-	fz_write_int32be(ctx, out, w); /* Page image in pixels */
-	fz_write_int32be(ctx, out, h);
-	fz_write_int32be(ctx, out, pwg ? pwg->media_type_num : 0);
-	fz_write_int32be(ctx, out, bpp < 8 ? 1 : 8); /* Bits per color */
-	fz_write_int32be(ctx, out, bpp); /* Bits per pixel */
-	fz_write_int32be(ctx, out, (w * bpp + 7)/8); /* Bytes per line */
-	fz_write_int32be(ctx, out, 0); /* Chunky pixels */
+	fz_write_int32_be(ctx, out, pwg ? pwg->manual_feed : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->media_position : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->media_weight : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->mirror_print : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->negative_print : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->num_copies : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->orientation : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->output_face_up : 0);
+	fz_write_int32_be(ctx, out, w * 72/ xres);	/* Page size in points */
+	fz_write_int32_be(ctx, out, h * 72/ yres);
+	fz_write_int32_be(ctx, out, pwg ? pwg->separations : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->tray_switch : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->tumble : 0);
+	fz_write_int32_be(ctx, out, w); /* Page image in pixels */
+	fz_write_int32_be(ctx, out, h);
+	fz_write_int32_be(ctx, out, pwg ? pwg->media_type_num : 0);
+	fz_write_int32_be(ctx, out, bpp < 8 ? 1 : 8); /* Bits per color */
+	fz_write_int32_be(ctx, out, bpp); /* Bits per pixel */
+	fz_write_int32_be(ctx, out, (w * bpp + 7)/8); /* Bytes per line */
+	fz_write_int32_be(ctx, out, 0); /* Chunky pixels */
 	switch (bpp)
 	{
-	case 1: fz_write_int32be(ctx, out, 3); /* Black */ break;
-	case 8: fz_write_int32be(ctx, out, 18); /* Sgray */ break;
-	case 24: fz_write_int32be(ctx, out, 19); /* Srgb */ break;
-	case 32: fz_write_int32be(ctx, out, 6); /* Cmyk */ break;
+	case 1: fz_write_int32_be(ctx, out, 3); /* Black */ break;
+	case 8: fz_write_int32_be(ctx, out, 18); /* Sgray */ break;
+	case 24: fz_write_int32_be(ctx, out, 19); /* Srgb */ break;
+	case 32: fz_write_int32_be(ctx, out, 6); /* Cmyk */ break;
 	default: fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap bpp must be 1, 8, 24 or 32 to write as pwg");
 	}
-	fz_write_int32be(ctx, out, pwg ? pwg->compression : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->row_count : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->row_feed : 0);
-	fz_write_int32be(ctx, out, pwg ? pwg->row_step : 0);
-	fz_write_int32be(ctx, out, bpp <= 8 ? 1 : 3); /* Num Colors */
+	fz_write_int32_be(ctx, out, pwg ? pwg->compression : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->row_count : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->row_feed : 0);
+	fz_write_int32_be(ctx, out, pwg ? pwg->row_step : 0);
+	fz_write_int32_be(ctx, out, bpp <= 8 ? 1 : 3); /* Num Colors */
 	for (i=424; i < 452; i += 4)
 		fz_write(ctx, out, zero, 4);
-	fz_write_int32be(ctx, out, 1); /* TotalPageCount */
-	fz_write_int32be(ctx, out, 1); /* CrossFeedTransform */
-	fz_write_int32be(ctx, out, 1); /* FeedTransform */
-	fz_write_int32be(ctx, out, 0); /* ImageBoxLeft */
-	fz_write_int32be(ctx, out, 0); /* ImageBoxTop */
-	fz_write_int32be(ctx, out, w); /* ImageBoxRight */
-	fz_write_int32be(ctx, out, h); /* ImageBoxBottom */
+	fz_write_int32_be(ctx, out, 1); /* TotalPageCount */
+	fz_write_int32_be(ctx, out, 1); /* CrossFeedTransform */
+	fz_write_int32_be(ctx, out, 1); /* FeedTransform */
+	fz_write_int32_be(ctx, out, 0); /* ImageBoxLeft */
+	fz_write_int32_be(ctx, out, 0); /* ImageBoxTop */
+	fz_write_int32_be(ctx, out, w); /* ImageBoxRight */
+	fz_write_int32_be(ctx, out, h); /* ImageBoxBottom */
 	for (i=480; i < 1668; i += 4)
 		fz_write(ctx, out, zero, 4);
 	fz_write(ctx, out, pwg ? pwg->rendering_intent : zero, 64);
@@ -91,35 +91,36 @@ void
 fz_write_pixmap_as_pwg_page(fz_context *ctx, fz_output *out, const fz_pixmap *pixmap, const fz_pwg_options *pwg)
 {
 	unsigned char *sp;
-	int y, x, sn, dn, ss;
+	int y, x, sn, dn, ss, ss2;
 
 	if (!out || !pixmap)
 		return;
 
-	if (pixmap->n != 1 && pixmap->n != 2 && pixmap->n != 4 && pixmap->n != 5)
+	if (pixmap->alpha != 0)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot write pwg with alpha");
+	sn = pixmap->n;
+	if (sn != 1 && sn != 3 && sn != 4)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap must be grayscale, rgb or cmyk to write as pwg");
 
-	sn = pixmap->n;
 	dn = pixmap->n;
-	if (dn > 1)
-		dn--;
 
 	fz_write_pwg_page_header(ctx, out, pwg, pixmap->xres, pixmap->yres, pixmap->w, pixmap->h, dn*8);
 
 	/* Now output the actual bitmap, using a packbits like compression */
 	sp = pixmap->samples;
 	ss = pixmap->w * sn;
+	ss2 = pixmap->stride;
 	y = 0;
 	while (y < pixmap->h)
 	{
 		int yrep;
 
-		assert(sp == pixmap->samples + y * ss);
+		assert(sp == pixmap->samples + y * ss2);
 
 		/* Count the number of times this line is repeated */
 		for (yrep = 1; yrep < 256 && y+yrep < pixmap->h; yrep++)
 		{
-			if (memcmp(sp, sp + yrep * ss, ss) != 0)
+			if (memcmp(sp, sp + yrep * ss2, ss) != 0)
 				break;
 		}
 		fz_write_byte(ctx, out, yrep-1);
@@ -130,7 +131,7 @@ fz_write_pixmap_as_pwg_page(fz_context *ctx, fz_output *out, const fz_pixmap *pi
 		{
 			int d;
 
-			assert(sp == pixmap->samples + y * ss + x * sn);
+			assert(sp == pixmap->samples + y * ss2 + x * sn);
 
 			/* How far do we have to look to find a repeated value? */
 			for (d = 1; d < 128 && x+d < pixmap->w; d++)
@@ -169,7 +170,7 @@ fz_write_pixmap_as_pwg_page(fz_context *ctx, fz_output *out, const fz_pixmap *pi
 		}
 
 		/* Move to the next line */
-		sp += ss*(yrep-1);
+		sp += ss2*(yrep-1);
 		y += yrep;
 	}
 }
