@@ -3,7 +3,7 @@
 fz_pixmap *
 fz_keep_pixmap(fz_context *ctx, fz_pixmap *pix)
 {
-	return (fz_pixmap *)fz_keep_storable(ctx, &pix->storable);
+	return fz_keep_storable(ctx, &pix->storable);
 }
 
 void
@@ -13,41 +13,52 @@ fz_drop_pixmap(fz_context *ctx, fz_pixmap *pix)
 }
 
 void
-fz_free_pixmap_imp(fz_context *ctx, fz_storable *pix_)
+fz_drop_pixmap_imp(fz_context *ctx, fz_storable *pix_)
 {
 	fz_pixmap *pix = (fz_pixmap *)pix_;
 
-	if (pix->colorspace)
-		fz_drop_colorspace(ctx, pix->colorspace);
+	fz_drop_colorspace(ctx, pix->colorspace);
 	if (pix->free_samples)
 		fz_free(ctx, pix->samples);
 	fz_free(ctx, pix);
 }
 
 fz_pixmap *
-fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h, unsigned char *samples)
+fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h, int alpha, int stride, unsigned char *samples)
 {
 	fz_pixmap *pix;
+	int n;
 
 	if (w < 0 || h < 0)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal dimensions for pixmap %d %d", w, h);
 
+	n = alpha + fz_colorspace_n(ctx, colorspace);
+	if (stride < n*w && stride > -n*w)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal stride for pixmap (n=%d w=%d, stride=%d)", n, w, stride);
+	if (samples == NULL && stride < n*w)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal -ve stride for pixmap without data");
+
 	pix = fz_malloc_struct(ctx, fz_pixmap);
-	FZ_INIT_STORABLE(pix, 1, fz_free_pixmap_imp);
+	FZ_INIT_STORABLE(pix, 1, fz_drop_pixmap_imp);
 	pix->x = 0;
 	pix->y = 0;
 	pix->w = w;
 	pix->h = h;
+	pix->alpha = alpha = !!alpha;
 	pix->interpolate = 1;
 	pix->xres = 96;
 	pix->yres = 96;
 	pix->colorspace = NULL;
-	pix->n = 1;
+	pix->n = n;
+	pix->stride = stride;
 
 	if (colorspace)
 	{
 		pix->colorspace = fz_keep_colorspace(ctx, colorspace);
-		pix->n = 1 + colorspace->n;
+	}
+	else
+	{
+		assert(alpha);
 	}
 
 	pix->samples = samples;
@@ -59,14 +70,13 @@ fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h
 	{
 		fz_try(ctx)
 		{
-			if (pix->w + pix->n - 1 > INT_MAX / pix->n)
+			if (pix->stride - 1 > INT_MAX / pix->n)
 				fz_throw(ctx, FZ_ERROR_GENERIC, "overly wide image");
-			pix->samples = fz_malloc_array(ctx, pix->h, pix->w * pix->n);
+			pix->samples = fz_malloc_array(ctx, pix->h, pix->stride);
 		}
 		fz_catch(ctx)
 		{
-			if (colorspace)
-				fz_drop_colorspace(ctx, colorspace);
+			fz_drop_colorspace(ctx, colorspace);
 			fz_free(ctx, pix);
 			fz_rethrow(ctx);
 		}
@@ -77,32 +87,35 @@ fz_new_pixmap_with_data(fz_context *ctx, fz_colorspace *colorspace, int w, int h
 }
 
 fz_pixmap *
-fz_new_pixmap(fz_context *ctx, fz_colorspace *colorspace, int w, int h)
+fz_new_pixmap(fz_context *ctx, fz_colorspace *colorspace, int w, int h, int alpha)
 {
-	return fz_new_pixmap_with_data(ctx, colorspace, w, h, NULL);
+	int stride = (fz_colorspace_n(ctx, colorspace) + alpha) * w;
+	return fz_new_pixmap_with_data(ctx, colorspace, w, h, alpha, stride, NULL);
 }
 
 fz_pixmap *
-fz_new_pixmap_with_bbox(fz_context *ctx, fz_colorspace *colorspace, const fz_irect *r)
+fz_new_pixmap_with_bbox(fz_context *ctx, fz_colorspace *colorspace, const fz_irect *r, int alpha)
 {
 	fz_pixmap *pixmap;
-	pixmap = fz_new_pixmap(ctx, colorspace, r->x1 - r->x0, r->y1 - r->y0);
+	pixmap = fz_new_pixmap(ctx, colorspace, r->x1 - r->x0, r->y1 - r->y0, alpha);
 	pixmap->x = r->x0;
 	pixmap->y = r->y0;
 	return pixmap;
 }
 
 fz_pixmap *
-fz_new_pixmap_with_bbox_and_data(fz_context *ctx, fz_colorspace *colorspace, const fz_irect *r, unsigned char *samples)
+fz_new_pixmap_with_bbox_and_data(fz_context *ctx, fz_colorspace *colorspace, const fz_irect *r, int alpha, unsigned char *samples)
 {
-	fz_pixmap *pixmap = fz_new_pixmap_with_data(ctx, colorspace, r->x1 - r->x0, r->y1 - r->y0, samples);
+	int w = r->x1 - r->x0;
+	int stride = (fz_colorspace_n(ctx, colorspace) + alpha) * w;
+	fz_pixmap *pixmap = fz_new_pixmap_with_data(ctx, colorspace, w, r->y1 - r->y0, alpha, stride, samples);
 	pixmap->x = r->x0;
 	pixmap->y = r->y0;
 	return pixmap;
 }
 
 fz_irect *
-fz_pixmap_bbox(fz_context *ctx, fz_pixmap *pix, fz_irect *bbox)
+fz_pixmap_bbox(fz_context *ctx, const fz_pixmap *pix, fz_irect *bbox)
 {
 	bbox->x0 = pix->x;
 	bbox->y0 = pix->y;
@@ -112,7 +125,7 @@ fz_pixmap_bbox(fz_context *ctx, fz_pixmap *pix, fz_irect *bbox)
 }
 
 fz_irect *
-fz_pixmap_bbox_no_ctx(fz_pixmap *pix, fz_irect *bbox)
+fz_pixmap_bbox_no_ctx(const fz_pixmap *pix, fz_irect *bbox)
 {
 	bbox->x0 = pix->x;
 	bbox->y0 = pix->y;
@@ -121,6 +134,26 @@ fz_pixmap_bbox_no_ctx(fz_pixmap *pix, fz_irect *bbox)
 	return bbox;
 }
 
+fz_colorspace *
+fz_pixmap_colorspace(fz_context *ctx, fz_pixmap *pix)
+{
+	if (!pix)
+		return NULL;
+	return pix->colorspace;
+}
+
+int
+fz_pixmap_x(fz_context *ctx, fz_pixmap *pix)
+{
+	return pix->x;
+}
+
+int
+fz_pixmap_y(fz_context *ctx, fz_pixmap *pix)
+{
+	return pix->y;
+}
+
 int
 fz_pixmap_width(fz_context *ctx, fz_pixmap *pix)
 {
@@ -133,52 +166,317 @@ fz_pixmap_height(fz_context *ctx, fz_pixmap *pix)
 	return pix->h;
 }
 
+int
+fz_pixmap_components(fz_context *ctx, fz_pixmap *pix)
+{
+	return pix->n;
+}
+
+int
+fz_pixmap_colorants(fz_context *ctx, fz_pixmap *pix)
+{
+	return pix->n - pix->alpha;
+}
+
+int
+fz_pixmap_stride(fz_context *ctx, fz_pixmap *pix)
+{
+	return pix->stride;
+}
+
+unsigned char *
+fz_pixmap_samples(fz_context *ctx, fz_pixmap *pix)
+{
+	if (!pix)
+		return NULL;
+	return pix->samples;
+}
+
+/*
+	The slowest routine in most CMYK rendering profiles.
+	We therefore spend some effort to improve it. Rather than
+	writing bytes, we write uint32_t's.
+*/
+#ifdef ARCH_ARM
+static void
+clear_cmyka_bitmap_ARM(uint32_t *samples, int c, int value)
+__attribute__((naked));
+
+static void
+clear_cmyka_bitmap_ARM(uint32_t *samples, int c, int value)
+{
+	asm volatile(
+	ENTER_ARM
+	"stmfd	r13!,{r4-r6,r14}					\n"
+	"@ r0 = samples							\n"
+	"@ r1 = c							\n"
+	"@ r2 = value							\n"
+	"mov	r3, #255						\n"
+	"mov	r12,#0			@ r12= 0			\n"
+	"subs	r1, r1, #3						\n"
+	"ble	2f							\n"
+	"str	r12,[r13,#-20]!						\n"
+	"str	r12,[r13,#4]						\n"
+	"str	r12,[r13,#8]						\n"
+	"str	r12,[r13,#12]						\n"
+	"str	r12,[r13,#16]						\n"
+	"strb	r2, [r13,#3]						\n"
+	"strb	r3, [r13,#4]						\n"
+	"strb	r2, [r13,#8]						\n"
+	"strb	r3, [r13,#9]						\n"
+	"strb	r2, [r13,#13]						\n"
+	"strb	r3, [r13,#14]						\n"
+	"strb	r2, [r13,#18]						\n"
+	"strb	r3, [r13,#19]						\n"
+	"ldmfd	r13!,{r4,r5,r6,r12,r14}					\n"
+	"1:								\n"
+	"stmia	r0!,{r4,r5,r6,r12,r14}					\n"
+	"subs	r1, r1, #4						\n"
+	"bgt	1b							\n"
+	"2:								\n"
+	"adds	r1, r1, #3						\n"
+	"ble	4f							\n"
+	"3:								\n"
+	"strb	r12,[r0], #1						\n"
+	"strb	r12,[r0], #1						\n"
+	"strb	r12,[r0], #1						\n"
+	"strb	r2, [r0], #1						\n"
+	"strb	r3, [r0], #1						\n"
+	"subs	r1, r1, #1						\n"
+	"bgt	3b							\n"
+	"4:								\n"
+	"ldmfd	r13!,{r4-r6,PC}						\n"
+	ENTER_THUMB
+	);
+}
+#endif
+
+static void
+clear_cmyk_bitmap(unsigned char *samples, int w, int h, int stride, int value, int alpha)
+{
+	uint32_t *s = (uint32_t *)(void *)samples;
+	uint8_t *t;
+
+	if (w < 0 || h < 0)
+		return;
+
+	if (alpha)
+	{
+		int c = w;
+		stride -= w*5;
+		if (stride == 0)
+		{
+#ifdef ARCH_ARM
+			clear_cmyka_bitmap_ARM(s, c, alpha);
+			return;
+#else
+			/* We can do it all fast (except for maybe a few stragglers) */
+			union
+			{
+				uint8_t bytes[20];
+				uint32_t words[5];
+			} d;
+
+			c *= h;
+			h = 1;
+
+			d.words[0] = 0;
+			d.words[1] = 0;
+			d.words[2] = 0;
+			d.words[3] = 0;
+			d.words[4] = 0;
+			d.bytes[3] = value;
+			d.bytes[4] = 255;
+			d.bytes[8] = value;
+			d.bytes[9] = 255;
+			d.bytes[13] = value;
+			d.bytes[14] = 255;
+			d.bytes[18] = value;
+			d.bytes[19] = 255;
+
+			c -= 3;
+			{
+				const uint32_t a0 = d.words[0];
+				const uint32_t a1 = d.words[1];
+				const uint32_t a2 = d.words[2];
+				const uint32_t a3 = d.words[3];
+				const uint32_t a4 = d.words[4];
+				while (c > 0)
+				{
+					*s++ = a0;
+					*s++ = a1;
+					*s++ = a2;
+					*s++ = a3;
+					*s++ = a4;
+					c -= 4;
+				}
+			}
+			c += 3;
+#endif
+		}
+		t = (unsigned char *)s;
+		w = c;
+		while (h--)
+		{
+			c = w;
+			while (c > 0)
+			{
+				*t++ = 0;
+				*t++ = 0;
+				*t++ = 0;
+				*t++ = value;
+				*t++ = 255;
+				c--;
+			}
+			t += stride;
+		}
+	}
+	else
+	{
+		stride -= w*4;
+		if ((stride & 3) == 0)
+		{
+			size_t W = w;
+			if (stride == 0)
+			{
+				W *= h;
+				h = 1;
+			}
+			W *= 4;
+			if (value == 0)
+			{
+				while (h--)
+				{
+					memset(s, 0, W);
+					s += (stride>>2);
+				}
+			}
+			else
+			{
+				/* We can do it all fast */
+				union
+				{
+					uint8_t bytes[4];
+					uint32_t word;
+				} d;
+
+				d.word = 0;
+				d.bytes[3] = value;
+				{
+					const uint32_t a0 = d.word;
+					while (h--)
+					{
+						size_t WW = W >> 2;
+						while (WW--)
+						{
+							*s++ = a0;
+						}
+						s += (stride>>2);
+					}
+				}
+			}
+		}
+		else
+		{
+			t = (unsigned char *)s;
+			while (h--)
+			{
+				int c = w;
+				while (c > 0)
+				{
+					*t++ = 0;
+					*t++ = 0;
+					*t++ = 0;
+					*t++ = value;
+					c--;
+				}
+				t += stride;
+			}
+		}
+	}
+}
+
 void
 fz_clear_pixmap(fz_context *ctx, fz_pixmap *pix)
 {
-	memset(pix->samples, 0, (unsigned int)(pix->w * pix->h * pix->n));
+	int stride = pix->w * pix->n;
+	int h = pix->h;
+	unsigned char *s = pix->samples;
+	if (stride == pix->stride)
+	{
+		stride *= h;
+		h = 1;
+	}
+	if (pix->alpha)
+	{
+		while (h--)
+		{
+			memset(s, 0, (unsigned int)stride);
+			s += pix->stride;
+		}
+	}
+	else
+	{
+		/* FIXME: Not right for CMYK or other subtractive spaces */
+		while (h--)
+		{
+			memset(s, 0xff, (unsigned int)stride);
+			s += pix->stride;
+		}
+	}
 }
 
 void
 fz_clear_pixmap_with_value(fz_context *ctx, fz_pixmap *pix, int value)
 {
-	/* CMYK needs special handling (and potentially any other subtractive colorspaces) */
-	if (pix->colorspace && pix->colorspace->n == 4)
-	{
-		int x, y;
-		unsigned char *s = pix->samples;
+	unsigned char *s;
+	int w, h, n, stride, len;
+	int alpha = pix->alpha;
 
-		value = 255 - value;
-		for (y = 0; y < pix->h; y++)
-		{
-			for (x = 0; x < pix->w; x++)
-			{
-				*s++ = 0;
-				*s++ = 0;
-				*s++ = 0;
-				*s++ = value;
-				*s++ = 255;
-			}
-		}
+	w = pix->w;
+	h = pix->h;
+	if (w < 0 || h < 0)
+		return;
+
+	/* CMYK needs special handling (and potentially any other subtractive colorspaces) */
+	if (fz_colorspace_n(ctx, pix->colorspace) == 4)
+	{
+		clear_cmyk_bitmap(pix->samples, w, h, pix->stride, 255-value, pix->alpha);
 		return;
 	}
 
-	if (value == 255)
+	n = pix->n;
+	stride = pix->stride;
+	len = w * n;
+
+	s = pix->samples;
+	if (value == 255 || !alpha)
 	{
-		memset(pix->samples, 255, (unsigned int)(pix->w * pix->h * pix->n));
+		if (stride == len)
+		{
+			len *= h;
+			h = 1;
+		}
+		while (h--)
+		{
+			memset(s, value, (unsigned int)len);
+			s += stride;
+		}
 	}
 	else
 	{
 		int k, x, y;
-		unsigned char *s = pix->samples;
+		stride -= len;
 		for (y = 0; y < pix->h; y++)
 		{
 			for (x = 0; x < pix->w; x++)
 			{
 				for (k = 0; k < pix->n - 1; k++)
 					*s++ = value;
-				*s++ = 255;
+				if (alpha)
+					*s++ = 255;
 			}
+			s += stride;
 		}
 	}
 }
@@ -199,9 +497,9 @@ fz_copy_pixmap_rect(fz_context *ctx, fz_pixmap *dest, fz_pixmap *src, const fz_i
 	if (w <= 0 || y <= 0)
 		return;
 
-	srcspan = src->w * src->n;
+	srcspan = src->stride;
 	srcp = src->samples + (unsigned int)(srcspan * (local_b.y0 - src->y) + src->n * (local_b.x0 - src->x));
-	destspan = dest->w * dest->n;
+	destspan = dest->stride;
 	destp = dest->samples + (unsigned int)(destspan * (local_b.y0 - dest->y) + dest->n * (local_b.x0 - dest->x));
 
 	if (src->n == dest->n)
@@ -236,6 +534,10 @@ fz_copy_pixmap_rect(fz_context *ctx, fz_pixmap *dest, fz_pixmap *src, const fz_i
 		}
 		while (--y);
 	}
+	else if (src->n == 1 + src->alpha && dest->n == 3 + dest->alpha)
+	{
+		assert("FIXME" == NULL);
+	}
 	else if (src->n == 4 && dest->n == 2)
 	{
 		/* Copy, and convert from rgb+alpha to grey+alpha */
@@ -257,6 +559,10 @@ fz_copy_pixmap_rect(fz_context *ctx, fz_pixmap *dest, fz_pixmap *src, const fz_i
 		}
 		while (--y);
 	}
+	else if (src->n == 3 + src->alpha && dest->n == 1 + dest->alpha)
+	{
+		assert("FIXME" == NULL);
+	}
 	else
 	{
 		/* FIXME: Crap conversion */
@@ -299,11 +605,11 @@ fz_clear_pixmap_rect_with_value(fz_context *ctx, fz_pixmap *dest, int value, con
 	if (w <= 0 || y <= 0)
 		return;
 
-	destspan = dest->w * dest->n;
+	destspan = dest->stride;
 	destp = dest->samples + (unsigned int)(destspan * (local_b.y0 - dest->y) + dest->n * (local_b.x0 - dest->x));
 
 	/* CMYK needs special handling (and potentially any other subtractive colorspaces) */
-	if (dest->colorspace && dest->colorspace->n == 4)
+	if (fz_colorspace_n(ctx, dest->colorspace) == 4)
 	{
 		value = 255 - value;
 		do
@@ -355,6 +661,10 @@ fz_premultiply_pixmap(fz_context *ctx, fz_pixmap *pix)
 	unsigned char *s = pix->samples;
 	unsigned char a;
 	int k, x, y;
+	int stride = pix->stride - pix->w * pix->n;
+
+	if (!pix->alpha)
+		return;
 
 	for (y = 0; y < pix->h; y++)
 	{
@@ -365,6 +675,7 @@ fz_premultiply_pixmap(fz_context *ctx, fz_pixmap *pix)
 				s[k] = fz_mul255(s[k], a);
 			s += pix->n;
 		}
+		s += stride;
 	}
 }
 
@@ -374,6 +685,10 @@ fz_unmultiply_pixmap(fz_context *ctx, fz_pixmap *pix)
 	unsigned char *s = pix->samples;
 	int a, inva;
 	int k, x, y;
+	int stride = pix->stride - pix->w * pix->n;
+
+	if (!pix->alpha)
+		return;
 
 	for (y = 0; y < pix->h; y++)
 	{
@@ -385,65 +700,122 @@ fz_unmultiply_pixmap(fz_context *ctx, fz_pixmap *pix)
 				s[k] = (s[k] * inva) >> 8;
 			s += pix->n;
 		}
+		s += stride;
 	}
 }
 
 fz_pixmap *
-fz_alpha_from_gray(fz_context *ctx, fz_pixmap *gray, int luminosity)
+fz_alpha_from_gray(fz_context *ctx, fz_pixmap *gray)
 {
 	fz_pixmap *alpha;
 	unsigned char *sp, *dp;
-	int len;
+	int w, h, sstride, dstride;
 	fz_irect bbox;
 
-	assert(gray->n == 2);
+	assert(gray->n == 1);
 
-	alpha = fz_new_pixmap_with_bbox(ctx, NULL, fz_pixmap_bbox(ctx, gray, &bbox));
+	alpha = fz_new_pixmap_with_bbox(ctx, NULL, fz_pixmap_bbox(ctx, gray, &bbox), 1);
 	dp = alpha->samples;
+	dstride = alpha->stride;
 	sp = gray->samples;
-	if (!luminosity)
-		sp ++;
+	sstride = gray->stride;
 
-	len = gray->w * gray->h;
-	while (len--)
+	h = gray->h;
+	w = gray->w;
+	while (h--)
 	{
-		*dp++ = sp[0];
-		sp += 2;
+		memcpy(dp, sp, w);
+		sp += sstride;
+		dp += dstride;
 	}
 
 	return alpha;
 }
 
+void
+fz_tint_pixmap(fz_context *ctx, fz_pixmap *pix, int r, int g, int b)
+{
+	unsigned char *s = pix->samples;
+	int x, y;
+
+	if (pix->colorspace == fz_device_bgr(ctx))
+	{
+		int save = r;
+		r = b;
+		b = save;
+	}
+	else if (pix->colorspace == fz_device_gray(ctx))
+	{
+		g = (r + g + b) / 3;
+	}
+	else if (pix->colorspace != fz_device_rgb(ctx))
+	{
+		fz_throw(ctx, FZ_ERROR_GENERIC, "can only tint RGB, BGR and Gray pixmaps");
+	}
+
+	if (pix->n == 4)
+	{
+		assert(pix->alpha);
+		for (y = 0; y < pix->h; y++)
+		{
+			for (x = 0; x < pix->w; x++)
+			{
+				s[0] = fz_mul255(s[0], r);
+				s[1] = fz_mul255(s[1], g);
+				s[2] = fz_mul255(s[2], b);
+				s += 4;
+			}
+			s += pix->stride - pix->w * 4;
+		}
+	}
+	else if (pix->n == 2)
+	{
+		assert(pix->alpha);
+		for (y = 0; y < pix->h; y++)
+		{
+			for (x = 0; x < pix->w; x++)
+			{
+				*s = fz_mul255(*s, g);
+				s += 2;
+			}
+			s += pix->stride - pix->w * 2;
+		}
+	}
+}
+
 void
 fz_invert_pixmap(fz_context *ctx, fz_pixmap *pix)
 {
 	unsigned char *s = pix->samples;
 	int k, x, y;
+	int n1 = pix->n - pix->alpha;
+	int n = pix->n;
 
 	for (y = 0; y < pix->h; y++)
 	{
 		for (x = 0; x < pix->w; x++)
 		{
-			for (k = 0; k < pix->n - 1; k++)
+			for (k = 0; k < n1; k++)
 				s[k] = 255 - s[k];
-			s += pix->n;
+			s += n;
 		}
+		s += pix->stride - pix->w * n;
 	}
 }
 
-void fz_invert_pixmap_rect(fz_pixmap *image, const fz_irect *rect)
+void fz_invert_pixmap_rect(fz_context *ctx, fz_pixmap *image, const fz_irect *rect)
 {
 	unsigned char *p;
 	int x, y, n;
 
-	int x0 = fz_clampi(rect->x0 - image->x, 0, image->w - 1);
-	int x1 = fz_clampi(rect->x1 - image->x, 0, image->w - 1);
-	int y0 = fz_clampi(rect->y0 - image->y, 0, image->h - 1);
-	int y1 = fz_clampi(rect->y1 - image->y, 0, image->h - 1);
+	int x0 = fz_clampi(rect->x0 - image->x, 0, image->w);
+	int x1 = fz_clampi(rect->x1 - image->x, 0, image->w);
+	int y0 = fz_clampi(rect->y0 - image->y, 0, image->h);
+	int y1 = fz_clampi(rect->y1 - image->y, 0, image->h);
 
 	for (y = y0; y < y1; y++)
 	{
-		p = image->samples + (unsigned int)((y * image->w + x0) * image->n);
+		p = image->samples + (unsigned int)((y * image->stride) + (x0 * image->n));
 		for (x = x0; x < x1; x++)
 		{
 			for (n = image->n; n > 1; n--, p++)
@@ -471,521 +843,11 @@ fz_gamma_pixmap(fz_context *ctx, fz_pixmap *pix, float gamma)
 				s[k] = gamma_map[s[k]];
 			s += pix->n;
 		}
+		s += pix->stride - pix->w * pix->n;
 	}
 }
 
-/*
- * Write pixmap to PNM file (without alpha channel)
- */
-
-void
-fz_output_pnm_header(fz_output *out, int w, int h, int n)
-{
-	fz_context *ctx = out->ctx;
-
-	if (n != 1 && n != 2 && n != 4)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap must be grayscale or rgb to write as pnm");
-
-	if (n == 1 || n == 2)
-		fz_printf(out, "P5\n");
-	if (n == 4)
-		fz_printf(out, "P6\n");
-	fz_printf(out, "%d %d\n", w, h);
-	fz_printf(out, "255\n");
-}
-
-void
-fz_output_pnm_band(fz_output *out, int w, int h, int n, int band, int bandheight, unsigned char *p)
-{
-	int len;
-	int start = band * bandheight;
-	int end = start + bandheight;
-
-	if (end > h)
-		end = h;
-	end -= start;
-
-	len = w * end;
-
-	switch (n)
-	{
-	case 1:
-		fz_write(out, p, len);
-		break;
-	case 2:
-		while (len--)
-		{
-			fz_putc(out, p[0]);
-			p += 2;
-		}
-		break;
-	case 4:
-		while (len--)
-		{
-			fz_putc(out, p[0]);
-			fz_putc(out, p[1]);
-			fz_putc(out, p[2]);
-			p += 4;
-		}
-	}
-}
-
-void
-fz_write_pnm(fz_context *ctx, fz_pixmap *pixmap, char *filename)
-{
-	fz_output *out = fz_new_output_to_filename(ctx, filename);
-	fz_output_pnm_header(out, pixmap->w, pixmap->h, pixmap->n);
-	fz_output_pnm_band(out, pixmap->w, pixmap->h, pixmap->n, 0, pixmap->h, pixmap->samples);
-	fz_close_output(out);
-}
-
-/*
- * Write pixmap to PAM file (with or without alpha channel)
- */
-
-void
-fz_output_pam_header(fz_output *out, int w, int h, int n, int savealpha)
-{
-	int sn = n;
-	int dn = n;
-	if (!savealpha && dn > 1)
-		dn--;
-
-	fz_printf(out, "P7\n");
-	fz_printf(out, "WIDTH %d\n", w);
-	fz_printf(out, "HEIGHT %d\n", h);
-	fz_printf(out, "DEPTH %d\n", dn);
-	fz_printf(out, "MAXVAL 255\n");
-	if (dn == 1) fz_printf(out, "TUPLTYPE GRAYSCALE\n");
-	else if (dn == 2 && sn == 2) fz_printf(out, "TUPLTYPE GRAYSCALE_ALPHA\n");
-	else if (dn == 3 && sn == 4) fz_printf(out, "TUPLTYPE RGB\n");
-	else if (dn == 4 && sn == 4) fz_printf(out, "TUPLTYPE RGB_ALPHA\n");
-	else if (dn == 4 && sn == 5) fz_printf(out, "TUPLTYPE CMYK\n");
-	else if (dn == 5 && sn == 5) fz_printf(out, "TUPLTYPE CMYK_ALPHA\n");
-	fz_printf(out, "ENDHDR\n");
-}
-
-void
-fz_output_pam_band(fz_output *out, int w, int h, int n, int band, int bandheight, unsigned char *sp, int savealpha)
-{
-	int y, x, k;
-	int start = band * bandheight;
-	int end = start + bandheight;
-	int sn = n;
-	int dn = n;
-	if (!savealpha && dn > 1)
-		dn--;
-
-	if (end > h)
-		end = h;
-	end -= start;
-
-	for (y = 0; y < end; y++)
-	{
-		x = w;
-		while (x--)
-		{
-			for (k = 0; k < dn; k++)
-				fz_putc(out, sp[k]);
-			sp += sn;
-		}
-	}
-}
-
-void
-fz_write_pam(fz_context *ctx, fz_pixmap *pixmap, char *filename, int savealpha)
-{
-	fz_output *out = fz_new_output_to_filename(ctx, filename);
-	fz_output_pam_header(out, pixmap->w, pixmap->h, pixmap->n, savealpha);
-	fz_output_pam_band(out, pixmap->w, pixmap->h, pixmap->n, 0, pixmap->h, pixmap->samples, savealpha);
-	fz_close_output(out);
-}
-
-/*
- * Write pixmap to PNG file (with or without alpha channel)
- */
-
-#include <zlib.h>
-
-static inline void big32(unsigned char *buf, unsigned int v)
-{
-	buf[0] = (v >> 24) & 0xff;
-	buf[1] = (v >> 16) & 0xff;
-	buf[2] = (v >> 8) & 0xff;
-	buf[3] = (v) & 0xff;
-}
-
-static void putchunk(char *tag, unsigned char *data, int size, fz_output *out)
-{
-	unsigned int sum;
-	fz_write_int32be(out, size);
-	fz_write(out, tag, 4);
-	fz_write(out, data, size);
-	sum = crc32(0, NULL, 0);
-	sum = crc32(sum, (unsigned char*)tag, 4);
-	sum = crc32(sum, data, size);
-	fz_write_int32be(out, sum);
-}
-
-void
-fz_write_png(fz_context *ctx, fz_pixmap *pixmap, char *filename, int savealpha)
-{
-	fz_output *out = fz_new_output_to_filename(ctx, filename);
-	fz_png_output_context *poc = NULL;
-
-	fz_var(poc);
-
-	fz_try(ctx)
-	{
-		poc = fz_output_png_header(out, pixmap->w, pixmap->h, pixmap->n, savealpha);
-		fz_output_png_band(out, pixmap->w, pixmap->h, pixmap->n, 0, pixmap->h, pixmap->samples, savealpha, poc);
-	}
-	fz_always(ctx)
-	{
-		fz_output_png_trailer(out, poc);
-		fz_close_output(out);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
-	}
-}
-
-void
-fz_output_png(fz_output *out, const fz_pixmap *pixmap, int savealpha)
-{
-	fz_png_output_context *poc;
-	fz_context *ctx;
-
-	if (!out)
-		return;
-
-	ctx = out->ctx;
-	poc = fz_output_png_header(out, pixmap->w, pixmap->h, pixmap->n, savealpha);
-
-	fz_try(ctx)
-	{
-		fz_output_png_band(out, pixmap->w, pixmap->h, pixmap->n, 0, pixmap->h, pixmap->samples, savealpha, poc);
-	}
-	fz_always(ctx)
-	{
-		fz_output_png_trailer(out, poc);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
-	}
-}
-
-struct fz_png_output_context_s
-{
-	unsigned char *udata;
-	unsigned char *cdata;
-	uLong usize, csize;
-	z_stream stream;
-};
-
-fz_png_output_context *
-fz_output_png_header(fz_output *out, int w, int h, int n, int savealpha)
-{
-	static const unsigned char pngsig[8] = { 137, 80, 78, 71, 13, 10, 26, 10 };
-	unsigned char head[13];
-	fz_context *ctx;
-	int color;
-	fz_png_output_context *poc;
-
-	if (!out)
-		return NULL;
-
-	ctx = out->ctx;
-
-	if (n != 1 && n != 2 && n != 4)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap must be grayscale or rgb to write as png");
-
-	poc = fz_malloc_struct(ctx, fz_png_output_context);
-
-	if (!savealpha && n > 1)
-		n--;
-
-	switch (n)
-	{
-	default:
-	case 1: color = 0; break;
-	case 2: color = 4; break;
-	case 3: color = 2; break;
-	case 4: color = 6; break;
-	}
-
-	big32(head+0, w);
-	big32(head+4, h);
-	head[8] = 8; /* depth */
-	head[9] = color;
-	head[10] = 0; /* compression */
-	head[11] = 0; /* filter */
-	head[12] = 0; /* interlace */
-
-	fz_write(out, pngsig, 8);
-	putchunk("IHDR", head, 13, out);
-
-	return poc;
-}
-
-void
-fz_output_png_band(fz_output *out, int w, int h, int n, int band, int bandheight, unsigned char *sp, int savealpha, fz_png_output_context *poc)
-{
-	unsigned char *dp;
-	int y, x, k, sn, dn, err, finalband;
-	fz_context *ctx;
-
-	if (!out || !sp || !poc)
-		return;
-
-	ctx = out->ctx;
-
-	if (n != 1 && n != 2 && n != 4)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap must be grayscale or rgb to write as png");
-
-	band *= bandheight;
-	finalband = (band+bandheight >= h);
-	if (finalband)
-		bandheight = h - band;
-
-	sn = n;
-	dn = n;
-	if (!savealpha && dn > 1)
-		dn--;
-
-	if (poc->udata == NULL)
-	{
-		poc->usize = (w * dn + 1) * bandheight;
-		/* Sadly the bound returned by compressBound is just for a
-		 * single usize chunk; if you compress a sequence of them
-		 * the buffering can result in you suddenly getting a block
-		 * larger than compressBound outputted in one go, even if you
-		 * take all the data out each time. */
-		poc->csize = compressBound(poc->usize);
-		fz_try(ctx)
-		{
-			poc->udata = fz_malloc(ctx, poc->usize);
-			poc->cdata = fz_malloc(ctx, poc->csize);
-		}
-		fz_catch(ctx)
-		{
-			fz_free(ctx, poc->udata);
-			poc->udata = NULL;
-			poc->cdata = NULL;
-			fz_rethrow(ctx);
-		}
-		err = deflateInit(&poc->stream, Z_DEFAULT_COMPRESSION);
-		if (err != Z_OK)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "compression error %d", err);
-	}
-
-	dp = poc->udata;
-	for (y = 0; y < bandheight; y++)
-	{
-		*dp++ = 1; /* sub prediction filter */
-		for (x = 0; x < w; x++)
-		{
-			for (k = 0; k < dn; k++)
-			{
-				if (x == 0)
-					dp[k] = sp[k];
-				else
-					dp[k] = sp[k] - sp[k-sn];
-			}
-			sp += sn;
-			dp += dn;
-		}
-	}
-
-	poc->stream.next_in = (Bytef*)poc->udata;
-	poc->stream.avail_in = (uInt)(dp - poc->udata);
-	do
-	{
-		poc->stream.next_out = poc->cdata;
-		poc->stream.avail_out = (uInt)poc->csize;
-
-		if (!finalband)
-		{
-			err = deflate(&poc->stream, Z_NO_FLUSH);
-			if (err != Z_OK)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "compression error %d", err);
-		}
-		else
-		{
-			err = deflate(&poc->stream, Z_FINISH);
-			if (err != Z_STREAM_END)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "compression error %d", err);
-		}
-
-		if (poc->stream.next_out != poc->cdata)
-			putchunk("IDAT", poc->cdata, poc->stream.next_out - poc->cdata, out);
-	}
-	while (poc->stream.avail_out == 0);
-}
-
-void
-fz_output_png_trailer(fz_output *out, fz_png_output_context *poc)
-{
-	unsigned char block[1];
-	int err;
-	fz_context *ctx;
-
-	if (!out || !poc)
-		return;
-
-	ctx = out->ctx;
-
-	err = deflateEnd(&poc->stream);
-	if (err != Z_OK)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "compression error %d", err);
-
-	fz_free(ctx, poc->cdata);
-	fz_free(ctx, poc->udata);
-	fz_free(ctx, poc);
-
-	putchunk("IEND", block, 0, out);
-}
-
-/* We use an auxiliary function to do pixmap_as_png, as it can enable us to
- * drop pix early in the case where we have to convert, potentially saving
- * us having to have 2 copies of the pixmap and a buffer open at once. */
-static fz_buffer *
-png_from_pixmap(fz_context *ctx, fz_pixmap *pix, int drop)
-{
-	fz_buffer *buf = NULL;
-	fz_output *out;
-	fz_pixmap *pix2 = NULL;
-
-	fz_var(buf);
-	fz_var(out);
-	fz_var(pix2);
-
-	if (pix->w == 0 || pix->h == 0)
-		return NULL;
-
-	fz_try(ctx)
-	{
-		if (pix->colorspace && pix->colorspace != fz_device_gray(ctx) && pix->colorspace != fz_device_rgb(ctx))
-		{
-			pix2 = fz_new_pixmap(ctx, fz_device_rgb(ctx), pix->w, pix->h);
-			fz_convert_pixmap(ctx, pix2, pix);
-			if (drop)
-				fz_drop_pixmap(ctx, pix);
-			pix = pix2;
-		}
-		buf = fz_new_buffer(ctx, 1024);
-		out = fz_new_output_with_buffer(ctx, buf);
-		fz_output_png(out, pix, 1);
-	}
-	fz_always(ctx)
-	{
-		fz_drop_pixmap(ctx, drop ? pix : pix2);
-		fz_close_output(out);
-	}
-	fz_catch(ctx)
-	{
-		fz_drop_buffer(ctx, buf);
-		fz_rethrow(ctx);
-	}
-	return buf;
-}
-
-fz_buffer *
-fz_new_png_from_image(fz_context *ctx, fz_image *image, int w, int h)
-{
-	fz_pixmap *pix = fz_image_get_pixmap(ctx, image, image->w, image->h);
-
-	return png_from_pixmap(ctx, pix, 1);
-}
-
-fz_buffer *
-fz_new_png_from_pixmap(fz_context *ctx, fz_pixmap *pix)
-{
-	return png_from_pixmap(ctx, pix, 0);
-}
-
-/*
- * Write pixmap to TGA file (with or without alpha channel)
- */
-
-static inline void tga_put_pixel(unsigned char *data, int n, int is_bgr, fz_output *out)
-{
-	if (n >= 3 && !is_bgr)
-	{
-		fz_putc(out, data[2]);
-		fz_putc(out, data[1]);
-		fz_putc(out, data[0]);
-		if (n == 4)
-			fz_putc(out, data[3]);
-		return;
-	}
-	if (n == 2)
-	{
-		fz_putc(out, data[0]);
-		fz_putc(out, data[0]);
-	}
-	fz_write(out, data, n);
-}
-
-void
-fz_write_tga(fz_context *ctx, fz_pixmap *pixmap, const char *filename, int savealpha)
-{
-	fz_output *out;
-	unsigned char head[18];
-	int n = pixmap->n;
-	int d = savealpha || n == 1 ? n : n - 1;
-	int is_bgr = pixmap->colorspace == fz_device_bgr(ctx);
-	int k;
-
-	if (pixmap->colorspace && pixmap->colorspace != fz_device_gray(ctx) &&
-		pixmap->colorspace != fz_device_rgb(ctx) && pixmap->colorspace != fz_device_bgr(ctx))
-	{
-		fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap must be grayscale or rgb to write as tga");
-	}
-
-	out = fz_new_output_to_filename(ctx, filename);
-
-	memset(head, 0, sizeof(head));
-	head[2] = n == 4 ? 10 : 11;
-	head[12] = pixmap->w & 0xFF; head[13] = (pixmap->w >> 8) & 0xFF;
-	head[14] = pixmap->h & 0xFF; head[15] = (pixmap->h >> 8) & 0xFF;
-	head[16] = d * 8;
-	head[17] = savealpha && n > 1 ? 8 : 0;
-	if (savealpha && d == 2)
-		head[16] = 32;
-
-	fz_write(out, head, sizeof(head));
-	for (k = 1; k <= pixmap->h; k++)
-	{
-		int i, j;
-		unsigned char *line = pixmap->samples + pixmap->w * n * (pixmap->h - k);
-		for (i = 0, j = 1; i < pixmap->w; i += j, j = 1)
-		{
-			for (; i + j < pixmap->w && j < 128 && !memcmp(line + i * n, line + (i + j) * n, d); j++);
-			if (j > 1)
-			{
-				fz_putc(out, j - 1 + 128);
-				tga_put_pixel(line + i * n, d, is_bgr, out);
-			}
-			else
-			{
-				for (; i + j < pixmap->w && j <= 128 && memcmp(line + (i + j - 1) * n, line + (i + j) * n, d) != 0; j++);
-				if (i + j < pixmap->w || j > 128)
-					j--;
-				fz_putc(out, j - 1);
-				for (; j > 0; j--, i++)
-					tga_put_pixel(line + i * n, d, is_bgr, out);
-			}
-		}
-	}
-	fz_write(out, "\0\0\0\0\0\0\0\0TRUEVISION-XFILE.\0", 26);
-
-	fz_close_output(out);
-}
-
-unsigned int
+size_t
 fz_pixmap_size(fz_context *ctx, fz_pixmap * pix)
 {
 	if (pix == NULL)
@@ -996,12 +858,17 @@ fz_pixmap_size(fz_context *ctx, fz_pixmap * pix)
 fz_pixmap *
 fz_new_pixmap_from_8bpp_data(fz_context *ctx, int x, int y, int w, int h, unsigned char *sp, int span)
 {
-	fz_pixmap *pixmap = fz_new_pixmap(ctx, NULL, w, h);
+	fz_pixmap *pixmap = fz_new_pixmap(ctx, NULL, w, h, 1);
+	int stride = pixmap->stride;
+	unsigned char *s = pixmap->samples;
 	pixmap->x = x;
 	pixmap->y = y;
 
 	for (y = 0; y < h; y++)
-		memcpy(pixmap->samples + y * w, sp + y * span, w);
+	{
+		memcpy(s, sp + y * span, w);
+		s += stride;
+	}
 
 	return pixmap;
 }
@@ -1009,7 +876,8 @@ fz_new_pixmap_from_8bpp_data(fz_context *ctx, int x, int y, int w, int h, unsign
 fz_pixmap *
 fz_new_pixmap_from_1bpp_data(fz_context *ctx, int x, int y, int w, int h, unsigned char *sp, int span)
 {
-	fz_pixmap *pixmap = fz_new_pixmap(ctx, NULL, w, h);
+	fz_pixmap *pixmap = fz_new_pixmap(ctx, NULL, w, h, 1);
+	int stride = pixmap->stride - pixmap->w;
 	pixmap->x = x;
 	pixmap->y = y;
 
@@ -1026,6 +894,7 @@ fz_new_pixmap_from_1bpp_data(fz_context *ctx, int x, int y, int w, int h, unsign
 			if (bit == 0)
 				bit = 0x80, in++;
 		}
+		out += stride;
 	}
 
 	return pixmap;
@@ -1215,11 +1084,17 @@ fz_subsample_pixmap_ARM(unsigned char *ptr, int w, int h, int f, int factor,
 void
 fz_subsample_pixmap(fz_context *ctx, fz_pixmap *tile, int factor)
 {
-	int dst_w, dst_h, w, h, fwd, fwd2, fwd3, back, back2, x, y, n, xx, yy, nn, f;
+	int dst_w, dst_h, w, h, fwd, fwd2, fwd3, back, back2, n, f;
 	unsigned char *s, *d;
+#ifndef ARCH_ARM
+	int x, y, xx, yy, nn;
+#endif
 
 	if (!tile)
 		return;
+
+	assert(tile->stride >= tile->w * tile->n);
+
 	s = d = tile->samples;
 	f = 1<<factor;
 	w = tile->w;
@@ -1227,11 +1102,11 @@ fz_subsample_pixmap(fz_context *ctx, fz_pixmap *tile, int factor)
 	n = tile->n;
 	dst_w = (w + f-1)>>factor;
 	dst_h = (h + f-1)>>factor;
-	fwd = w*n;
+	fwd = tile->stride;
 	back = f*fwd-n;
 	back2 = f*n-1;
 	fwd2 = (f-1)*n;
-	fwd3 = (f-1)*fwd;
+	fwd3 = (f-1)*fwd + tile->stride - w * n;
 	factor *= 2;
 #ifdef ARCH_ARM
 	{
@@ -1345,23 +1220,61 @@ fz_subsample_pixmap(fz_context *ctx, fz_pixmap *tile, int factor)
 #endif
 	tile->w = dst_w;
 	tile->h = dst_h;
+	tile->stride = dst_w * n;
 	tile->samples = fz_resize_array(ctx, tile->samples, dst_w * n, dst_h);
 }
 
 void
-fz_pixmap_set_resolution(fz_pixmap *pix, int res)
+fz_set_pixmap_resolution(fz_context *ctx, fz_pixmap *pix, int xres, int yres)
 {
-	pix->xres = res;
-	pix->yres = res;
+	pix->xres = xres;
+	pix->yres = yres;
 }
 
 void
-fz_md5_pixmap(fz_pixmap *pix, unsigned char digest[16])
+fz_md5_pixmap(fz_context *ctx, fz_pixmap *pix, unsigned char digest[16])
 {
 	fz_md5 md5;
 
 	fz_md5_init(&md5);
 	if (pix)
-		fz_md5_update(&md5, pix->samples, pix->w * pix->h * pix->n);
+	{
+		unsigned char *s = pix->samples;
+		int h = pix->h;
+		int ss = pix->stride;
+		int len = pix->w * pix->n;
+		while (h--)
+		{
+			fz_md5_update(&md5, s, len);
+			s += ss;
+		}
+	}
 	fz_md5_final(&md5, digest);
 }
+
+#ifdef HAVE_VALGRIND
+int fz_valgrind_pixmap(const fz_pixmap *pix)
+{
+	int w, h, n, total;
+	int ww, hh, nn;
+	int stride;
+	const unsigned char *p = pix->samples;
+
+	if (pix == NULL)
+		return 0;
+
+	total = 0;
+	ww = pix->w;
+	hh = pix->h;
+	nn = pix->n;
+	stride = pix->stride - ww*nn;
+	for (h = 0; h < hh; h++)
+	{
+		for (w = 0; w < ww; w++)
+			for (n = 0; n < nn; n++)
+				if (*p++) total ++;
+		p += stride;
+	}
+	return total;
+}
+#endif /* HAVE_VALGRIND */
