@@ -3,7 +3,7 @@
 /* Extract text into an unsorted span soup. */
 
 #define LINE_DIST 0.9f
-#define SPACE_DIST 0.2f
+#define SPACE_DIST 0.15f
 #define SPACE_MAX_DIST 0.8f
 #define PARAGRAPH_DIST 0.5f
 
@@ -18,20 +18,27 @@
 #include FT_FREETYPE_H
 #include FT_ADVANCES_H
 
-typedef struct fz_text_device_s fz_text_device;
+typedef struct fz_stext_device_s fz_stext_device;
 
 typedef struct span_soup_s span_soup;
 
-struct fz_text_device_s
+struct fz_stext_device_s
 {
 	fz_device super;
-	fz_text_sheet *sheet;
-	fz_text_page *page;
+	fz_stext_sheet *sheet;
+	fz_stext_page *page;
 	span_soup *spans;
-	fz_text_span *cur_span;
+	fz_stext_span *cur_span;
 	int lastchar;
+	int flags;
 };
 
+const char *fz_stext_options_usage =
+	"Structured text output options:\n"
+	"\tpreserve-ligatures: do not expand all ligatures into constituent characters\n"
+	"\tpreserve-whitespace: do not convert all whitespace characters into spaces\n"
+	"\n";
+
 static fz_rect *
 add_point_to_rect(fz_rect *a, const fz_point *p)
 {
@@ -47,11 +54,11 @@ add_point_to_rect(fz_rect *a, const fz_point *p)
 }
 
 fz_rect *
-fz_text_char_bbox(fz_context *ctx, fz_rect *bbox, fz_text_span *span, int i)
+fz_stext_char_bbox(fz_context *ctx, fz_rect *bbox, fz_stext_span *span, int i)
 {
 	fz_point a, d;
 	const fz_point *max;
-	fz_text_char *ch;
+	fz_stext_char *ch;
 
 	if (!span || i >= span->len)
 	{
@@ -94,7 +101,7 @@ fz_text_char_bbox(fz_context *ctx, fz_rect *bbox, fz_text_span *span, int i)
 }
 
 static void
-add_bbox_to_span(fz_text_span *span)
+add_bbox_to_span(fz_stext_span *span)
 {
 	fz_point a, d;
 	fz_rect *bbox = &span->bbox;
@@ -133,7 +140,7 @@ add_bbox_to_span(fz_text_span *span)
 struct span_soup_s
 {
 	int len, cap;
-	fz_text_span **spans;
+	fz_stext_span **spans;
 };
 
 static span_soup *
@@ -162,7 +169,7 @@ free_span_soup(fz_context *ctx, span_soup *soup)
 }
 
 static void
-add_span_to_soup(fz_context *ctx, span_soup *soup, fz_text_span *span)
+add_span_to_soup(fz_context *ctx, span_soup *soup, fz_stext_span *span)
 {
 	if (span == NULL)
 		return;
@@ -176,12 +183,12 @@ add_span_to_soup(fz_context *ctx, span_soup *soup, fz_text_span *span)
 	soup->spans[soup->len++] = span;
 }
 
-static fz_text_line *
-push_span(fz_context *ctx, fz_text_device *tdev, fz_text_span *span, int new_line, float distance)
+static fz_stext_line *
+push_span(fz_context *ctx, fz_stext_device *tdev, fz_stext_span *span, int new_line, float distance)
 {
-	fz_text_line *line;
-	fz_text_block *block;
-	fz_text_page *page = tdev->page;
+	fz_stext_line *line;
+	fz_stext_block *block;
+	fz_stext_page *page = tdev->page;
 	int prev_not_text = 0;
 
 	if (page->len == 0 || page->blocks[page->len-1].type != FZ_PAGE_BLOCK_TEXT)
@@ -200,7 +207,7 @@ push_span(fz_context *ctx, fz_text_device *tdev, fz_text_span *span, int new_lin
 				page->blocks = fz_resize_array(ctx, page->blocks, newcap, sizeof(*page->blocks));
 				page->cap = newcap;
 			}
-			block = fz_malloc_struct(ctx, fz_text_block);
+			block = fz_malloc_struct(ctx, fz_stext_block);
 			page->blocks[page->len].type = FZ_PAGE_BLOCK_TEXT;
 			page->blocks[page->len].u.text = block;
 			block->cap = 0;
@@ -250,7 +257,7 @@ push_span(fz_context *ctx, fz_text_device *tdev, fz_text_span *span, int new_lin
 
 #if defined(DEBUG_SPANS) || defined(DEBUG_ALIGN) || defined(DEBUG_INDENTS)
 static void
-dump_span(fz_text_span *s)
+dump_span(fz_stext_span *s)
 {
 	int i;
 	for (i=0; i < s->len; i++)
@@ -262,12 +269,12 @@ dump_span(fz_text_span *s)
 
 #ifdef DEBUG_ALIGN
 static void
-dump_line(fz_text_line *line)
+dump_line(fz_stext_line *line)
 {
 	int i;
 	for (i=0; i < line->len; i++)
 	{
-		fz_text_span *s = line->spans[i];
+		fz_stext_span *s = line->spans[i];
 		if (s->spacing > 1)
 			printf(" ");
 		dump_span(s);
@@ -277,11 +284,11 @@ dump_line(fz_text_line *line)
 #endif
 
 static void
-strain_soup(fz_context *ctx, fz_text_device *tdev)
+strain_soup(fz_context *ctx, fz_stext_device *tdev)
 {
 	span_soup *soup = tdev->spans;
-	fz_text_line *last_line = NULL;
-	fz_text_span *last_span = NULL;
+	fz_stext_line *last_line = NULL;
+	fz_stext_span *last_span = NULL;
 	int span_num;
 
 	if (soup == NULL)
@@ -290,7 +297,7 @@ strain_soup(fz_context *ctx, fz_text_device *tdev)
 	/* Really dumb implementation to match what we had before */
 	for (span_num=0; span_num < soup->len; span_num++)
 	{
-		fz_text_span *span = soup->spans[span_num];
+		fz_stext_span *span = soup->spans[span_num];
 		int new_line = 1;
 		float distance = 0;
 		float spacing = 0;
@@ -379,19 +386,19 @@ strain_soup(fz_context *ctx, fz_text_device *tdev)
 	}
 }
 
-fz_text_sheet *
-fz_new_text_sheet(fz_context *ctx)
+fz_stext_sheet *
+fz_new_stext_sheet(fz_context *ctx)
 {
-	fz_text_sheet *sheet = fz_malloc(ctx, sizeof *sheet);
+	fz_stext_sheet *sheet = fz_malloc(ctx, sizeof *sheet);
 	sheet->maxid = 0;
 	sheet->style = NULL;
 	return sheet;
 }
 
 void
-fz_drop_text_sheet(fz_context *ctx, fz_text_sheet *sheet)
+fz_drop_stext_sheet(fz_context *ctx, fz_stext_sheet *sheet)
 {
-	fz_text_style *style;
+	fz_stext_style *style;
 
 	if (sheet == NULL)
 		return;
@@ -399,7 +406,7 @@ fz_drop_text_sheet(fz_context *ctx, fz_text_sheet *sheet)
 	style = sheet->style;
 	while (style)
 	{
-		fz_text_style *next = style->next;
+		fz_stext_style *next = style->next;
 		fz_drop_font(ctx, style->font);
 		fz_free(ctx, style);
 		style = next;
@@ -407,11 +414,11 @@ fz_drop_text_sheet(fz_context *ctx, fz_text_sheet *sheet)
 	fz_free(ctx, sheet);
 }
 
-static fz_text_style *
-fz_lookup_text_style_imp(fz_context *ctx, fz_text_sheet *sheet,
+static fz_stext_style *
+fz_lookup_stext_style_imp(fz_context *ctx, fz_stext_sheet *sheet,
 	float size, fz_font *font, int wmode, int script)
 {
-	fz_text_style *style;
+	fz_stext_style *style;
 
 	for (style = sheet->style; style; style = style->next)
 	{
@@ -436,30 +443,30 @@ fz_lookup_text_style_imp(fz_context *ctx, fz_text_sheet *sheet,
 	return style;
 }
 
-static fz_text_style *
-fz_lookup_text_style(fz_context *ctx, fz_text_sheet *sheet, fz_text *text, const fz_matrix *ctm,
-	fz_colorspace *colorspace, float *color, float alpha, fz_stroke_state *stroke)
+static fz_stext_style *
+fz_lookup_stext_style(fz_context *ctx, fz_stext_sheet *sheet, fz_text_span *span, const fz_matrix *ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, const fz_stroke_state *stroke)
 {
 	float size = 1.0f;
-	fz_font *font = text ? text->font : NULL;
-	int wmode = text ? text->wmode : 0;
-	if (ctm && text)
+	fz_font *font = span ? span->font : NULL;
+	int wmode = span ? span->wmode : 0;
+	if (ctm && span)
 	{
-		fz_matrix tm = text->trm;
+		fz_matrix tm = span->trm;
 		fz_matrix trm;
 		tm.e = 0;
 		tm.f = 0;
 		fz_concat(&trm, &tm, ctm);
 		size = fz_matrix_expansion(&trm);
 	}
-	return fz_lookup_text_style_imp(ctx, sheet, size, font, wmode, 0);
+	return fz_lookup_stext_style_imp(ctx, sheet, size, font, wmode, 0);
 }
 
-fz_text_page *
-fz_new_text_page(fz_context *ctx)
+fz_stext_page *
+fz_new_stext_page(fz_context *ctx, const fz_rect *mediabox)
 {
-	fz_text_page *page = fz_malloc(ctx, sizeof(*page));
-	page->mediabox = fz_empty_rect;
+	fz_stext_page *page = fz_malloc(ctx, sizeof(*page));
+	page->mediabox = *mediabox;
 	page->len = 0;
 	page->cap = 0;
 	page->blocks = NULL;
@@ -468,9 +475,9 @@ fz_new_text_page(fz_context *ctx)
 }
 
 static void
-fz_drop_text_line_contents(fz_context *ctx, fz_text_line *line)
+fz_drop_stext_line_contents(fz_context *ctx, fz_stext_line *line)
 {
-	fz_text_span *span, *next;
+	fz_stext_span *span, *next;
 	for (span = line->first_span; span; span=next)
 	{
 		next = span->next;
@@ -480,13 +487,13 @@ fz_drop_text_line_contents(fz_context *ctx, fz_text_line *line)
 }
 
 static void
-fz_drop_text_block(fz_context *ctx, fz_text_block *block)
+fz_drop_stext_block(fz_context *ctx, fz_stext_block *block)
 {
-	fz_text_line *line;
+	fz_stext_line *line;
 	if (block == NULL)
 		return;
 	for (line = block->lines; line < block->lines + block->len; line++)
-		fz_drop_text_line_contents(ctx, line);
+		fz_drop_stext_line_contents(ctx, line);
 	fz_free(ctx, block->lines);
 	fz_free(ctx, block);
 }
@@ -502,7 +509,7 @@ fz_drop_image_block(fz_context *ctx, fz_image_block *block)
 }
 
 void
-fz_drop_text_page(fz_context *ctx, fz_text_page *page)
+fz_drop_stext_page(fz_context *ctx, fz_stext_page *page)
 {
 	fz_page_block *block;
 	if (page == NULL)
@@ -512,7 +519,7 @@ fz_drop_text_page(fz_context *ctx, fz_text_page *page)
 		switch (block->type)
 		{
 		case FZ_PAGE_BLOCK_TEXT:
-			fz_drop_text_block(ctx, block->u.text);
+			fz_drop_stext_block(ctx, block->u.text);
 			break;
 		case FZ_PAGE_BLOCK_IMAGE:
 			fz_drop_image_block(ctx, block->u.image);
@@ -523,10 +530,10 @@ fz_drop_text_page(fz_context *ctx, fz_text_page *page)
 	fz_free(ctx, page);
 }
 
-static fz_text_span *
-fz_new_text_span(fz_context *ctx, const fz_point *p, int wmode, const fz_matrix *trm)
+static fz_stext_span *
+fz_new_stext_span(fz_context *ctx, const fz_point *p, int wmode, const fz_matrix *trm)
 {
-	fz_text_span *span = fz_malloc_struct(ctx, fz_text_span);
+	fz_stext_span *span = fz_malloc_struct(ctx, fz_stext_span);
 	span->ascender_max = 0;
 	span->descender_min = 0;
 	span->cap = 0;
@@ -546,12 +553,12 @@ fz_new_text_span(fz_context *ctx, const fz_point *p, int wmode, const fz_matrix
 }
 
 static void
-add_char_to_span(fz_context *ctx, fz_text_span *span, int c, fz_point *p, fz_point *max, fz_text_style *style)
+add_char_to_span(fz_context *ctx, fz_stext_span *span, int c, fz_point *p, fz_point *max, fz_stext_style *style)
 {
 	if (span->len == span->cap)
 	{
 		int newcap = (span->cap ? span->cap * 2 : 16);
-		span->text = fz_resize_array(ctx, span->text, newcap, sizeof(fz_text_char));
+		span->text = fz_resize_array(ctx, span->text, newcap, sizeof(fz_stext_char));
 		span->cap = newcap;
 		span->bbox = fz_empty_rect;
 	}
@@ -567,7 +574,7 @@ add_char_to_span(fz_context *ctx, fz_text_span *span, int c, fz_point *p, fz_poi
 }
 
 static void
-fz_add_text_char_imp(fz_context *ctx, fz_text_device *dev, fz_text_style *style, int c, fz_matrix *trm, float adv, int wmode)
+fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_stext_style *style, int c, int glyph, fz_matrix *trm, float adv, int wmode)
 {
 	int can_append = 1;
 	int add_space = 0;
@@ -624,6 +631,14 @@ fz_add_text_char_imp(fz_context *ctx, fz_text_device *dev, fz_text_style *style,
 		q.y = trm->f;
 	}
 
+	if (glyph < 0)
+	{
+		/* Don't reset 'pen' to start of no-glyph characters in cluster */
+		if (dev->cur_span)
+			q = dev->cur_span->max;
+		goto no_glyph;
+	}
+
 	if (dev->cur_span == NULL ||
 		trm->a != dev->cur_span->transform.a || trm->b != dev->cur_span->transform.b ||
 		trm->c != dev->cur_span->transform.c || trm->d != dev->cur_span->transform.d ||
@@ -638,6 +653,11 @@ fz_add_text_char_imp(fz_context *ctx, fz_text_device *dev, fz_text_style *style,
 	}
 	else
 	{
+		delta.x = q.x - dev->cur_span->max.x;
+		delta.y = q.y - dev->cur_span->max.y;
+		if (delta.x < FLT_EPSILON && delta.y < FLT_EPSILON && c == dev->lastchar)
+			return;
+
 		/* Calculate how far we've moved since the end of the current
 		 * span. */
 		delta.x = p.x - dev->cur_span->max.x;
@@ -651,11 +671,10 @@ fz_add_text_char_imp(fz_context *ctx, fz_text_device *dev, fz_text_style *style,
 		base_offset = -ndir.y * delta.x + ndir.x * delta.y;
 
 		spacing /= size * SPACE_DIST;
-		spacing = fabsf(spacing);
 		if (fabsf(base_offset) < size * 0.1)
 		{
 			/* Only a small amount off the baseline - we'll take this */
-			if (spacing < 1.0)
+			if (fabsf(spacing) < 1.0)
 			{
 				/* Motion is in line, and small. */
 			}
@@ -685,78 +704,110 @@ fz_add_text_char_imp(fz_context *ctx, fz_text_device *dev, fz_text_style *style,
 	printf("%c%c append=%d space=%d size=%g spacing=%g base_offset=%g\n", dev->lastchar, c, can_append, add_space, size, spacing, base_offset);
 #endif
 
-	if (can_append == 0)
+	/* Start a new span */
+	if (!can_append)
 	{
-		/* Start a new span */
 		add_span_to_soup(ctx, dev->spans, dev->cur_span);
 		dev->cur_span = NULL;
-		dev->cur_span = fz_new_text_span(ctx, &p, wmode, trm);
+		dev->cur_span = fz_new_stext_span(ctx, &p, wmode, trm);
 		dev->cur_span->spacing = 0;
 	}
+
+	/* Add synthetic space */
 	if (add_space)
 	{
-		r.x = - 0.2f;
-		r.y = 0;
-		fz_transform_point(&r, trm);
-		add_char_to_span(ctx, dev->cur_span, ' ', &p, &r, style);
+		/* We know we always have a cur_span here */
+		r = dev->cur_span->max;
+		add_char_to_span(ctx, dev->cur_span, ' ', &r, &p, style);
 	}
+
+no_glyph:
 	add_char_to_span(ctx, dev->cur_span, c, &p, &q, style);
+	dev->lastchar = c;
 }
 
 static void
-fz_add_text_char(fz_context *ctx, fz_text_device *dev, fz_text_style *style, int c, fz_matrix *trm, float adv, int wmode)
+fz_add_stext_char(fz_context *ctx, fz_stext_device *dev, fz_stext_style *style, int c, int glyph, fz_matrix *trm, float adv, int wmode)
 {
-	switch (c)
-	{
-	case -1: /* ignore when one unicode character maps to multiple glyphs */
-		break;
-	case 0xFB00: /* ff */
-		fz_add_text_char_imp(ctx, dev, style, 'f', trm, adv/2, wmode);
-		fz_add_text_char_imp(ctx, dev, style, 'f', trm, adv/2, wmode);
-		break;
-	case 0xFB01: /* fi */
-		fz_add_text_char_imp(ctx, dev, style, 'f', trm, adv/2, wmode);
-		fz_add_text_char_imp(ctx, dev, style, 'i', trm, adv/2, wmode);
-		break;
-	case 0xFB02: /* fl */
-		fz_add_text_char_imp(ctx, dev, style, 'f', trm, adv/2, wmode);
-		fz_add_text_char_imp(ctx, dev, style, 'l', trm, adv/2, wmode);
-		break;
-	case 0xFB03: /* ffi */
-		fz_add_text_char_imp(ctx, dev, style, 'f', trm, adv/3, wmode);
-		fz_add_text_char_imp(ctx, dev, style, 'f', trm, adv/3, wmode);
-		fz_add_text_char_imp(ctx, dev, style, 'i', trm, adv/3, wmode);
-		break;
-	case 0xFB04: /* ffl */
-		fz_add_text_char_imp(ctx, dev, style, 'f', trm, adv/3, wmode);
-		fz_add_text_char_imp(ctx, dev, style, 'f', trm, adv/3, wmode);
-		fz_add_text_char_imp(ctx, dev, style, 'l', trm, adv/3, wmode);
-		break;
-	case 0xFB05: /* long st */
-	case 0xFB06: /* st */
-		fz_add_text_char_imp(ctx, dev, style, 's', trm, adv/2, wmode);
-		fz_add_text_char_imp(ctx, dev, style, 't', trm, adv/2, wmode);
-		break;
-	default:
-		fz_add_text_char_imp(ctx, dev, style, c, trm, adv, wmode);
-		break;
-	}
+	/* ignore when one unicode character maps to multiple glyphs */
+	if (c == -1)
+		return;
+
+	if (!(dev->flags & FZ_STEXT_PRESERVE_LIGATURES))
+		switch (c)
+		{
+		case 0xFB00: /* ff */
+			fz_add_stext_char_imp(ctx, dev, style, 'f', glyph, trm, adv, wmode);
+			fz_add_stext_char_imp(ctx, dev, style, 'f', -1, trm, 0, wmode);
+			return;
+		case 0xFB01: /* fi */
+			fz_add_stext_char_imp(ctx, dev, style, 'f', glyph, trm, adv, wmode);
+			fz_add_stext_char_imp(ctx, dev, style, 'i', -1, trm, 0, wmode);
+			return;
+		case 0xFB02: /* fl */
+			fz_add_stext_char_imp(ctx, dev, style, 'f', glyph, trm, adv, wmode);
+			fz_add_stext_char_imp(ctx, dev, style, 'l', -1, trm, 0, wmode);
+			return;
+		case 0xFB03: /* ffi */
+			fz_add_stext_char_imp(ctx, dev, style, 'f', glyph, trm, adv, wmode);
+			fz_add_stext_char_imp(ctx, dev, style, 'f', -1, trm, 0, wmode);
+			fz_add_stext_char_imp(ctx, dev, style, 'i', -1, trm, 0, wmode);
+			return;
+		case 0xFB04: /* ffl */
+			fz_add_stext_char_imp(ctx, dev, style, 'f', glyph, trm, adv, wmode);
+			fz_add_stext_char_imp(ctx, dev, style, 'f', -1, trm, 0, wmode);
+			fz_add_stext_char_imp(ctx, dev, style, 'l', -1, trm, 0, wmode);
+			return;
+		case 0xFB05: /* long st */
+		case 0xFB06: /* st */
+			fz_add_stext_char_imp(ctx, dev, style, 's', glyph, trm, adv, wmode);
+			fz_add_stext_char_imp(ctx, dev, style, 't', -1, trm, 0, wmode);
+			return;
+		}
+
+	if (!(dev->flags & FZ_STEXT_PRESERVE_WHITESPACE))
+		switch (c)
+		{
+		case 0x0009: /* tab */
+		case 0x0020: /* space */
+		case 0x00A0: /* no-break space */
+		case 0x1680: /* ogham space mark */
+		case 0x180E: /* mongolian vowel separator */
+		case 0x2000: /* en quad */
+		case 0x2001: /* em quad */
+		case 0x2002: /* en space */
+		case 0x2003: /* em space */
+		case 0x2004: /* three-per-em space */
+		case 0x2005: /* four-per-em space */
+		case 0x2006: /* six-per-em space */
+		case 0x2007: /* figure space */
+		case 0x2008: /* punctuation space */
+		case 0x2009: /* thin space */
+		case 0x200A: /* hair space */
+		case 0x202F: /* narrow no-break space */
+		case 0x205F: /* medium mathematical space */
+		case 0x3000: /* ideographic space */
+			c = ' ';
+		}
+
+	fz_add_stext_char_imp(ctx, dev, style, c, glyph, trm, adv, wmode);
 }
 
 static void
-fz_text_extract(fz_context *ctx, fz_text_device *dev, fz_text *text, const fz_matrix *ctm, fz_text_style *style)
+fz_stext_extract(fz_context *ctx, fz_stext_device *dev, fz_text_span *span, const fz_matrix *ctm, fz_stext_style *style)
 {
-	fz_font *font = text->font;
-	FT_Face face = font->ft_face;
-	fz_matrix tm = text->trm;
+	fz_font *font = span->font;
+	FT_Face face = fz_font_ft_face(ctx, font);
+	fz_buffer **t3procs = fz_font_t3_procs(ctx, font);
+	fz_rect *bbox = fz_font_bbox(ctx, font);
+	fz_matrix tm = span->trm;
 	fz_matrix trm;
 	float adv;
 	float ascender = 1;
 	float descender = 0;
-	int multi;
-	int i, j, err;
+	int i, err;
 
-	if (text->len == 0)
+	if (span->len == 0)
 		return;
 
 	if (dev->spans == NULL)
@@ -764,26 +815,26 @@ fz_text_extract(fz_context *ctx, fz_text_device *dev, fz_text *text, const fz_ma
 
 	if (style->wmode == 0)
 	{
-		if (font->ft_face)
+		if (face)
 		{
 			fz_lock(ctx, FZ_LOCK_FREETYPE);
-			err = FT_Set_Char_Size(font->ft_face, 64, 64, 72, 72);
+			err = FT_Set_Char_Size(face, 64, 64, 72, 72);
 			if (err)
 				fz_warn(ctx, "freetype set character size: %s", ft_error_string(err));
 			ascender = (float)face->ascender / face->units_per_EM;
 			descender = (float)face->descender / face->units_per_EM;
 			fz_unlock(ctx, FZ_LOCK_FREETYPE);
 		}
-		else if (font->t3procs && !fz_is_empty_rect(&font->bbox))
+		else if (t3procs && !fz_is_empty_rect(bbox))
 		{
-			ascender = font->bbox.y1;
-			descender = font->bbox.y0;
+			ascender = bbox->y1;
+			descender = bbox->y0;
 		}
 	}
 	else
 	{
-		ascender = font->bbox.x1;
-		descender = font->bbox.x0;
+		ascender = bbox->x1;
+		descender = bbox->x0;
 	}
 	style->ascender = ascender;
 	style->descender = descender;
@@ -792,92 +843,96 @@ fz_text_extract(fz_context *ctx, fz_text_device *dev, fz_text *text, const fz_ma
 	tm.f = 0;
 	fz_concat(&trm, &tm, ctm);
 
-	for (i = 0; i < text->len; i++)
+	for (i = 0; i < span->len; i++)
 	{
 		/* Calculate new pen location and delta */
-		tm.e = text->items[i].x;
-		tm.f = text->items[i].y;
+		tm.e = span->items[i].x;
+		tm.f = span->items[i].y;
 		fz_concat(&trm, &tm, ctm);
 
 		/* Calculate bounding box and new pen position based on font metrics */
-		adv = fz_advance_glyph(ctx, font, text->items[i].gid);
-
-		/* Check for one glyph to many char mapping */
-		for (j = i + 1; j < text->len; j++)
-			if (text->items[j].gid >= 0)
-				break;
-		multi = j - i;
-
-		if (multi == 1)
-		{
-			fz_add_text_char(ctx, dev, style, text->items[i].ucs, &trm, adv, text->wmode);
-		}
+		if (span->items[i].gid >= 0)
+			adv = fz_advance_glyph(ctx, font, span->items[i].gid, style->wmode);
 		else
-		{
-			for (j = 0; j < multi; j++)
-			{
-				fz_add_text_char(ctx, dev, style, text->items[i + j].ucs, &trm, adv/multi, text->wmode);
-			}
-			i += j - 1;
-		}
+			adv = 0;
 
-		dev->lastchar = text->items[i].ucs;
+		fz_add_stext_char(ctx, dev, style, span->items[i].ucs, span->items[i].gid, &trm, adv, span->wmode);
 	}
 }
 
 static void
-fz_text_fill_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm,
-	fz_colorspace *colorspace, float *color, float alpha)
+fz_stext_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm,
+	fz_colorspace *colorspace, const float *color, float alpha)
 {
-	fz_text_device *tdev = (fz_text_device*)dev;
-	fz_text_style *style;
-	style = fz_lookup_text_style(ctx, tdev->sheet, text, ctm, colorspace, color, alpha, NULL);
-	fz_text_extract(ctx, tdev, text, ctm, style);
+	fz_stext_device *tdev = (fz_stext_device*)dev;
+	fz_stext_style *style;
+	fz_text_span *span;
+	for (span = text->head; span; span = span->next)
+	{
+		style = fz_lookup_stext_style(ctx, tdev->sheet, span, ctm, colorspace, color, alpha, NULL);
+		fz_stext_extract(ctx, tdev, span, ctm, style);
+	}
 }
 
 static void
-fz_text_stroke_text(fz_context *ctx, fz_device *dev, fz_text *text, fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, float *color, float alpha)
+fz_stext_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm,
+	fz_colorspace *colorspace, const float *color, float alpha)
 {
-	fz_text_device *tdev = (fz_text_device*)dev;
-	fz_text_style *style;
-	style = fz_lookup_text_style(ctx, tdev->sheet, text, ctm, colorspace, color, alpha, stroke);
-	fz_text_extract(ctx, tdev, text, ctm, style);
+	fz_stext_device *tdev = (fz_stext_device*)dev;
+	fz_stext_style *style;
+	fz_text_span *span;
+	for (span = text->head; span; span = span->next)
+	{
+		style = fz_lookup_stext_style(ctx, tdev->sheet, span, ctm, colorspace, color, alpha, stroke);
+		fz_stext_extract(ctx, tdev, span, ctm, style);
+	}
 }
 
 static void
-fz_text_clip_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm, int accumulate)
+fz_stext_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm, const fz_rect *scissor)
 {
-	fz_text_device *tdev = (fz_text_device*)dev;
-	fz_text_style *style;
-	style = fz_lookup_text_style(ctx, tdev->sheet, text, ctm, NULL, NULL, 0, NULL);
-	fz_text_extract(ctx, tdev, text, ctm, style);
+	fz_stext_device *tdev = (fz_stext_device*)dev;
+	fz_stext_style *style;
+	fz_text_span *span;
+	for (span = text->head; span; span = span->next)
+	{
+		style = fz_lookup_stext_style(ctx, tdev->sheet, span, ctm, NULL, NULL, 0, NULL);
+		fz_stext_extract(ctx, tdev, span, ctm, style);
+	}
 }
 
 static void
-fz_text_clip_stroke_text(fz_context *ctx, fz_device *dev, fz_text *text, fz_stroke_state *stroke, const fz_matrix *ctm)
+fz_stext_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm, const fz_rect *scissor)
 {
-	fz_text_device *tdev = (fz_text_device*)dev;
-	fz_text_style *style;
-	style = fz_lookup_text_style(ctx, tdev->sheet, text, ctm, NULL, NULL, 0, stroke);
-	fz_text_extract(ctx, tdev, text, ctm, style);
+	fz_stext_device *tdev = (fz_stext_device*)dev;
+	fz_stext_style *style;
+	fz_text_span *span;
+	for (span = text->head; span; span = span->next)
+	{
+		style = fz_lookup_stext_style(ctx, tdev->sheet, span, ctm, NULL, NULL, 0, stroke);
+		fz_stext_extract(ctx, tdev, span, ctm, style);
+	}
 }
 
 static void
-fz_text_ignore_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm)
+fz_stext_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm)
 {
-	fz_text_device *tdev = (fz_text_device*)dev;
-	fz_text_style *style;
-	style = fz_lookup_text_style(ctx, tdev->sheet, text, ctm, NULL, NULL, 0, NULL);
-	fz_text_extract(ctx, tdev, text, ctm, style);
+	fz_stext_device *tdev = (fz_stext_device*)dev;
+	fz_stext_style *style;
+	fz_text_span *span;
+	for (span = text->head; span; span = span->next)
+	{
+		style = fz_lookup_stext_style(ctx, tdev->sheet, span, ctm, NULL, NULL, 0, NULL);
+		fz_stext_extract(ctx, tdev, span, ctm, style);
+	}
 }
 
 static void
-fz_text_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *img, const fz_matrix *ctm,
-		fz_colorspace *cspace, float *color, float alpha)
+fz_stext_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *img, const fz_matrix *ctm,
+		fz_colorspace *cspace, const float *color, float alpha)
 {
-	fz_text_device *tdev = (fz_text_device*)dev;
-	fz_text_page *page = tdev->page;
+	fz_stext_device *tdev = (fz_stext_device*)dev;
+	fz_stext_page *page = tdev->page;
 	fz_image_block *block;
 
 	/* If the alpha is less than 50% then it's probably a watermark or
@@ -898,7 +953,7 @@ fz_text_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *img, const fz
 	block->image = fz_keep_image(ctx, img);
 	block->cspace = fz_keep_colorspace(ctx, cspace);
 	if (cspace)
-		memcpy(block->colors, color, sizeof(block->colors[0])*cspace->n);
+		memcpy(block->colors, color, sizeof(block->colors[0])*fz_colorspace_n(ctx, cspace));
 	block->mat = *ctm;
 	block->bbox.x0 = 0;
 	block->bbox.y0 = 0;
@@ -909,13 +964,13 @@ fz_text_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *img, const fz
 }
 
 static void
-fz_text_fill_image(fz_context *ctx, fz_device *dev, fz_image *img, const fz_matrix *ctm, float alpha)
+fz_stext_fill_image(fz_context *ctx, fz_device *dev, fz_image *img, const fz_matrix *ctm, float alpha)
 {
-	fz_text_fill_image_mask(ctx, dev, img, ctm, NULL, NULL, alpha);
+	fz_stext_fill_image_mask(ctx, dev, img, ctm, NULL, NULL, alpha);
 }
 
 static int
-fz_bidi_direction(int bidiclass, int curdir)
+direction_from_bidi_class(int bidiclass, int curdir)
 {
 	switch (bidiclass)
 	{
@@ -948,11 +1003,11 @@ fz_bidi_direction(int bidiclass, int curdir)
 }
 
 static void
-fz_bidi_reorder_run(fz_text_span *span, int a, int b, int dir)
+fz_bidi_reorder_run(fz_stext_span *span, int a, int b, int dir)
 {
 	if (a < b && dir == -1)
 	{
-		fz_text_char c;
+		fz_stext_char c;
 		int m = a + (b - a) / 2;
 		while (a < m)
 		{
@@ -966,7 +1021,7 @@ fz_bidi_reorder_run(fz_text_span *span, int a, int b, int dir)
 }
 
 static void
-fz_bidi_reorder_span(fz_text_span *span)
+fz_bidi_reorder_span(fz_stext_span *span)
 {
 	int a, b, dir, curdir;
 
@@ -974,7 +1029,7 @@ fz_bidi_reorder_span(fz_text_span *span)
 	curdir = 1;
 	for (b = 0; b < span->len; b++)
 	{
-		dir = fz_bidi_direction(ucdn_get_bidi_class(span->text[b].c), curdir);
+		dir = direction_from_bidi_class(ucdn_get_bidi_class(span->text[b].c), curdir);
 		if (dir != curdir)
 		{
 			fz_bidi_reorder_run(span, a, b, curdir);
@@ -986,12 +1041,12 @@ fz_bidi_reorder_span(fz_text_span *span)
 }
 
 static void
-fz_bidi_reorder_text_page(fz_context *ctx, fz_text_page *page)
+fz_bidi_reorder_stext_page(fz_context *ctx, fz_stext_page *page)
 {
 	fz_page_block *pageblock;
-	fz_text_block *block;
-	fz_text_line *line;
-	fz_text_span *span;
+	fz_stext_block *block;
+	fz_stext_line *line;
+	fz_stext_span *span;
 
 	for (pageblock = page->blocks; pageblock < page->blocks + page->len; pageblock++)
 		if (pageblock->type == FZ_PAGE_BLOCK_TEXT)
@@ -1001,68 +1056,69 @@ fz_bidi_reorder_text_page(fz_context *ctx, fz_text_page *page)
 }
 
 static void
-fz_text_begin_page(fz_context *ctx, fz_device *dev, const fz_rect *mediabox, const fz_matrix *ctm)
+fz_stext_close_device(fz_context *ctx, fz_device *dev)
 {
-	fz_text_device *tdev = (fz_text_device*)dev;
-
-	if (tdev->page->len)
-	{
-		tdev->page->next = fz_new_text_page(ctx);
-		tdev->page = tdev->page->next;
-	}
-
-	tdev->page->mediabox = *mediabox;
-	fz_transform_rect(&tdev->page->mediabox, ctm);
-}
-
-static void
-fz_text_end_page(fz_context *ctx, fz_device *dev)
-{
-	fz_text_device *tdev = (fz_text_device*)dev;
+	fz_stext_device *tdev = (fz_stext_device*)dev;
 
 	add_span_to_soup(ctx, tdev->spans, tdev->cur_span);
 	tdev->cur_span = NULL;
 
 	strain_soup(ctx, tdev);
-	free_span_soup(ctx, tdev->spans);
-	tdev->spans = NULL;
 
 	/* TODO: smart sorting of blocks in reading order */
 	/* TODO: unicode NFC normalization */
 
-	fz_bidi_reorder_text_page(ctx, tdev->page);
+	fz_bidi_reorder_stext_page(ctx, tdev->page);
 }
 
 static void
-fz_text_drop_imp(fz_context *ctx, fz_device *dev)
+fz_stext_drop_device(fz_context *ctx, fz_device *dev)
 {
-	fz_text_device *tdev = (fz_text_device*)dev;
+	fz_stext_device *tdev = (fz_stext_device*)dev;
 	free_span_soup(ctx, tdev->spans);
+	tdev->spans = NULL;
+}
+
+fz_stext_options *
+fz_parse_stext_options(fz_context *ctx, fz_stext_options *opts, const char *string)
+{
+	const char *val;
+
+	memset(opts, 0, sizeof *opts);
+
+	if (fz_has_option(ctx, string, "preserve-ligatures", &val) && fz_option_eq(val, "yes"))
+		opts->flags |= FZ_STEXT_PRESERVE_LIGATURES;
+	if (fz_has_option(ctx, string, "preserve-whitespace", &val) && fz_option_eq(val, "yes"))
+		opts->flags |= FZ_STEXT_PRESERVE_WHITESPACE;
+
+	return opts;
 }
 
 fz_device *
-fz_new_text_device(fz_context *ctx, fz_text_sheet *sheet, fz_text_page *page)
+fz_new_stext_device(fz_context *ctx, fz_stext_sheet *sheet, fz_stext_page *page, const fz_stext_options *opts)
 {
-	fz_text_device *dev = fz_new_device(ctx, sizeof *dev);
+	fz_stext_device *dev = fz_new_device(ctx, sizeof *dev);
 
 	dev->super.hints = FZ_IGNORE_IMAGE | FZ_IGNORE_SHADE;
 
-	dev->super.drop_imp = fz_text_drop_imp;
-	dev->super.begin_page = fz_text_begin_page;
-	dev->super.end_page = fz_text_end_page;
-	dev->super.fill_text = fz_text_fill_text;
-	dev->super.stroke_text = fz_text_stroke_text;
-	dev->super.clip_text = fz_text_clip_text;
-	dev->super.clip_stroke_text = fz_text_clip_stroke_text;
-	dev->super.ignore_text = fz_text_ignore_text;
-	dev->super.fill_image = fz_text_fill_image;
-	dev->super.fill_image_mask = fz_text_fill_image_mask;
+	dev->super.close_device = fz_stext_close_device;
+	dev->super.drop_device = fz_stext_drop_device;
+
+	dev->super.fill_text = fz_stext_fill_text;
+	dev->super.stroke_text = fz_stext_stroke_text;
+	dev->super.clip_text = fz_stext_clip_text;
+	dev->super.clip_stroke_text = fz_stext_clip_stroke_text;
+	dev->super.ignore_text = fz_stext_ignore_text;
+	dev->super.fill_image = fz_stext_fill_image;
+	dev->super.fill_image_mask = fz_stext_fill_image_mask;
 
 	dev->sheet = sheet;
 	dev->page = page;
 	dev->spans = NULL;
 	dev->cur_span = NULL;
 	dev->lastchar = ' ';
+	if (opts)
+		dev->flags = opts->flags;
 
 	return (fz_device*)dev;
 }
