@@ -1,7 +1,7 @@
 #include "mupdf/fitz.h"
 
 void
-fz_write_gproof_file(fz_context *ctx, const char *pdf_file, fz_document *doc, const char *filename, int res,
+fz_save_gproof(fz_context *ctx, const char *pdf_file, fz_document *doc, const char *filename, int res,
 				const char *print_profile, const char *display_profile)
 {
 	int i;
@@ -10,26 +10,26 @@ fz_write_gproof_file(fz_context *ctx, const char *pdf_file, fz_document *doc, co
 	fz_page *page = NULL;
 
 	fz_var(page);
-	
+
 	if (num_pages <= 0)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot write a 0 page GProof skeleton file");
 
-	out = fz_new_output_to_filename(ctx, filename);
+	out = fz_new_output_with_path(ctx, filename, 0);
 
 	fz_try(ctx)
 	{
 		/* File Signature: GPRO */
-		fz_write_int32le(ctx, out, 0x4f525047);
+		fz_write_int32_le(ctx, out, 0x4f525047);
 
 		/* Version = 1 */
 		fz_write_byte(ctx, out, 1);
 		fz_write_byte(ctx, out, 0);
 
 		/* Resolution */
-		fz_write_int32le(ctx, out, res);
+		fz_write_int32_le(ctx, out, res);
 
 		/* Num Pages */
-		fz_write_int32le(ctx, out, num_pages);
+		fz_write_int32_le(ctx, out, num_pages);
 
 		for (i = 0; i < num_pages; i++)
 		{
@@ -41,10 +41,11 @@ fz_write_gproof_file(fz_context *ctx, const char *pdf_file, fz_document *doc, co
 			fz_drop_page(ctx, page);
 			page = NULL;
 
-			w = (int)((rect.x1 - rect.x0) * res / 72.0 + 0.5);
-			h = (int)((rect.y1 - rect.y0) * res / 72.0 + 0.5);
-			fz_write_int32le(ctx, out, w);
-			fz_write_int32le(ctx, out, h);
+			/* Same lack of rounding as gs uses */
+			w = (int)((rect.x1 - rect.x0) * res / 72.0);
+			h = (int)((rect.y1 - rect.y0) * res / 72.0);
+			fz_write_int32_le(ctx, out, w);
+			fz_write_int32_le(ctx, out, h);
 		}
 
 		/* Filenames */
