@@ -36,14 +36,15 @@ struct gstate_s
 	fz_matrix tm;
 };
 
-typedef struct image_entry_s image_entry;
-
-struct image_entry_s
-{
-	char digest[16];
-	int id;
-	pdf_obj *ref;
-};
+/* The image digest information, object reference, as well as indirect reference
+ * ID are all stored in doc->resources->image, and so they are maintained
+ * through the life of the document not just this page level device. As we
+ * encounter images on a page, we will add to the hash table if they are not
+ * already present.  When we have an image on a particular page, the resource
+ * dict will be updated with the proper indirect reference across the document.
+ * We do need to maintain some information as to what image resources we have
+ * already specified for this page which is the purpose of image_indices
+ */
 
 typedef struct alpha_entry_s alpha_entry;
 
@@ -53,13 +54,6 @@ struct alpha_entry_s
 	int stroke;
 };
 
-typedef struct font_entry_s font_entry;
-
-struct font_entry_s
-{
-	fz_font *font;
-};
-
 typedef struct group_entry_s group_entry;
 
 struct group_entry_s
@@ -76,7 +70,6 @@ struct pdf_device_s
 	fz_device super;
 
 	pdf_document *doc;
-	pdf_obj *contents;
 	pdf_obj *resources;
 	fz_buffer *buffer;
 
@@ -91,16 +84,16 @@ struct pdf_device_s
 
 	int num_imgs;
 	int max_imgs;
-	image_entry *images;
+	int *image_indices;
+
+	int num_cid_fonts;
+	int max_cid_fonts;
+	fz_font **cid_fonts;
 
 	int num_alphas;
 	int max_alphas;
 	alpha_entry *alphas;
 
-	int num_fonts;
-	int max_fonts;
-	font_entry *fonts;
-
 	int num_groups;
 	int max_groups;
 	group_entry *groups;
@@ -109,210 +102,8 @@ struct pdf_device_s
 #define CURRENT_GSTATE(pdev) (&(pdev)->gstates[(pdev)->num_gstates-1])
 
 /* Helper functions */
-
-static int
-send_image(fz_context *ctx, pdf_device *pdev, fz_image *image, int mask, int smask)
-{
-	fz_pixmap *pixmap = NULL;
-	pdf_obj *imobj = NULL;
-	pdf_obj *imref = NULL;
-	fz_compressed_buffer *cbuffer = NULL;
-	fz_compression_params *cp = NULL;
-	fz_buffer *buffer = NULL;
-	int i, num;
-	fz_md5 state;
-	unsigned char digest[16];
-	fz_colorspace *colorspace = image->colorspace;
-	pdf_document *doc = pdev->doc;
-
-	/* If we can maintain compression, do so */
-	cbuffer = image->buffer;
-
-	fz_var(pixmap);
-	fz_var(buffer);
-	fz_var(imobj);
-	fz_var(imref);
-
-	fz_try(ctx)
-	{
-		if (cbuffer != NULL && cbuffer->params.type != FZ_IMAGE_PNG && cbuffer->params.type != FZ_IMAGE_TIFF)
-		{
-			buffer = fz_keep_buffer(ctx, cbuffer->buffer);
-			cp = &cbuffer->params;
-		}
-		else
-		{
-			unsigned int size;
-			int n;
-			/* Currently, set to maintain resolution; should we consider
-			 * subsampling here according to desired output res? */
-			pixmap = fz_image_get_pixmap(ctx, image, image->w, image->h);
-			colorspace = pixmap->colorspace; /* May be different to image->colorspace! */
-			n = (pixmap->n == 1 ? 1 : pixmap->n-1);
-			size = image->w * image->h * n;
-			buffer = fz_new_buffer(ctx, size);
-			buffer->len = size;
-			if (pixmap->n == 1)
-			{
-				memcpy(buffer->data, pixmap->samples, size);
-			}
-			else
-			{
-				/* Need to remove the alpha plane */
-				unsigned char *d = buffer->data;
-				unsigned char *s = pixmap->samples;
-				int mod = n;
-				while (size--)
-				{
-					*d++ = *s++;
-					mod--;
-					if (mod == 0)
-						s++, mod = n;
-				}
-			}
-		}
-
-		fz_md5_init(&state);
-		fz_md5_update(&state, buffer->data, buffer->len);
-		fz_md5_final(&state, digest);
-		for(i=0; i < pdev->num_imgs; i++)
-		{
-			if (!memcmp(&digest, pdev->images[i].digest, sizeof(digest)))
-			{
-				num = i;
-				break;
-			}
-		}
-
-		if (i < pdev->num_imgs)
-			break;
-
-		if (pdev->num_imgs == pdev->max_imgs)
-		{
-			int newmax = pdev->max_imgs * 2;
-			if (newmax == 0)
-				newmax = 4;
-			pdev->images = fz_resize_array(ctx, pdev->images, newmax, sizeof(*pdev->images));
-			pdev->max_imgs = newmax;
-		}
-		num = pdev->num_imgs++;
-		memcpy(pdev->images[num].digest,digest,16);
-		pdev->images[num].ref = NULL; /* Will be filled in later */
-
-		imobj = pdf_new_dict(ctx, doc, 3);
-		pdf_dict_put_drop(ctx, imobj, PDF_NAME_Type, PDF_NAME_XObject);
-		pdf_dict_put_drop(ctx, imobj, PDF_NAME_Subtype, PDF_NAME_Image);
-		pdf_dict_put_drop(ctx, imobj, PDF_NAME_Width, pdf_new_int(ctx, doc, image->w));
-		pdf_dict_put_drop(ctx, imobj, PDF_NAME_Height, pdf_new_int(ctx, doc, image->h));
-		if (mask)
-		{}
-		else if (!colorspace || colorspace->n == 1)
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_ColorSpace, PDF_NAME_DeviceGray);
-		else if (colorspace->n == 3)
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_ColorSpace, PDF_NAME_DeviceRGB);
-		else if (colorspace->n == 4)
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_ColorSpace, PDF_NAME_DeviceCMYK);
-		if (!mask)
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_BitsPerComponent, pdf_new_int(ctx, doc, image->bpc));
-		switch (cp ? cp->type : FZ_IMAGE_UNKNOWN)
-		{
-		case FZ_IMAGE_UNKNOWN: /* Unknown also means raw */
-		default:
-			break;
-		case FZ_IMAGE_JPEG:
-			if (cp->u.jpeg.color_transform != -1)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_ColorTransform, pdf_new_int(ctx, doc, cp->u.jpeg.color_transform));
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_Filter, PDF_NAME_DCTDecode);
-			break;
-		case FZ_IMAGE_JPX:
-			if (cp->u.jpx.smask_in_data)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_SMaskInData, pdf_new_int(ctx, doc, cp->u.jpx.smask_in_data));
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_Filter, PDF_NAME_JPXDecode);
-			break;
-		case FZ_IMAGE_FAX:
-			if (cp->u.fax.columns)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_Columns, pdf_new_int(ctx, doc, cp->u.fax.columns));
-			if (cp->u.fax.rows)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_Rows, pdf_new_int(ctx, doc, cp->u.fax.rows));
-			if (cp->u.fax.k)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_K, pdf_new_int(ctx, doc, cp->u.fax.k));
-			if (cp->u.fax.end_of_line)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_EndOfLine, pdf_new_int(ctx, doc, cp->u.fax.end_of_line));
-			if (cp->u.fax.encoded_byte_align)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_EncodedByteAlign, pdf_new_int(ctx, doc, cp->u.fax.encoded_byte_align));
-			if (cp->u.fax.end_of_block)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_EndOfBlock, pdf_new_int(ctx, doc, cp->u.fax.end_of_block));
-			if (cp->u.fax.black_is_1)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_BlackIs1, pdf_new_int(ctx, doc, cp->u.fax.black_is_1));
-			if (cp->u.fax.damaged_rows_before_error)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_DamagedRowsBeforeError, pdf_new_int(ctx, doc, cp->u.fax.damaged_rows_before_error));
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_Filter, PDF_NAME_CCITTFaxDecode);
-			break;
-		case FZ_IMAGE_JBIG2:
-			/* FIXME - jbig2globals */
-			cp->type = FZ_IMAGE_UNKNOWN;
-			break;
-		case FZ_IMAGE_FLATE:
-			if (cp->u.flate.columns)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_Columns, pdf_new_int(ctx, doc, cp->u.flate.columns));
-			if (cp->u.flate.colors)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_Colors, pdf_new_int(ctx, doc, cp->u.flate.colors));
-			if (cp->u.flate.predictor)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_Predictor, pdf_new_int(ctx, doc, cp->u.flate.predictor));
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_Filter, PDF_NAME_FlateDecode);
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_BitsPerComponent, pdf_new_int(ctx, doc, image->bpc));
-			break;
-		case FZ_IMAGE_LZW:
-			if (cp->u.lzw.columns)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_Columns, pdf_new_int(ctx, doc, cp->u.lzw.columns));
-			if (cp->u.lzw.colors)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_Colors, pdf_new_int(ctx, doc, cp->u.lzw.colors));
-			if (cp->u.lzw.predictor)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_Predictor, pdf_new_int(ctx, doc, cp->u.lzw.predictor));
-			if (cp->u.lzw.early_change)
-				pdf_dict_put_drop(ctx, imobj, PDF_NAME_EarlyChange, pdf_new_int(ctx, doc, cp->u.lzw.early_change));
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_Filter, PDF_NAME_LZWDecode);
-			break;
-		case FZ_IMAGE_RLD:
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_Filter, PDF_NAME_RunLengthDecode);
-			break;
-		}
-		if (mask)
-		{
-			pdf_dict_put_drop(ctx, imobj, PDF_NAME_ImageMask, pdf_new_bool(ctx, doc, 1));
-		}
-		if (image->mask)
-		{
-			int smasknum = send_image(ctx, pdev, image->mask, 0, 1);
-			pdf_dict_put(ctx, imobj, PDF_NAME_SMask, pdev->images[smasknum].ref);
-		}
-
-		imref = pdf_new_ref(ctx, doc, imobj);
-		pdf_update_stream(ctx, doc, imref, buffer, 1);
-
-		{
-			char text[32];
-			snprintf(text, sizeof(text), "XObject/Img%d", num);
-			pdf_dict_putp(ctx, pdev->resources, text, imref);
-		}
-		pdev->images[num].ref = imref;
-	}
-	fz_always(ctx)
-	{
-		fz_drop_buffer(ctx, buffer);
-		pdf_drop_obj(ctx, imobj);
-		fz_drop_pixmap(ctx, pixmap);
-	}
-	fz_catch(ctx)
-	{
-		pdf_drop_obj(ctx, imref);
-		fz_rethrow(ctx);
-	}
-	return num;
-}
-
 static void
-pdf_dev_stroke_state(fz_context *ctx, pdf_device *pdev, fz_stroke_state *stroke_state)
+pdf_dev_stroke_state(fz_context *ctx, pdf_device *pdev, const fz_stroke_state *stroke_state)
 {
 	gstate *gs = CURRENT_GSTATE(pdev);
 
@@ -398,7 +189,7 @@ pdf_dev_path_close(fz_context *ctx, void *arg)
 	fz_buffer_printf(ctx, buf, "h\n");
 }
 
-static const fz_path_processor pdf_dev_path_proc =
+static const fz_path_walker pdf_dev_path_proc =
 {
 	pdf_dev_path_moveto,
 	pdf_dev_path_lineto,
@@ -407,11 +198,11 @@ static const fz_path_processor pdf_dev_path_proc =
 };
 
 static void
-pdf_dev_path(fz_context *ctx, pdf_device *pdev, fz_path *path)
+pdf_dev_path(fz_context *ctx, pdf_device *pdev, const fz_path *path)
 {
 	gstate *gs = CURRENT_GSTATE(pdev);
 
-	fz_process_path(ctx, &pdf_dev_path_proc, (void *)gs->buf, path);
+	fz_walk_path(ctx, path, &pdf_dev_path_proc, (void *)gs->buf);
 }
 
 static void
@@ -424,12 +215,12 @@ pdf_dev_ctm(fz_context *ctx, pdf_device *pdev, const fz_matrix *ctm)
 		return;
 	fz_invert_matrix(&inverse, &gs->ctm);
 	fz_concat(&inverse, ctm, &inverse);
-	memcpy(&gs->ctm, ctm, sizeof(*ctm));
-	fz_buffer_printf(ctx, gs->buf, "%f %f %f %f %f %f cm\n", inverse.a, inverse.b, inverse.c, inverse.d, inverse.e, inverse.f);
+	gs->ctm = *ctm;
+	fz_buffer_printf(ctx, gs->buf, "%M cm\n", &inverse);
 }
 
 static void
-pdf_dev_color(fz_context *ctx, pdf_device *pdev, fz_colorspace *colorspace, float *color, int stroke)
+pdf_dev_color(fz_context *ctx, pdf_device *pdev, fz_colorspace *colorspace, const float *color, int stroke)
 {
 	int diff = 0;
 	int i;
@@ -447,9 +238,10 @@ pdf_dev_color(fz_context *ctx, pdf_device *pdev, fz_colorspace *colorspace, floa
 	if (cspace == 0)
 	{
 		/* If it's an unknown colorspace, fallback to rgb */
-		colorspace->to_rgb(ctx, colorspace, color, rgb);
+		fz_convert_color(ctx, fz_device_rgb(ctx), rgb, colorspace, color);
 		color = rgb;
 		colorspace = fz_device_rgb(ctx);
+		cspace = 3;
 	}
 
 	if (gs->colorspace[stroke] != colorspace)
@@ -458,7 +250,7 @@ pdf_dev_color(fz_context *ctx, pdf_device *pdev, fz_colorspace *colorspace, floa
 		diff = 1;
 	}
 
-	for (i=0; i < colorspace->n; i++)
+	for (i=0; i < cspace; i++)
 		if (gs->color[stroke][i] != color[i])
 		{
 			gs->color[stroke][i] = color[i];
@@ -531,8 +323,8 @@ pdf_dev_alpha(fz_context *ctx, pdf_device *pdev, float alpha, int stroke)
 		{
 			char text[32];
 			pdf_dict_put_drop(ctx, o, (stroke ? PDF_NAME_CA : PDF_NAME_ca), pdf_new_real(ctx, doc, alpha));
-			ref = pdf_new_ref(ctx, doc, o);
-			snprintf(text, sizeof(text), "ExtGState/Alp%d", i);
+			ref = pdf_add_object(ctx, doc, o);
+			fz_snprintf(text, sizeof(text), "ExtGState/Alp%d", i);
 			pdf_dict_putp(ctx, pdev->resources, text, ref);
 		}
 		fz_always(ctx)
@@ -549,78 +341,59 @@ pdf_dev_alpha(fz_context *ctx, pdf_device *pdev, float alpha, int stroke)
 	fz_buffer_printf(ctx, gs->buf, "/Alp%d gs\n", i);
 }
 
-static void
-pdf_dev_font(fz_context *ctx, pdf_device *pdev, fz_font *font, float size)
+static int
+pdf_dev_add_font_res(fz_context *ctx, pdf_device *pdev, fz_font *font)
 {
-	int i;
-	pdf_document *doc = pdev->doc;
-	gstate *gs = CURRENT_GSTATE(pdev);
+	pdf_obj *fres;
+	char text[32];
+	int k;
+	int num;
 
-	/* If the font is unchanged, nothing to do */
-	if (gs->font >= 0 && pdev->fonts[gs->font].font == font)
-		return;
+	/* Check if we already had this one */
+	for (k = 0; k < pdev->num_cid_fonts; k++)
+		if (pdev->cid_fonts[k] == font)
+			return k;
 
-	if (font->ft_buffer != NULL || font->ft_substitute)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "pdf device supports only base 14 fonts currently");
+	/* This will add it to the xref if needed */
+	fres = pdf_add_cid_font(ctx, pdev->doc, font);
 
-	/* Have we sent such a font before? */
-	for (i = 0; i < pdev->num_fonts; i++)
-		if (pdev->fonts[i].font == font)
-			break;
+	/* Not there so add to resources */
+	fz_snprintf(text, sizeof(text), "Font/F%d", pdev->num_cid_fonts);
+	pdf_dict_putp_drop(ctx, pdev->resources, text, fres);
 
-	if (i == pdev->num_fonts)
+	/* And add index to our list for this page */
+	if (pdev->num_cid_fonts == pdev->max_cid_fonts)
 	{
-		pdf_obj *o;
-		pdf_obj *ref = NULL;
-
-		fz_var(ref);
-
-		/* No. Need to make a new one */
-		if (pdev->num_fonts == pdev->max_fonts)
-		{
-			int newmax = pdev->max_fonts * 2;
-			if (newmax == 0)
-				newmax = 4;
-			pdev->fonts = fz_resize_array(ctx, pdev->fonts, newmax, sizeof(*pdev->fonts));
-			pdev->max_fonts = newmax;
-		}
-		pdev->fonts[i].font = fz_keep_font(ctx, font);
-
-		o = pdf_new_dict(ctx, doc, 3);
-		fz_try(ctx)
-		{
-			char text[32];
-			pdf_dict_put_drop(ctx, o, PDF_NAME_Type, PDF_NAME_Font);
-			pdf_dict_put_drop(ctx, o, PDF_NAME_Subtype, PDF_NAME_Type1);
-			pdf_dict_put_drop(ctx, o, PDF_NAME_BaseFont, pdf_new_name(ctx, doc, font->name));
-			pdf_dict_put_drop(ctx, o, PDF_NAME_Encoding, PDF_NAME_WinAnsiEncoding);
-			ref = pdf_new_ref(ctx, doc, o);
-			snprintf(text, sizeof(text), "Font/F%d", i);
-			pdf_dict_putp(ctx, pdev->resources, text, ref);
-		}
-		fz_always(ctx)
-		{
-			pdf_drop_obj(ctx, o);
-			pdf_drop_obj(ctx, ref);
-		}
-		fz_catch(ctx)
-		{
-			fz_rethrow(ctx);
-		}
-		pdev->num_fonts++;
+		int newmax = pdev->max_cid_fonts * 2;
+		if (newmax == 0)
+			newmax = 4;
+		pdev->cid_fonts = fz_resize_array(ctx, pdev->cid_fonts, newmax, sizeof(*pdev->cid_fonts));
+		pdev->max_cid_fonts = newmax;
 	}
-	fz_buffer_printf(ctx, gs->buf, "/F%d %f Tf\n", i, size);
+	num = pdev->num_cid_fonts++;
+	pdev->cid_fonts[num] = fz_keep_font(ctx, font);
+	return num;
 }
 
 static void
-pdf_dev_tm(fz_context *ctx, pdf_device *pdev, const fz_matrix *tm)
+pdf_dev_font(fz_context *ctx, pdf_device *pdev, fz_font *font)
 {
 	gstate *gs = CURRENT_GSTATE(pdev);
 
-	if (memcmp(&gs->tm, tm, sizeof(*tm)) == 0)
+	/* If the font is unchanged, nothing to do */
+	if (gs->font >= 0 && pdev->cid_fonts[gs->font] == font)
 		return;
-	fz_buffer_printf(ctx, gs->buf, "%f %f %f %f %f %f Tm\n", tm->a, tm->b, tm->c, tm->d, tm->e, tm->f);
-	gs->tm = *tm;
+
+	if (fz_font_t3_procs(ctx, font))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "pdf device does not support type 3 fonts");
+	if (fz_font_flags(font)->ft_substitute)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "pdf device does not support substitute fonts");
+	if (!pdf_font_writing_supported(font))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "pdf device does not support font types found in this file");
+
+	gs->font = pdf_dev_add_font_res(ctx, pdev, font);
+
+	fz_buffer_printf(ctx, gs->buf, "/F%d 1 Tf\n", gs->font);
 }
 
 static void
@@ -667,70 +440,47 @@ pdf_dev_pop(fz_context *ctx, pdf_device *pdev)
 }
 
 static void
-pdf_dev_text(fz_context *ctx, pdf_device *pdev, fz_text *text, float size)
+pdf_dev_text_span(fz_context *ctx, pdf_device *pdev, fz_text_span *span)
 {
-	int mask = FT_LOAD_NO_SCALE | FT_LOAD_IGNORE_TRANSFORM;
-	int i;
-	fz_matrix trm;
-	fz_matrix inverse;
 	gstate *gs = CURRENT_GSTATE(pdev);
-	fz_matrix trunc_trm;
+	fz_matrix tm;
+	float dx, dy, adv;
+	int i;
 
-	trm = gs->tm;
-	trunc_trm.a = trm.a;
-	trunc_trm.b = trm.b;
-	trunc_trm.c = trm.c;
-	trunc_trm.d = trm.d;
-	trunc_trm.e = 0;
-	trunc_trm.f = 0;
-	fz_invert_matrix(&inverse, &trunc_trm);
+	if (span->len == 0)
+		return;
 
-	i = 0;
-	while (i < text->len)
+	tm = span->trm;
+	tm.e = span->items[0].x;
+	tm.f = span->items[0].y;
+
+	fz_buffer_printf(ctx, gs->buf, "%f %f %f %f %f %f Tm\n", tm.a, tm.b, tm.c, tm.d, tm.e, tm.f);
+
+	for (i = 0; i < span->len; ++i)
 	{
-		fz_text_item *it = &text->items[i];
-		fz_point delta;
-		float x;
-		int j;
+		fz_text_item *it = &span->items[i];
+		if (it->gid < 0)
+			continue;
 
-		delta.x = it->x - trm.e;
-		delta.y = it->y - trm.f;
-		fz_transform_point(&delta, &inverse);
-		if (delta.x != 0 || delta.y != 0)
-		{
-			fz_buffer_printf(ctx, gs->buf, "%f %f Td ", delta.x, delta.y);
-			trm.e = it->x;
-			trm.f = it->y;
-		}
+		dx = it->x - tm.e;
+		dy = it->y - tm.f;
+		tm.e = it->x;
+		tm.f = it->y;
 
-		j = i+1;
-		if (text->font->ft_face)
-		{
-			/* Find prefix of text for which the advance of each character accounts
-			 * for the position offset */
-			x = it->x;
-			while (j < text->len)
-			{
-				FT_Fixed adv;
-				FT_Get_Advance(text->font->ft_face, text->items[j-1].gid, mask, &adv);
-				x += (float)adv * size /((FT_Face)text->font->ft_face)->units_per_EM;
-				if (fabs(x - text->items[j].x) > ALLOWED_TEXT_POS_ERROR || fabs(it->y - text->items[j].y) > ALLOWED_TEXT_POS_ERROR)
-					break;
-				j++;
-			}
-		}
+		if (fabsf(dx) > 0 || fabsf(dy) > 0)
+			fz_buffer_printf(ctx, gs->buf, "%f %f %f %f %f %f Tm\n", tm.a, tm.b, tm.c, tm.d, tm.e, tm.f);
 
-		fz_buffer_printf(ctx, gs->buf, "<");
-		for (/* i from its current value */; i < j; i++)
-		{
-			/* FIXME: should use it->gid, rather than it->ucs, and convert
-			* to the correct encoding */
-			fz_buffer_printf(ctx, gs->buf, "%02x", text->items[i].ucs);
-		}
-		fz_buffer_printf(ctx, gs->buf, "> Tj\n");
+		if (fz_font_t3_procs(ctx, span->font))
+			fz_buffer_printf(ctx, gs->buf, "<%02x> Tj\n", it->gid);
+		else
+			fz_buffer_printf(ctx, gs->buf, "<%04x> Tj\n", it->gid);
+
+		adv = fz_advance_glyph(ctx, span->font, it->gid, span->wmode);
+		if (span->wmode == 0)
+			fz_pre_translate(&tm, adv, 0);
+		else
+			fz_pre_translate(&tm, 0, adv);
 	}
-	gs->tm.e = trm.e;
-	gs->tm.f = trm.f;
 }
 
 static void
@@ -760,7 +510,6 @@ pdf_dev_begin_text(fz_context *ctx, pdf_device *pdev, const fz_matrix *tm, int t
 		gs->tm.f = 0;
 		pdev->in_text = 1;
 	}
-	pdf_dev_tm(ctx, pdev, tm);
 }
 
 static void
@@ -816,19 +565,20 @@ pdf_dev_new_form(fz_context *ctx, pdf_obj **form_ref, pdf_device *pdev, const fz
 		group = pdf_new_dict(ctx, doc, 5);
 		fz_try(ctx)
 		{
+			int n = fz_colorspace_n(ctx, colorspace);
 			pdf_dict_put_drop(ctx, group, PDF_NAME_Type, PDF_NAME_Group);
 			pdf_dict_put_drop(ctx, group, PDF_NAME_S, PDF_NAME_Transparency);
 			pdf_dict_put_drop(ctx, group, PDF_NAME_K, pdf_new_bool(ctx, doc, knockout));
 			pdf_dict_put_drop(ctx, group, PDF_NAME_I, pdf_new_bool(ctx, doc, isolated));
-			if (!colorspace)
+			if (n == 0)
 			{}
-			else if (colorspace->n == 1)
+			if (n == 1)
 				pdf_dict_put_drop(ctx, group, PDF_NAME_CS, PDF_NAME_DeviceGray);
-			else if (colorspace->n == 4)
+			else if (n == 4)
 				pdf_dict_put_drop(ctx, group, PDF_NAME_CS, PDF_NAME_DeviceCMYK);
 			else
 				pdf_dict_put_drop(ctx, group, PDF_NAME_CS, PDF_NAME_DeviceRGB);
-			group_ref = pdev->groups[num].ref = pdf_new_ref(ctx, doc, group);
+			group_ref = pdev->groups[num].ref = pdf_add_object(ctx, doc, group);
 		}
 		fz_always(ctx)
 		{
@@ -849,11 +599,14 @@ pdf_dev_new_form(fz_context *ctx, pdf_obj **form_ref, pdf_device *pdev, const fz
 		pdf_dict_put(ctx, form, PDF_NAME_Group, group_ref);
 		pdf_dict_put_drop(ctx, form, PDF_NAME_FormType, pdf_new_int(ctx, doc, 1));
 		pdf_dict_put_drop(ctx, form, PDF_NAME_BBox, pdf_new_rect(ctx, doc, bbox));
-		*form_ref = pdf_new_ref(ctx, doc, form);
+		*form_ref = pdf_add_object(ctx, doc, form);
+	}
+	fz_always(ctx)
+	{
+		pdf_drop_obj(ctx, form);
 	}
 	fz_catch(ctx)
 	{
-		pdf_drop_obj(ctx, form);
 		fz_rethrow(ctx);
 	}
 
@@ -861,7 +614,7 @@ pdf_dev_new_form(fz_context *ctx, pdf_obj **form_ref, pdf_device *pdev, const fz
 	{
 		char text[32];
 		num = pdev->num_forms++;
-		snprintf(text, sizeof(text), "XObject/Fm%d", num);
+		fz_snprintf(text, sizeof(text), "XObject/Fm%d", num);
 		pdf_dict_putp(ctx, pdev->resources, text, *form_ref);
 	}
 
@@ -871,8 +624,8 @@ pdf_dev_new_form(fz_context *ctx, pdf_obj **form_ref, pdf_device *pdev, const fz
 /* Entry points */
 
 static void
-pdf_dev_fill_path(fz_context *ctx, fz_device *dev, fz_path *path, int even_odd, const fz_matrix *ctm,
-	fz_colorspace *colorspace, float *color, float alpha)
+pdf_dev_fill_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even_odd, const fz_matrix *ctm,
+	fz_colorspace *colorspace, const float *color, float alpha)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	gstate *gs = CURRENT_GSTATE(pdev);
@@ -886,8 +639,8 @@ pdf_dev_fill_path(fz_context *ctx, fz_device *dev, fz_path *path, int even_odd,
 }
 
 static void
-pdf_dev_stroke_path(fz_context *ctx, fz_device *dev, fz_path *path, fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, float *color, float alpha)
+pdf_dev_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const fz_stroke_state *stroke, const fz_matrix *ctm,
+	fz_colorspace *colorspace, const float *color, float alpha)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	gstate *gs = CURRENT_GSTATE(pdev);
@@ -902,7 +655,7 @@ pdf_dev_stroke_path(fz_context *ctx, fz_device *dev, fz_path *path, fz_stroke_st
 }
 
 static void
-pdf_dev_clip_path(fz_context *ctx, fz_device *dev, fz_path *path, const fz_rect *rect, int even_odd, const fz_matrix *ctm)
+pdf_dev_clip_path(fz_context *ctx, fz_device *dev, const fz_path *path, int even_odd, const fz_matrix *ctm, const fz_rect *scissor)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	gstate *gs;
@@ -916,7 +669,7 @@ pdf_dev_clip_path(fz_context *ctx, fz_device *dev, fz_path *path, const fz_rect
 }
 
 static void
-pdf_dev_clip_stroke_path(fz_context *ctx, fz_device *dev, fz_path *path, const fz_rect *rect, fz_stroke_state *stroke, const fz_matrix *ctm)
+pdf_dev_clip_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const fz_stroke_state *stroke, const fz_matrix *ctm, const fz_rect *scissor)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	gstate *gs;
@@ -934,102 +687,142 @@ pdf_dev_clip_stroke_path(fz_context *ctx, fz_device *dev, fz_path *path, const f
 }
 
 static void
-pdf_dev_fill_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm,
-	fz_colorspace *colorspace, float *color, float alpha)
+pdf_dev_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm,
+		fz_colorspace *colorspace, const float *color, float alpha)
 {
 	pdf_device *pdev = (pdf_device*)dev;
-	fz_matrix trm = text->trm;
-	float size = fz_matrix_expansion(&trm);
+	fz_text_span *span;
 
-	fz_pre_scale(&trm, 1/size, 1/size);
-
-	pdf_dev_begin_text(ctx, pdev, &trm, 0);
-	pdf_dev_font(ctx, pdev, text->font, size);
-	pdf_dev_ctm(ctx, pdev, ctm);
-	pdf_dev_alpha(ctx, pdev, alpha, 0);
-	pdf_dev_color(ctx, pdev, colorspace, color, 0);
-	pdf_dev_text(ctx, pdev, text, size);
+	for (span = text->head; span; span = span->next)
+	{
+		fz_matrix trm = span->trm;
+		pdf_dev_begin_text(ctx, pdev, &trm, 0);
+		pdf_dev_font(ctx, pdev, span->font);
+		pdf_dev_ctm(ctx, pdev, ctm);
+		pdf_dev_alpha(ctx, pdev, alpha, 0);
+		pdf_dev_color(ctx, pdev, colorspace, color, 0);
+		pdf_dev_text_span(ctx, pdev, span);
+	}
 }
 
 static void
-pdf_dev_stroke_text(fz_context *ctx, fz_device *dev, fz_text *text, fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, float *color, float alpha)
+pdf_dev_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm,
+		fz_colorspace *colorspace, const float *color, float alpha)
 {
 	pdf_device *pdev = (pdf_device*)dev;
-	fz_matrix trm = text->trm;
-	float size = fz_matrix_expansion(&trm);
+	fz_text_span *span;
 
-	fz_pre_scale(&trm, 1/size, 1/size);
-
-	pdf_dev_begin_text(ctx, pdev, &text->trm, 1);
-	pdf_dev_font(ctx, pdev, text->font, 1);
-	pdf_dev_ctm(ctx, pdev, ctm);
-	pdf_dev_alpha(ctx, pdev, alpha, 1);
-	pdf_dev_color(ctx, pdev, colorspace, color, 1);
-	pdf_dev_text(ctx, pdev, text, size);
+	for (span = text->head; span; span = span->next)
+	{
+		pdf_dev_begin_text(ctx, pdev, &span->trm, 1);
+		pdf_dev_font(ctx, pdev, span->font);
+		pdf_dev_ctm(ctx, pdev, ctm);
+		pdf_dev_alpha(ctx, pdev, alpha, 1);
+		pdf_dev_color(ctx, pdev, colorspace, color, 1);
+		pdf_dev_text_span(ctx, pdev, span);
+	}
 }
 
 static void
-pdf_dev_clip_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm, int accumulate)
+pdf_dev_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm, const fz_rect *scissor)
 {
 	pdf_device *pdev = (pdf_device*)dev;
-	fz_matrix trm = text->trm;
-	float size = fz_matrix_expansion(&trm);
-
-	fz_pre_scale(&trm, 1/size, 1/size);
-
-	pdf_dev_begin_text(ctx, pdev, &text->trm, 0);
-	pdf_dev_ctm(ctx, pdev, ctm);
-	pdf_dev_font(ctx, pdev, text->font, 7);
-	pdf_dev_text(ctx, pdev, text, size);
+	fz_text_span *span;
+	for (span = text->head; span; span = span->next)
+	{
+		pdf_dev_begin_text(ctx, pdev, &span->trm, 0);
+		pdf_dev_ctm(ctx, pdev, ctm);
+		pdf_dev_font(ctx, pdev, span->font);
+		pdf_dev_text_span(ctx, pdev, span);
+	}
 }
 
 static void
-pdf_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, fz_text *text, fz_stroke_state *stroke, const fz_matrix *ctm)
+pdf_dev_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm, const fz_rect *scissor)
 {
 	pdf_device *pdev = (pdf_device*)dev;
-	fz_matrix trm = text->trm;
-	float size = fz_matrix_expansion(&trm);
-
-	fz_pre_scale(&trm, 1/size, 1/size);
-
-	pdf_dev_begin_text(ctx, pdev, &text->trm, 0);
-	pdf_dev_font(ctx, pdev, text->font, 5);
-	pdf_dev_ctm(ctx, pdev, ctm);
-	pdf_dev_text(ctx, pdev, text, size);
+	fz_text_span *span;
+	for (span = text->head; span; span = span->next)
+	{
+		pdf_dev_begin_text(ctx, pdev, &span->trm, 0);
+		pdf_dev_font(ctx, pdev, span->font);
+		pdf_dev_ctm(ctx, pdev, ctm);
+		pdf_dev_text_span(ctx, pdev, span);
+	}
 }
 
 static void
-pdf_dev_ignore_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm)
+pdf_dev_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm)
 {
 	pdf_device *pdev = (pdf_device*)dev;
-	fz_matrix trm = text->trm;
-	float size = fz_matrix_expansion(&trm);
+	fz_text_span *span;
+	for (span = text->head; span; span = span->next)
+	{
+		pdf_dev_begin_text(ctx, pdev, &span->trm, 0);
+		pdf_dev_ctm(ctx, pdev, ctm);
+		pdf_dev_font(ctx, pdev, span->font);
+		pdf_dev_text_span(ctx, pdev, span);
+	}
+}
 
-	fz_pre_scale(&trm, 1/size, 1/size);
+static void
+pdf_dev_add_image_res(fz_context *ctx, fz_device *dev, pdf_obj *im_res)
+{
+	char text[32];
+	pdf_device *pdev = (pdf_device*)dev;
+	int k;
+	int num;
 
-	pdf_dev_begin_text(ctx, pdev, &text->trm, 0);
-	pdf_dev_ctm(ctx, pdev, ctm);
-	pdf_dev_font(ctx, pdev, text->font, 3);
-	pdf_dev_text(ctx, pdev, text, size);
+	/* Check if we already had this one */
+	for (k = 0; k < pdev->num_imgs; k++)
+	{
+		if (pdev->image_indices[k] == pdf_to_num(ctx, im_res))
+			return;
+	}
+
+	/* Not there so add to resources */
+	fz_snprintf(text, sizeof(text), "XObject/Img%d", pdf_to_num(ctx, im_res));
+	pdf_dict_putp(ctx, pdev->resources, text, im_res);
+
+	/* And add index to our list for this page */
+	if (pdev->num_imgs == pdev->max_imgs)
+	{
+		int newmax = pdev->max_imgs * 2;
+		if (newmax == 0)
+			newmax = 4;
+		pdev->image_indices = fz_resize_array(ctx, pdev->image_indices, newmax, sizeof(*pdev->image_indices));
+		pdev->max_imgs = newmax;
+	}
+	num = pdev->num_imgs++;
+	pdev->image_indices[num] = pdf_to_num(ctx, im_res);
 }
 
 static void
 pdf_dev_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm, float alpha)
 {
 	pdf_device *pdev = (pdf_device*)dev;
-	int num;
+	pdf_obj *im_res;
 	gstate *gs = CURRENT_GSTATE(pdev);
 	fz_matrix local_ctm = *ctm;
 
 	pdf_dev_end_text(ctx, pdev);
-	num = send_image(ctx, pdev, image, 0, 0);
+	im_res = pdf_add_image(ctx, pdev->doc, image, 0);
+	if (im_res == NULL)
+	{
+		fz_warn(ctx, "pdf_add_image: problem adding image resource");
+		return;
+	}
 	pdf_dev_alpha(ctx, pdev, alpha, 0);
+
 	/* PDF images are upside down, so fiddle the ctm */
 	fz_pre_scale(&local_ctm, 1, -1);
 	fz_pre_translate(&local_ctm, 0, -1);
 	pdf_dev_ctm(ctx, pdev, &local_ctm);
-	fz_buffer_printf(ctx, gs->buf, "/Img%d Do\n", num);
+	fz_buffer_printf(ctx, gs->buf, "/Img%d Do\n", pdf_to_num(ctx, im_res));
+
+	/* Possibly add to page resources */
+	pdf_dev_add_image_res(ctx, dev, im_res);
+	pdf_drop_obj(ctx, im_res);
 }
 
 static void
@@ -1043,27 +836,37 @@ pdf_dev_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, const fz_ma
 
 static void
 pdf_dev_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm,
-fz_colorspace *colorspace, float *color, float alpha)
+		fz_colorspace *colorspace, const float *color, float alpha)
 {
 	pdf_device *pdev = (pdf_device*)dev;
+	pdf_obj *im_res = NULL;
 	gstate *gs = CURRENT_GSTATE(pdev);
-	int num;
 	fz_matrix local_ctm = *ctm;
 
 	pdf_dev_end_text(ctx, pdev);
-	num = send_image(ctx, pdev, image, 1, 0);
+	im_res = pdf_add_image(ctx, pdev->doc, image, 1);
+	if (im_res == NULL)
+	{
+		fz_warn(ctx, "pdf_add_image: problem adding image resource");
+		return;
+	}
 	fz_buffer_printf(ctx, gs->buf, "q\n");
 	pdf_dev_alpha(ctx, pdev, alpha, 0);
 	pdf_dev_color(ctx, pdev, colorspace, color, 0);
+
 	/* PDF images are upside down, so fiddle the ctm */
 	fz_pre_scale(&local_ctm, 1, -1);
 	fz_pre_translate(&local_ctm, 0, -1);
 	pdf_dev_ctm(ctx, pdev, &local_ctm);
-	fz_buffer_printf(ctx, gs->buf, "/Img%d Do Q\n", num);
+	fz_buffer_printf(ctx, gs->buf, "/Img%d Do Q\n", pdf_to_num(ctx, im_res));
+
+	/* Possibly add to page resources */
+	pdf_dev_add_image_res(ctx, dev, im_res);
+	pdf_drop_obj(ctx, im_res);
 }
 
 static void
-pdf_dev_clip_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_rect *rect, const fz_matrix *ctm)
+pdf_dev_clip_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm, const fz_rect *scissor)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 
@@ -1083,7 +886,7 @@ pdf_dev_pop_clip(fz_context *ctx, fz_device *dev)
 }
 
 static void
-pdf_dev_begin_mask(fz_context *ctx, fz_device *dev, const fz_rect *bbox, int luminosity, fz_colorspace *colorspace, float *color)
+pdf_dev_begin_mask(fz_context *ctx, fz_device *dev, const fz_rect *bbox, int luminosity, fz_colorspace *colorspace, const float *color)
 {
 	pdf_device *pdev = (pdf_device*)dev;
 	pdf_document *doc = pdev->doc;
@@ -1106,24 +909,25 @@ pdf_dev_begin_mask(fz_context *ctx, fz_device *dev, const fz_rect *bbox, int lum
 
 	fz_try(ctx)
 	{
+		int n = fz_colorspace_n(ctx, colorspace);
 		smask = pdf_new_dict(ctx, doc, 4);
 		pdf_dict_put_drop(ctx, smask, PDF_NAME_Type, PDF_NAME_Mask);
 		pdf_dict_put_drop(ctx, smask, PDF_NAME_S, (luminosity ? PDF_NAME_Luminosity : PDF_NAME_Alpha));
 		pdf_dict_put(ctx, smask, PDF_NAME_G, form_ref);
-		color_obj = pdf_new_array(ctx, doc, colorspace->n);
-		for (i = 0; i < colorspace->n; i++)
-			pdf_array_push(ctx, color_obj, pdf_new_real(ctx, doc, color[i]));
+		color_obj = pdf_new_array(ctx, doc, n);
+		for (i = 0; i < n; i++)
+			pdf_array_push_drop(ctx, color_obj, pdf_new_real(ctx, doc, color[i]));
 		pdf_dict_put_drop(ctx, smask, PDF_NAME_BC, color_obj);
 		color_obj = NULL;
 
 		egs = pdf_new_dict(ctx, doc, 5);
 		pdf_dict_put_drop(ctx, egs, PDF_NAME_Type, PDF_NAME_ExtGState);
-		pdf_dict_put_drop(ctx, egs, PDF_NAME_SMask, pdf_new_ref(ctx, doc, smask));
-		egs_ref = pdf_new_ref(ctx, doc, egs);
+		pdf_dict_put_drop(ctx, egs, PDF_NAME_SMask, pdf_add_object(ctx, doc, smask));
+		egs_ref = pdf_add_object(ctx, doc, egs);
 
 		{
 			char text[32];
-			snprintf(text, sizeof(text), "ExtGState/SM%d", pdev->num_smasks++);
+			fz_snprintf(text, sizeof(text), "ExtGState/SM%d", pdev->num_smasks++);
 			pdf_dict_putp(ctx, pdev->resources, text, egs_ref);
 			pdf_drop_obj(ctx, egs_ref);
 		}
@@ -1183,7 +987,7 @@ pdf_dev_begin_group(fz_context *ctx, fz_device *dev, const fz_rect *bbox, int is
 	{
 		char text[32];
 		pdf_obj *obj;
-		snprintf(text, sizeof(text), "ExtGState/BlendMode%d", blendmode);
+		fz_snprintf(text, sizeof(text), "ExtGState/BlendMode%d", blendmode);
 		obj = pdf_dict_getp(ctx, pdev->resources, text);
 		if (obj == NULL)
 		{
@@ -1240,52 +1044,42 @@ pdf_dev_end_tile(fz_context *ctx, fz_device *dev)
 }
 
 static void
-pdf_dev_drop_imp(fz_context *ctx, fz_device *dev)
+pdf_dev_close_device(fz_context *ctx, fz_device *dev)
+{
+	pdf_device *pdev = (pdf_device*)dev;
+	pdf_dev_end_text(ctx, pdev);
+}
+
+static void
+pdf_dev_drop_device(fz_context *ctx, fz_device *dev)
 {
 	pdf_device *pdev = (pdf_device*)dev;
-	pdf_document *doc = pdev->doc;
 	int i;
 
-	pdf_dev_end_text(ctx, pdev);
-
 	for (i = pdev->num_gstates-1; i >= 0; i--)
-	{
 		fz_drop_stroke_state(ctx, pdev->gstates[i].stroke_state);
-	}
 
-	for (i = pdev->num_fonts-1; i >= 0; i--)
-	{
-		fz_drop_font(ctx, pdev->fonts[i].font);
-	}
+	for (i = pdev->num_cid_fonts-1; i >= 0; i--)
+		fz_drop_font(ctx, pdev->cid_fonts[i]);
 
-	for (i = pdev->num_imgs-1; i >= 0; i--)
-	{
-		pdf_drop_obj(ctx, pdev->images[i].ref);
-	}
-
-	if (pdev->contents)
-	{
-		pdf_update_stream(ctx, doc, pdev->contents, pdev->gstates[0].buf, 0);
-		pdf_drop_obj(ctx, pdev->contents);
-	}
-
-	if (pdev->buffer != pdev->gstates[0].buf)
-	{
-		fz_drop_buffer(ctx, pdev->gstates[0].buf);
-	}
+	for (i = pdev->num_groups - 1; i >= 0; i--)
+		pdf_drop_obj(ctx, pdev->groups[i].ref);
 
+	fz_drop_buffer(ctx, pdev->buffer);
 	pdf_drop_obj(ctx, pdev->resources);
-
-	fz_free(ctx, pdev->images);
+	fz_free(ctx, pdev->cid_fonts);
+	fz_free(ctx, pdev->image_indices);
+	fz_free(ctx, pdev->groups);
 	fz_free(ctx, pdev->alphas);
 	fz_free(ctx, pdev->gstates);
 }
 
-fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, pdf_obj *contents, pdf_obj *resources, const fz_matrix *ctm, fz_buffer *buf)
+fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, const fz_matrix *topctm, const fz_rect *mediabox, pdf_obj *resources, fz_buffer *buf)
 {
 	pdf_device *dev = fz_new_device(ctx, sizeof *dev);
 
-	dev->super.drop_imp = pdf_dev_drop_imp;
+	dev->super.close_device = pdf_dev_close_device;
+	dev->super.drop_device = pdf_dev_drop_device;
 
 	dev->super.fill_path = pdf_dev_fill_path;
 	dev->super.stroke_path = pdf_dev_stroke_path;
@@ -1319,11 +1113,10 @@ fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, pdf_obj *conte
 		if (!buf)
 			buf = fz_new_buffer(ctx, 256);
 		dev->doc = doc;
-		dev->contents = pdf_keep_obj(ctx, contents);
 		dev->resources = pdf_keep_obj(ctx, resources);
 		dev->gstates = fz_malloc_struct(ctx, gstate);
 		dev->gstates[0].buf = buf;
-		dev->gstates[0].ctm = *ctm;
+		dev->gstates[0].ctm = fz_identity; // XXX
 		dev->gstates[0].colorspace[0] = fz_device_gray(ctx);
 		dev->gstates[0].colorspace[1] = fz_device_gray(ctx);
 		dev->gstates[0].color[0][0] = 1;
@@ -1334,6 +1127,9 @@ fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, pdf_obj *conte
 		dev->gstates[0].horizontal_scaling = 100;
 		dev->num_gstates = 1;
 		dev->max_gstates = 1;
+
+		if (topctm != &fz_identity)
+			fz_buffer_printf(ctx, buf, "%M cm\n", topctm);
 	}
 	fz_catch(ctx)
 	{
@@ -1346,38 +1142,11 @@ fz_device *pdf_new_pdf_device(fz_context *ctx, pdf_document *doc, pdf_obj *conte
 	return (fz_device*)dev;
 }
 
-fz_device *pdf_page_write(fz_context *ctx, pdf_document *doc, pdf_page *page)
+fz_device *pdf_page_write(fz_context *ctx, pdf_document *doc,
+	const fz_rect *mediabox, pdf_obj **presources, fz_buffer **pcontents)
 {
-	pdf_obj *resources = pdf_dict_get(ctx, page->me, PDF_NAME_Resources);
-	fz_matrix ctm;
-	pdf_obj *obj;
-
-	fz_pre_translate(fz_scale(&ctm, 1, -1), 0, page->mediabox.y0-page->mediabox.y1);
-
-	if (resources == NULL)
-	{
-		resources = pdf_new_dict(ctx, doc, 0);
-		pdf_dict_put_drop(ctx, page->me, PDF_NAME_Resources, resources);
-	}
-
-	/* We always make a new object for page->contents here, in case
-	 * the existing one is an array, or is shared. */
-	obj = pdf_new_dict(ctx, doc, 0);
-	fz_try(ctx)
-	{
-		pdf_obj *new_contents = pdf_new_ref(ctx, doc, obj);
-		pdf_dict_put(ctx, page->me, PDF_NAME_Contents, new_contents);
-		pdf_drop_obj(ctx, page->contents);
-		page->contents = new_contents;
-	}
-	fz_always(ctx)
-	{
-		pdf_drop_obj(ctx, obj);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
-	}
-
-	return pdf_new_pdf_device(ctx, doc, page->contents, resources, &ctm, NULL);
+	fz_matrix pagectm = { 1, 0, 0, -1, -mediabox->x0, mediabox->y1 };
+	*presources = pdf_new_dict(ctx, doc, 0);
+	*pcontents = fz_new_buffer(ctx, 0);
+	return pdf_new_pdf_device(ctx, doc, &pagectm, mediabox, *presources, *pcontents);
 }
