@@ -48,7 +48,7 @@ typedef struct pdf_obj_num_s
 typedef struct pdf_obj_string_s
 {
 	pdf_obj super;
-	unsigned short len;
+	unsigned int len;
 	char buf[1];
 } pdf_obj_string;
 
@@ -142,14 +142,19 @@ pdf_new_real(fz_context *ctx, pdf_document *doc, float f)
 }
 
 pdf_obj *
-pdf_new_string(fz_context *ctx, pdf_document *doc, const char *str, int len)
+pdf_new_string(fz_context *ctx, pdf_document *doc, const char *str, size_t len)
 {
 	pdf_obj_string *obj;
+	unsigned int l = (unsigned int)len;
+
+	if ((size_t)l != len)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Overflow in pdf string");
+
 	obj = Memento_label(fz_malloc(ctx, offsetof(pdf_obj_string, buf) + len + 1), "pdf_obj(string)");
 	obj->super.refs = 1;
 	obj->super.kind = PDF_STRING;
 	obj->super.flags = 0;
-	obj->len = len;
+	obj->len = l;
 	memcpy(obj->buf, str, len);
 	obj->buf[len] = '\0';
 	return &obj->super;
@@ -197,73 +202,94 @@ pdf_obj *
 pdf_keep_obj(fz_context *ctx, pdf_obj *obj)
 {
 	if (obj >= PDF_OBJ__LIMIT)
-		obj->refs ++;
+	{
+		(void)fz_keep_imp16(ctx, obj, &obj->refs);
+	}
 	return obj;
 }
 
-int pdf_is_indirect(fz_context *ctx, pdf_obj *obj)
-{
-	return obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_INDIRECT : 0;
-}
+#define OBJ_IS_NULL(obj) \
+	(obj == PDF_OBJ_NULL)
+#define OBJ_IS_BOOL(obj) \
+	(obj == PDF_OBJ_TRUE || obj == PDF_OBJ_FALSE)
+#define OBJ_IS_INT(obj) \
+	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_INT : 0)
+#define OBJ_IS_REAL(obj) \
+	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_REAL : 0)
+#define OBJ_IS_NUMBER(obj) \
+	(obj >= PDF_OBJ__LIMIT ? (obj->kind == PDF_REAL || obj->kind == PDF_INT) : 0)
+#define OBJ_IS_STRING(obj) \
+	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_STRING : 0)
+#define OBJ_IS_NAME(obj) \
+	((obj >= PDF_OBJ__LIMIT) ? (obj->kind == PDF_NAME) : (obj != NULL && obj < PDF_OBJ_NAME__LIMIT))
+#define OBJ_IS_ARRAY(obj) \
+	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_ARRAY : 0)
+#define OBJ_IS_DICT(obj) \
+	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_DICT : 0)
+#define OBJ_IS_INDIRECT(obj) \
+	(obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_INDIRECT : 0)
 
 #define RESOLVE(obj) \
-	if (obj >= PDF_OBJ__LIMIT && obj->kind == PDF_INDIRECT) \
-		obj = pdf_resolve_indirect(ctx, obj); \
+	if (OBJ_IS_INDIRECT(obj)) \
+		obj = pdf_resolve_indirect_chain(ctx, obj); \
+
+int pdf_is_indirect(fz_context *ctx, pdf_obj *obj)
+{
+	return OBJ_IS_INDIRECT(obj);
+}
 
 int pdf_is_null(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj == PDF_OBJ_NULL;
+	return OBJ_IS_NULL(obj);
 }
 
 int pdf_is_bool(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj == PDF_OBJ_TRUE || obj == PDF_OBJ_FALSE;
+	return OBJ_IS_BOOL(obj);
 }
 
 int pdf_is_int(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_INT : 0;
+	return OBJ_IS_INT(obj);
 }
 
 int pdf_is_real(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_REAL : 0;
+	return OBJ_IS_REAL(obj);
 }
 
 int pdf_is_number(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj >= PDF_OBJ__LIMIT ? (obj->kind == PDF_REAL || obj->kind == PDF_INT) : 0;
+	return OBJ_IS_NUMBER(obj);
 }
 
 int pdf_is_string(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_STRING : 0;
+	return OBJ_IS_STRING(obj);
 }
 
 int pdf_is_name(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT)
-		return obj != NULL && obj < PDF_OBJ_NAME__LIMIT;
-	return obj->kind == PDF_NAME;
+	return OBJ_IS_NAME(obj);
 }
 
 int pdf_is_array(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_ARRAY : 0;
+	return OBJ_IS_ARRAY(obj);
 }
 
 int pdf_is_dict(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	return obj >= PDF_OBJ__LIMIT ? obj->kind == PDF_DICT : 0;
+	return OBJ_IS_DICT(obj);
 }
 
 int pdf_to_bool(fz_context *ctx, pdf_obj *obj)
@@ -311,19 +337,17 @@ float pdf_to_real(fz_context *ctx, pdf_obj *obj)
 char *pdf_to_name(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (!obj)
+	if (!OBJ_IS_NAME(obj))
 		return "";
 	if (obj < PDF_OBJ_NAME__LIMIT)
 		return PDF_NAMES[(intptr_t)obj];
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_NAME)
-		return "";
 	return NAME(obj)->n;
 }
 
 char *pdf_to_str_buf(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_STRING)
+	if (!OBJ_IS_STRING(obj))
 		return "";
 	return STRING(obj)->buf;
 }
@@ -331,21 +355,21 @@ char *pdf_to_str_buf(fz_context *ctx, pdf_obj *obj)
 int pdf_to_str_len(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_STRING)
+	if (!OBJ_IS_STRING(obj))
 		return 0;
 	return STRING(obj)->len;
 }
 
 void pdf_set_int(fz_context *ctx, pdf_obj *obj, int i)
 {
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_INT)
+	if (!OBJ_IS_INT(obj))
 		return;
 	NUM(obj)->u.i = i;
 }
 
 void pdf_set_int_offset(fz_context *ctx, pdf_obj *obj, fz_off_t i)
 {
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_INT)
+	if (!OBJ_IS_INT(obj))
 		return;
 	NUM(obj)->u.i = i;
 }
@@ -354,9 +378,9 @@ void pdf_set_int_offset(fz_context *ctx, pdf_obj *obj, fz_off_t i)
 void pdf_set_str_len(fz_context *ctx, pdf_obj *obj, int newlen)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_STRING)
+	if (!OBJ_IS_STRING(obj))
 		return; /* This should never happen */
-	if (newlen > STRING(obj)->len)
+	if (newlen < 0 || (unsigned int)newlen > STRING(obj)->len)
 		return; /* This should never happen */
 	STRING(obj)->len = newlen;
 }
@@ -369,25 +393,38 @@ pdf_obj *pdf_to_dict(fz_context *ctx, pdf_obj *obj)
 
 int pdf_to_num(fz_context *ctx, pdf_obj *obj)
 {
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_INDIRECT)
+	if (!OBJ_IS_INDIRECT(obj))
 		return 0;
 	return REF(obj)->num;
 }
 
 int pdf_to_gen(fz_context *ctx, pdf_obj *obj)
 {
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_INDIRECT)
+	if (!OBJ_IS_INDIRECT(obj))
 		return 0;
 	return REF(obj)->gen;
 }
 
 pdf_document *pdf_get_indirect_document(fz_context *ctx, pdf_obj *obj)
 {
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_INDIRECT)
+	if (!OBJ_IS_INDIRECT(obj))
 		return NULL;
 	return REF(obj)->doc;
 }
 
+pdf_document *pdf_get_bound_document(fz_context *ctx, pdf_obj *obj)
+{
+	if (obj < PDF_OBJ__LIMIT)
+		return NULL;
+	if (obj->kind == PDF_INDIRECT)
+		return REF(obj)->doc;
+	if (obj->kind == PDF_ARRAY)
+		return ARRAY(obj)->doc;
+	if (obj->kind == PDF_DICT)
+		return DICT(obj)->doc;
+	return NULL;
+}
+
 int pdf_objcmp_resolve(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 {
 	RESOLVE(a);
@@ -569,8 +606,8 @@ pdf_copy_array(fz_context *ctx, pdf_obj *obj)
 	int n;
 
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_ARRAY)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "assert: not an array (%s)", pdf_objkindstr(obj));
+	if (!OBJ_IS_ARRAY(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not an array (%s)", pdf_objkindstr(obj));
 
 	doc = ARRAY(obj)->doc;
 
@@ -586,7 +623,7 @@ int
 pdf_array_len(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_ARRAY)
+	if (!OBJ_IS_ARRAY(obj))
 		return 0;
 	return ARRAY(obj)->len;
 }
@@ -595,7 +632,7 @@ pdf_obj *
 pdf_array_get(fz_context *ctx, pdf_obj *obj, int i)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_ARRAY)
+	if (!OBJ_IS_ARRAY(obj))
 		return NULL;
 	if (i < 0 || i >= ARRAY(obj)->len)
 		return NULL;
@@ -604,7 +641,7 @@ pdf_array_get(fz_context *ctx, pdf_obj *obj, int i)
 
 static void prepare_object_for_alteration(fz_context *ctx, pdf_obj *obj, pdf_obj *val)
 {
-	pdf_document *doc;
+	pdf_document *doc, *val_doc;
 	int parent;
 
 	/*
@@ -627,6 +664,14 @@ static void prepare_object_for_alteration(fz_context *ctx, pdf_obj *obj, pdf_obj
 	default:
 		return;
 	}
+
+	if (val)
+	{
+		val_doc = pdf_get_bound_document(ctx, val);
+		if (doc && val_doc && val_doc != doc)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "container and item belong to different documents");
+	}
+
 	/*
 		parent_num = 0 while an object is being parsed from the file.
 		No further action is necessary.
@@ -647,23 +692,17 @@ void
 pdf_array_put(fz_context *ctx, pdf_obj *obj, int i, pdf_obj *item)
 {
 	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		prepare_object_for_alteration(ctx, obj, item);
+	if (!OBJ_IS_ARRAY(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not an array (%s)", pdf_objkindstr(obj));
+	if (i < 0 || i >= ARRAY(obj)->len)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "index out of bounds");
 
-		if (obj->kind != PDF_ARRAY)
-			fz_warn(ctx, "assert: not an array (%s)", pdf_objkindstr(obj));
-		else if (i < 0)
-			fz_warn(ctx, "assert: index %d < 0", i);
-		else if (i >= ARRAY(obj)->len)
-			fz_warn(ctx, "assert: index %d > length %d", i, ARRAY(obj)->len);
-		else
-		{
-			pdf_drop_obj(ctx, ARRAY(obj)->items[i]);
-			ARRAY(obj)->items[i] = pdf_keep_obj(ctx, item);
-		}
-	}
-	return; /* Can't warn :( */
+	if (!item)
+		item = PDF_OBJ_NULL;
+
+	prepare_object_for_alteration(ctx, obj, item);
+	pdf_drop_obj(ctx, ARRAY(obj)->items[i]);
+	ARRAY(obj)->items[i] = pdf_keep_obj(ctx, item);
 }
 
 void
@@ -677,21 +716,17 @@ void
 pdf_array_push(fz_context *ctx, pdf_obj *obj, pdf_obj *item)
 {
 	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		prepare_object_for_alteration(ctx, obj, item);
+	if (!OBJ_IS_ARRAY(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not an array (%s)", pdf_objkindstr(obj));
 
-		if (obj->kind != PDF_ARRAY)
-			fz_warn(ctx, "assert: not an array (%s)", pdf_objkindstr(obj));
-		else
-		{
-			if (ARRAY(obj)->len + 1 > ARRAY(obj)->cap)
-				pdf_array_grow(ctx, ARRAY(obj));
-			ARRAY(obj)->items[ARRAY(obj)->len] = pdf_keep_obj(ctx, item);
-			ARRAY(obj)->len++;
-		}
-	}
-	return; /* Can't warn :( */
+	if (!item)
+		item = PDF_OBJ_NULL;
+
+	prepare_object_for_alteration(ctx, obj, item);
+	if (ARRAY(obj)->len + 1 > ARRAY(obj)->cap)
+		pdf_array_grow(ctx, ARRAY(obj));
+	ARRAY(obj)->items[ARRAY(obj)->len] = pdf_keep_obj(ctx, item);
+	ARRAY(obj)->len++;
 }
 
 void
@@ -713,24 +748,20 @@ void
 pdf_array_insert(fz_context *ctx, pdf_obj *obj, pdf_obj *item, int i)
 {
 	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		prepare_object_for_alteration(ctx, obj, item);
+	if (!OBJ_IS_ARRAY(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not an array (%s)", pdf_objkindstr(obj));
+	if (i < 0 || i > ARRAY(obj)->len)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "index out of bounds");
 
-		if (obj->kind != PDF_ARRAY)
-			fz_warn(ctx, "assert: not an array (%s)", pdf_objkindstr(obj));
-		else
-		{
-			if (i < 0 || i > ARRAY(obj)->len)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "attempt to insert object %d in array of length %d", i, ARRAY(obj)->len);
-			if (ARRAY(obj)->len + 1 > ARRAY(obj)->cap)
-				pdf_array_grow(ctx, ARRAY(obj));
-			memmove(ARRAY(obj)->items + i + 1, ARRAY(obj)->items + i, (ARRAY(obj)->len - i) * sizeof(pdf_obj*));
-			ARRAY(obj)->items[i] = pdf_keep_obj(ctx, item);
-			ARRAY(obj)->len++;
-		}
-	}
-	return; /* Can't warn :( */
+	if (!item)
+		item = PDF_OBJ_NULL;
+
+	prepare_object_for_alteration(ctx, obj, item);
+	if (ARRAY(obj)->len + 1 > ARRAY(obj)->cap)
+		pdf_array_grow(ctx, ARRAY(obj));
+	memmove(ARRAY(obj)->items + i + 1, ARRAY(obj)->items + i, (ARRAY(obj)->len - i) * sizeof(pdf_obj*));
+	ARRAY(obj)->items[i] = pdf_keep_obj(ctx, item);
+	ARRAY(obj)->len++;
 }
 
 void
@@ -752,19 +783,15 @@ void
 pdf_array_delete(fz_context *ctx, pdf_obj *obj, int i)
 {
 	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		if (obj->kind != PDF_ARRAY)
-			fz_warn(ctx, "assert: not an array (%s)", pdf_objkindstr(obj));
-		else
-		{
-			pdf_drop_obj(ctx, ARRAY(obj)->items[i]);
-			ARRAY(obj)->items[i] = 0;
-			ARRAY(obj)->len--;
-			memmove(ARRAY(obj)->items + i, ARRAY(obj)->items + i + 1, (ARRAY(obj)->len - i) * sizeof(pdf_obj*));
-		}
-	}
-	return; /* Can't warn :( */
+	if (!OBJ_IS_ARRAY(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not an array (%s)", pdf_objkindstr(obj));
+	if (i < 0 || i >= ARRAY(obj)->len)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "index out of bounds");
+
+	pdf_drop_obj(ctx, ARRAY(obj)->items[i]);
+	ARRAY(obj)->items[i] = 0;
+	ARRAY(obj)->len--;
+	memmove(ARRAY(obj)->items + i, ARRAY(obj)->items + i + 1, (ARRAY(obj)->len - i) * sizeof(pdf_obj*));
 }
 
 int
@@ -780,6 +807,19 @@ pdf_array_contains(fz_context *ctx, pdf_obj *arr, pdf_obj *obj)
 	return 0;
 }
 
+int
+pdf_array_find(fz_context *ctx, pdf_obj *arr, pdf_obj *obj)
+{
+	int i, len;
+
+	len = pdf_array_len(ctx, arr);
+	for (i = 0; i < len; i++)
+		if (!pdf_objcmp(ctx, pdf_array_get(ctx, arr, i), obj))
+			return i;
+
+	return -1;
+}
+
 pdf_obj *pdf_new_rect(fz_context *ctx, pdf_document *doc, const fz_rect *rect)
 {
 	pdf_obj *arr = NULL;
@@ -953,32 +993,28 @@ pdf_dict_grow(fz_context *ctx, pdf_obj *obj)
 pdf_obj *
 pdf_copy_dict(fz_context *ctx, pdf_obj *obj)
 {
+	pdf_document *doc;
 	pdf_obj *dict;
 	int i, n;
 
 	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		pdf_document *doc = DICT(obj)->doc;
+	if (!OBJ_IS_DICT(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
 
-		if (obj->kind != PDF_DICT)
-			fz_warn(ctx, "assert: not a dict (%s)", pdf_objkindstr(obj));
+	doc = DICT(obj)->doc;
+	n = pdf_dict_len(ctx, obj);
+	dict = pdf_new_dict(ctx, doc, n);
+	for (i = 0; i < n; i++)
+		pdf_dict_put(ctx, dict, pdf_dict_get_key(ctx, obj, i), pdf_dict_get_val(ctx, obj, i));
 
-		n = pdf_dict_len(ctx, obj);
-		dict = pdf_new_dict(ctx, doc, n);
-		for (i = 0; i < n; i++)
-			pdf_dict_put(ctx, dict, pdf_dict_get_key(ctx, obj, i), pdf_dict_get_val(ctx, obj, i));
-
-		return dict;
-	}
-	return NULL; /* Can't warn :( */
+	return dict;
 }
 
 int
 pdf_dict_len(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
+	if (!OBJ_IS_DICT(obj))
 		return 0;
 	return DICT(obj)->len;
 }
@@ -987,7 +1023,7 @@ pdf_obj *
 pdf_dict_get_key(fz_context *ctx, pdf_obj *obj, int i)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
+	if (!OBJ_IS_DICT(obj))
 		return NULL;
 	if (i < 0 || i >= DICT(obj)->len)
 		return NULL;
@@ -998,7 +1034,7 @@ pdf_obj *
 pdf_dict_get_val(fz_context *ctx, pdf_obj *obj, int i)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
+	if (!OBJ_IS_DICT(obj))
 		return NULL;
 	if (i < 0 || i >= DICT(obj)->len)
 		return NULL;
@@ -1006,22 +1042,26 @@ pdf_dict_get_val(fz_context *ctx, pdf_obj *obj, int i)
 }
 
 void
-pdf_dict_put_val_drop(fz_context *ctx, pdf_obj *obj, int i, pdf_obj *new_obj)
+pdf_dict_put_val_drop(fz_context *ctx, pdf_obj *obj, int i, pdf_obj *val)
 {
 	RESOLVE(obj);
-	if (!obj || obj->kind != PDF_DICT)
+	if (!OBJ_IS_DICT(obj))
 	{
-		pdf_drop_obj(ctx, new_obj);
-		return;
+		pdf_drop_obj(ctx, val);
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
 	}
 	if (i < 0 || i >= DICT(obj)->len)
 	{
 		/* FIXME: Should probably extend the dict here */
-		pdf_drop_obj(ctx, new_obj);
-		return;
+		pdf_drop_obj(ctx, val);
+		fz_throw(ctx, FZ_ERROR_GENERIC, "index out of bounds");
 	}
+
+	if (!val)
+		val = PDF_OBJ_NULL;
+
 	pdf_drop_obj(ctx, DICT(obj)->items[i].v);
-	DICT(obj)->items[i].v = new_obj;
+	DICT(obj)->items[i].v = val;
 }
 
 /* Returns 0 <= i < len for key found. Returns -1-len < i <= -1 for key
@@ -1124,49 +1164,48 @@ pdf_dict_gets(fz_context *ctx, pdf_obj *obj, const char *key)
 	int i;
 
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
+	if (!OBJ_IS_DICT(obj))
+		return NULL;
+	if (!key)
 		return NULL;
 
 	i = pdf_dict_finds(ctx, obj, key);
 	if (i >= 0)
 		return DICT(obj)->items[i].v;
-
 	return NULL;
 }
 
 pdf_obj *
 pdf_dict_getp(fz_context *ctx, pdf_obj *obj, const char *keys)
 {
+	char buf[256];
+	char *k, *e;
+
 	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
+	if (!OBJ_IS_DICT(obj))
+		return NULL;
+	if (strlen(keys)+1 > 256)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "path too long");
+
+	strcpy(buf, keys);
+
+	e = buf;
+	while (*e && obj)
 	{
-		char buf[256];
-		char *k, *e;
+		k = e;
+		while (*e != '/' && *e != '\0')
+			e++;
 
-		if (strlen(keys)+1 > 256)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "buffer overflow in pdf_dict_getp");
-
-		strcpy(buf, keys);
-
-		e = buf;
-		while (*e && obj)
+		if (*e == '/')
 		{
-			k = e;
-			while (*e != '/' && *e != '\0')
-				e++;
-
-			if (*e == '/')
-			{
-				*e = '\0';
-				e++;
-			}
-
-			obj = pdf_dict_gets(ctx, obj, k);
+			*e = '\0';
+			e++;
 		}
 
-		return obj;
+		obj = pdf_dict_gets(ctx, obj, k);
 	}
-	return NULL; /* Can't warn */
+
+	return obj;
 }
 
 pdf_obj *
@@ -1192,10 +1231,12 @@ pdf_dict_get(fz_context *ctx, pdf_obj *obj, pdf_obj *key)
 	int i;
 
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
+	if (!OBJ_IS_DICT(obj))
+		return NULL;
+	if (!OBJ_IS_NAME(key))
 		return NULL;
 
-	if (key < PDF_OBJ__LIMIT)
+	if (key < PDF_OBJ_NAME__LIMIT)
 		i = pdf_dict_find(ctx, obj, key);
 	else
 		i = pdf_dict_finds(ctx, obj, pdf_to_name(ctx, key));
@@ -1224,75 +1265,84 @@ pdf_dict_geta(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *abbrev)
 	return pdf_dict_get(ctx, obj, abbrev);
 }
 
+static void
+pdf_dict_get_put(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *val, pdf_obj **old_val)
+{
+	int i;
+
+	if (old_val)
+		*old_val = NULL;
+
+	RESOLVE(obj);
+	if (!OBJ_IS_DICT(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
+	if (!OBJ_IS_NAME(key))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "key is not a name (%s)", pdf_objkindstr(obj));
+
+	if (!val)
+		val = PDF_OBJ_NULL;
+
+	if (DICT(obj)->len > 100 && !(obj->flags & PDF_FLAGS_SORTED))
+		pdf_sort_dict(ctx, obj);
+
+	if (key < PDF_OBJ_NAME__LIMIT)
+		i = pdf_dict_find(ctx, obj, key);
+	else
+		i = pdf_dict_finds(ctx, obj, pdf_to_name(ctx, key));
+
+	prepare_object_for_alteration(ctx, obj, val);
+
+	if (i >= 0 && i < DICT(obj)->len)
+	{
+		if (DICT(obj)->items[i].v != val)
+		{
+			pdf_obj *d = DICT(obj)->items[i].v;
+			DICT(obj)->items[i].v = pdf_keep_obj(ctx, val);
+			if (old_val)
+				*old_val = d;
+			else
+				pdf_drop_obj(ctx, d);
+		}
+	}
+	else
+	{
+		if (DICT(obj)->len + 1 > DICT(obj)->cap)
+			pdf_dict_grow(ctx, obj);
+
+		i = -1-i;
+		if ((obj->flags & PDF_FLAGS_SORTED) && DICT(obj)->len > 0)
+			memmove(&DICT(obj)->items[i + 1],
+					&DICT(obj)->items[i],
+					(DICT(obj)->len - i) * sizeof(struct keyval));
+
+		DICT(obj)->items[i].k = pdf_keep_obj(ctx, key);
+		DICT(obj)->items[i].v = pdf_keep_obj(ctx, val);
+		DICT(obj)->len ++;
+	}
+}
+
 void
 pdf_dict_put(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *val)
 {
-	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		int i;
-
-		if (obj->kind != PDF_DICT)
-		{
-			fz_warn(ctx, "assert: not a dict (%s)", pdf_objkindstr(obj));
-			return;
-		}
-
-		RESOLVE(key);
-		if (!key || (key >= PDF_OBJ__LIMIT && key->kind != PDF_NAME))
-		{
-			fz_warn(ctx, "assert: key is not a name (%s)", pdf_objkindstr(obj));
-			return;
-		}
-
-		if (!val)
-		{
-			fz_warn(ctx, "assert: val does not exist for key (%s)", pdf_to_name(ctx, key));
-			return;
-		}
-
-		prepare_object_for_alteration(ctx, obj, val);
-
-		if (DICT(obj)->len > 100 && !(obj->flags & PDF_FLAGS_SORTED))
-			pdf_sort_dict(ctx, obj);
-
-		if (key < PDF_OBJ__LIMIT)
-			i = pdf_dict_find(ctx, obj, key);
-		else
-			i = pdf_dict_finds(ctx, obj, pdf_to_name(ctx, key));
-		if (i >= 0 && i < DICT(obj)->len)
-		{
-			if (DICT(obj)->items[i].v != val)
-			{
-				pdf_obj *d = DICT(obj)->items[i].v;
-				DICT(obj)->items[i].v = pdf_keep_obj(ctx, val);
-				pdf_drop_obj(ctx, d);
-			}
-		}
-		else
-		{
-			if (DICT(obj)->len + 1 > DICT(obj)->cap)
-				pdf_dict_grow(ctx, obj);
-
-			i = -1-i;
-			if ((obj->flags & PDF_FLAGS_SORTED) && DICT(obj)->len > 0)
-				memmove(&DICT(obj)->items[i + 1],
-						&DICT(obj)->items[i],
-						(DICT(obj)->len - i) * sizeof(struct keyval));
-
-			DICT(obj)->items[i].k = pdf_keep_obj(ctx, key);
-			DICT(obj)->items[i].v = pdf_keep_obj(ctx, val);
-			DICT(obj)->len ++;
-		}
-	}
-	return; /* Can't warn :( */
+	pdf_dict_get_put(ctx, obj, key, val, NULL);
 }
 
 void
 pdf_dict_put_drop(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *val)
 {
 	fz_try(ctx)
-		pdf_dict_put(ctx, obj, key, val);
+		pdf_dict_get_put(ctx, obj, key, val, NULL);
+	fz_always(ctx)
+		pdf_drop_obj(ctx, val);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
+void
+pdf_dict_get_put_drop(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *val, pdf_obj **old_val)
+{
+	fz_try(ctx)
+		pdf_dict_get_put(ctx, obj, key, val, old_val);
 	fz_always(ctx)
 		pdf_drop_obj(ctx, val);
 	fz_catch(ctx)
@@ -1306,8 +1356,8 @@ pdf_dict_puts(fz_context *ctx, pdf_obj *obj, const char *key, pdf_obj *val)
 	pdf_obj *keyobj;
 
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dictionary (%s)", pdf_objkindstr(obj));
+	if (!OBJ_IS_DICT(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
 
 	doc = DICT(obj)->doc;
 	keyobj = pdf_new_name(ctx, doc, key);
@@ -1327,8 +1377,8 @@ pdf_dict_puts_drop(fz_context *ctx, pdf_obj *obj, const char *key, pdf_obj *val)
 	pdf_obj *keyobj;
 
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dictionary (%s)", pdf_objkindstr(obj));
+	if (!OBJ_IS_DICT(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
 
 	doc = DICT(obj)->doc;
 	keyobj = pdf_new_name(ctx, doc, key);
@@ -1357,14 +1407,12 @@ pdf_dict_putp(fz_context *ctx, pdf_obj *obj, const char *keys, pdf_obj *val)
 	pdf_obj *cobj = NULL;
 
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dictionary (%s)", pdf_objkindstr(obj));
-
-	doc = DICT(obj)->doc;
-
+	if (!OBJ_IS_DICT(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
 	if (strlen(keys)+1 > 256)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "buffer overflow in pdf_dict_putp");
 
+	doc = DICT(obj)->doc;
 	strcpy(buf, keys);
 
 	e = buf;
@@ -1429,8 +1477,8 @@ pdf_dict_vputl(fz_context *ctx, pdf_obj *obj, pdf_obj *val, va_list keys)
 	pdf_document *doc;
 
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dictionary (%s)", pdf_objkindstr(obj));
+	if (!OBJ_IS_DICT(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
 
 	doc = DICT(obj)->doc;
 
@@ -1499,48 +1547,43 @@ pdf_dict_putl_drop(fz_context *ctx, pdf_obj *obj, pdf_obj *val, ...)
 void
 pdf_dict_dels(fz_context *ctx, pdf_obj *obj, const char *key)
 {
-	RESOLVE(obj);
-	if (obj >= PDF_OBJ__LIMIT)
-	{
-		prepare_object_for_alteration(ctx, obj, NULL);
+	int i;
 
-		if (obj->kind != PDF_DICT)
-			fz_warn(ctx, "assert: not a dict (%s)", pdf_objkindstr(obj));
-		else
-		{
-			int i = pdf_dict_finds(ctx, obj, key);
-			if (i >= 0)
-			{
-				pdf_drop_obj(ctx, DICT(obj)->items[i].k);
-				pdf_drop_obj(ctx, DICT(obj)->items[i].v);
-				obj->flags &= ~PDF_FLAGS_SORTED;
-				DICT(obj)->items[i] = DICT(obj)->items[DICT(obj)->len-1];
-				DICT(obj)->len --;
-			}
-		}
+	RESOLVE(obj);
+	if (!OBJ_IS_DICT(obj))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "not a dict (%s)", pdf_objkindstr(obj));
+	if (!key)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "key is null");
+
+	prepare_object_for_alteration(ctx, obj, NULL);
+	i = pdf_dict_finds(ctx, obj, key);
+	if (i >= 0)
+	{
+		pdf_drop_obj(ctx, DICT(obj)->items[i].k);
+		pdf_drop_obj(ctx, DICT(obj)->items[i].v);
+		obj->flags &= ~PDF_FLAGS_SORTED;
+		DICT(obj)->items[i] = DICT(obj)->items[DICT(obj)->len-1];
+		DICT(obj)->len --;
 	}
-	return; /* Can't warn :( */
 }
 
 void
 pdf_dict_del(fz_context *ctx, pdf_obj *obj, pdf_obj *key)
 {
-	RESOLVE(key);
-	if (!key)
-		return; /* Can't warn */
+	if (!OBJ_IS_NAME(key))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "key is not a name (%s)", pdf_objkindstr(key));
 
-	if (key < PDF_OBJ__LIMIT)
+	if (key < PDF_OBJ_NAME__LIMIT)
 		pdf_dict_dels(ctx, obj, PDF_NAMES[(intptr_t)key]);
-	else if (key->kind == PDF_NAME)
+	else
 		pdf_dict_dels(ctx, obj, NAME(key)->n);
-	/* else Can't warn */
 }
 
 void
 pdf_sort_dict(fz_context *ctx, pdf_obj *obj)
 {
 	RESOLVE(obj);
-	if (obj < PDF_OBJ__LIMIT || obj->kind != PDF_DICT)
+	if (OBJ_IS_DICT(obj))
 		return;
 	if (!(obj->flags & PDF_FLAGS_SORTED))
 	{
@@ -1702,14 +1745,15 @@ pdf_drop_obj(fz_context *ctx, pdf_obj *obj)
 {
 	if (obj >= PDF_OBJ__LIMIT)
 	{
-		if (--obj->refs)
-			return;
-		if (obj->kind == PDF_ARRAY)
-			pdf_drop_array(ctx, obj);
-		else if (obj->kind == PDF_DICT)
-			pdf_drop_dict(ctx, obj);
-		else
-			fz_free(ctx, obj);
+		if (fz_drop_imp16(ctx, obj, &obj->refs))
+		{
+			if (obj->kind == PDF_ARRAY)
+				pdf_drop_array(ctx, obj);
+			else if (obj->kind == PDF_DICT)
+				pdf_drop_dict(ctx, obj);
+			else
+				fz_free(ctx, obj);
+		}
 	}
 }
 
@@ -1745,6 +1789,8 @@ int pdf_obj_parent_num(fz_context *ctx, pdf_obj *obj)
 
 	switch(obj->kind)
 	{
+	case PDF_INDIRECT:
+		return REF(obj)->num;
 	case PDF_ARRAY:
 		return ARRAY(obj)->parent_num;
 	case PDF_DICT:
@@ -1951,15 +1997,19 @@ static void fmt_array(fz_context *ctx, struct fmt *fmt, pdf_obj *obj)
 		fmt_putc(ctx, fmt, ']');
 	}
 	else {
-		fmt_puts(ctx, fmt, "[ ");
+		fmt_putc(ctx, fmt, '[');
+		fmt->indent ++;
 		for (i = 0; i < n; i++) {
 			if (fmt->col > 60) {
 				fmt_putc(ctx, fmt, '\n');
 				fmt_indent(ctx, fmt);
+			} else {
+				fmt_putc(ctx, fmt, ' ');
 			}
 			fmt_obj(ctx, fmt, pdf_array_get(ctx, obj, i));
-			fmt_putc(ctx, fmt, ' ');
 		}
+		fmt->indent --;
+		fmt_putc(ctx, fmt, ' ');
 		fmt_putc(ctx, fmt, ']');
 		fmt_sep(ctx, fmt);
 	}
@@ -2078,8 +2128,7 @@ pdf_sprint_obj(fz_context *ctx, char *s, int n, pdf_obj *obj, int tight)
 	return fmt.len;
 }
 
-int
-pdf_fprint_obj(fz_context *ctx, FILE *fp, pdf_obj *obj, int tight)
+int pdf_print_obj(fz_context *ctx, fz_output *out, pdf_obj *obj, int tight)
 {
 	char buf[1024];
 	char *ptr;
@@ -2089,56 +2138,18 @@ pdf_fprint_obj(fz_context *ctx, FILE *fp, pdf_obj *obj, int tight)
 	if ((n + 1) < sizeof buf)
 	{
 		pdf_sprint_obj(ctx, buf, sizeof buf, obj, tight);
-		fputs(buf, fp);
-		fputc('\n', fp);
+		fz_write(ctx, out, buf, n);
 	}
 	else
 	{
 		ptr = fz_malloc(ctx, n + 1);
 		pdf_sprint_obj(ctx, ptr, n + 1, obj, tight);
-		fputs(ptr, fp);
-		fputc('\n', fp);
+		fz_write(ctx, out, ptr, n);
 		fz_free(ctx, ptr);
 	}
 	return n;
 }
 
-int pdf_output_obj(fz_context *ctx, fz_output *out, pdf_obj *obj, int tight)
-{
-	char buf[1024];
-	char *ptr;
-	int n;
-
-	n = pdf_sprint_obj(ctx, NULL, 0, obj, tight);
-	if ((n + 1) < sizeof buf)
-	{
-		pdf_sprint_obj(ctx, buf, sizeof buf, obj, tight);
-		fz_puts(ctx, out, buf);
-	}
-	else
-	{
-		ptr = fz_malloc(ctx, n + 1);
-		pdf_sprint_obj(ctx, ptr, n + 1, obj, tight);
-		fz_puts(ctx, out, buf);
-		fz_free(ctx, ptr);
-	}
-	return n;
-}
-
-#ifndef NDEBUG
-void
-pdf_print_obj(fz_context *ctx, pdf_obj *obj)
-{
-	pdf_fprint_obj(ctx, stdout, obj, 0);
-}
-
-void
-pdf_print_ref(fz_context *ctx, pdf_obj *ref)
-{
-	pdf_print_obj(ctx, pdf_resolve_indirect(ctx, ref));
-}
-#endif
-
 int pdf_obj_refs(fz_context *ctx, pdf_obj *ref)
 {
 	return (ref >= PDF_OBJ__LIMIT ? ref->refs : 0);
