@@ -548,10 +548,12 @@ pdf_out_BI(fz_context *ctx, pdf_processor *proc, fz_image *img)
 	fz_compressed_buffer *cbuf;
 	fz_buffer *buf;
 	int i;
+	unsigned char *data;
+	size_t len;
 
 	if (img == NULL)
 		return;
-	cbuf = img->buffer;
+	cbuf = fz_compressed_image_buffer(ctx, img);
 	if (cbuf == NULL)
 		return;
 	buf = cbuf->buffer;
@@ -662,21 +664,23 @@ pdf_out_BI(fz_context *ctx, pdf_processor *proc, fz_image *img)
 	}
 
 	fz_printf(ctx, out, "ID\n");
+	len = fz_buffer_storage(ctx, buf, &data);
 	if (ahx)
 	{
-		for (i = 0; i < buf->len; ++i)
+		size_t z;
+		for (z = 0; z < len; ++z)
 		{
-			int c = buf->data[i];
+			int c = data[z];
 			fz_putc(ctx, out, "0123456789abcdef"[(c >> 4) & 0xf]);
 			fz_putc(ctx, out, "0123456789abcdef"[c & 0xf]);
-			if ((i & 31) == 31)
+			if ((z & 31) == 31)
 				fz_putc(ctx, out, '\n');
 		}
 		fz_putc(ctx, out, '>');
 	}
 	else
 	{
-		fz_write(ctx, out, buf->data, buf->len);
+		fz_write(ctx, out, data, len);
 	}
 	fz_printf(ctx, out, "\nEI\n");
 }
@@ -760,7 +764,7 @@ pdf_out_EX(fz_context *ctx, pdf_processor *proc)
 }
 
 static void
-pdf_drop_imp_output_processor(fz_context *ctx, pdf_processor *proc)
+pdf_drop_output_processor(fz_context *ctx, pdf_processor *proc)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
 	fz_drop_output(ctx, out);
@@ -771,7 +775,7 @@ pdf_new_output_processor(fz_context *ctx, fz_output *out, int ahxencode)
 {
 	pdf_output_processor *proc = pdf_new_processor(ctx, sizeof *proc);
 	{
-		proc->super.drop_imp = pdf_drop_imp_output_processor;
+		proc->super.drop_processor = pdf_drop_output_processor;
 
 		/* general graphics state */
 		proc->super.op_w = pdf_out_w;
